/*
 *  $Id: netlogwriter.h,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: NetLogger File Writer
 *
 */


#ifndef NETLOGWRITER_H
#define NETLOGWRITER_H


#include "tdsystem.h"
#include "timedthread.h"
#include "networkmonitorinterface.h"
#include "streammonitorreport.h"


namespace Coral {


/**
 * This class implements a writer for NetworkMonitorReports. The reports will
 * be written to a file in given intervals.
 *
 * @short   Network Log Writer
 * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
 * @version 1.0
 */       
class NetLogWriter : public TimedThread
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor for a new NetLogWriter. Note: NetLogWriter will *not*
     * be started automatically. You have to call start()!
     *
     * @param name Name of log file to write.
     * @param interval Interval in microseonds the log is written.
     * @param maxCount Number of entries to be written.
     * @param monitor NetworkMonitorInterface to get the reports from.
     * @param smReport StreamMonitorReport (default NULL).
     *
     * @see Thread#start
     */
   NetLogWriter(const char*              name,
                const card64             interval,
                NetworkMonitorInterface* monitor,
                StreamMonitorReport*     smReport = NULL);

   /**
     * Destructor.
     */
   ~NetLogWriter();


   // ====== Status functions ===============================================
   /**
     * Check, if NetLogWriter is ready.
     *
     * @return true, if ready; false, if not.
     */  
   inline bool ready() const;

   /**
     * Get time stamp of file (seconds since 01-Jan-1970).
     *
     * @return Time stamp.
     */
   inline card64 getTimeStamp() const; 

   /**
     * Get interval of the entries in file.
     *
     * @return Interval in microseconds.
     */
   inline card64 getInterval() const;

   /**
     * Get number of entries written into file.
     *
     * @return Number of entries written.
     */
   inline card64 getCount() const;
   

   // ====== Private data ===================================================
   private:
   void timerEvent();
   
   
   FILE*                    FileDescriptor;
   NetworkMonitorInterface* Monitor;
   StreamMonitorReport*     SMReport;
   card64                   Count;
   card64                   TimeStamp;
   card64                   Interval;
};


}


#include "netlogwriter.icc"


#endif
