/*
 *  $Id: streammonitorreport.cc,v 1.2 2001/11/02 12:38:20 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Stream Monitor Report
 *
 */


#include "tdsystem.h"
#include "streammonitorreport.h"
#include "streamreport.h"
#include "tools.h"


namespace Coral {


// ###### Constructor #######################################################
StreamMonitorReport::StreamMonitorReport(const cardinal maxCount)
{
   AcceptedSources       = NULL;
   AcceptedDestinations  = NULL;
   AcceptAllSources      = false;
   AcceptAllDestinations = false;
   PrintAccepted         = false;
   PrintRejected         = false;
   MaxCount              = maxCount;
   Count                 = 0;
   Report                = new StreamReport*[maxCount];
   if(Report == NULL) {
      cerr << "WARNING: StreamMonitorReport::StreamMonitorReport() - Out of memory!" << endl;
      exit(1);
   }
   for(cardinal i = 0;i < MaxCount;i++) {
      Report[i] = NULL;
   }
   for(cardinal i = 0;i < MaxCount;i++) {
      Report[i] = new StreamReport;
      if(Report[i] == NULL) {
         cerr << "WARNING: StreamMonitorReport::StreamMonitorReport() - Out of memory!" << endl;
         for(cardinal j = 0;j < MaxCount;j++) {
            if(Report[i]) delete Report[i];
         }
         delete Report;
         Report = NULL;
         return;
      }
   }
   reset();
}


// ###### Destructor ########################################################
StreamMonitorReport::~StreamMonitorReport()
{
   if(Report != NULL) {
      for(cardinal i = 0;i < MaxCount;i++) {
         if(Report[i]) delete Report[i];
      }
      delete Report;
      Report = NULL;
   }
}


// ###### Reset #############################################################
void StreamMonitorReport::reset()
{
   if(Report == NULL)
      return;

   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      TotalBytesRaw[i]     = 0;
      TotalBytesPayload[i] = 0;
      TotalPackets[i]      = 0;
   }
   Count = 0;
}


// ###### Check, if address is accepted by accept list ######################
bool StreamMonitorReport::isAccepted(multiset<PortableAddress>* set, const PortableAddress& address)
{
   // ====== First, try to find address with port 0 (all ports) in accept list.
   PortableAddress address2 = address;
   address2.Port = 0;
   multiset<PortableAddress>::iterator found1 = set->find(address2);
   if(found1 != set->end()) {
      return(true);
   }

   // ====== Second, try to find address with port in accept list.
   multiset<PortableAddress>::iterator found2 = set->find(address);
   if(found2 != set->end()) {
      return(true);
   }

   return(false);
}


// ###### Update report with new packet #####################################
void StreamMonitorReport::update(const card8            protocol3,
                                 const card8            protocol4,
                                 const PortableAddress& source,
                                 const PortableAddress& destination,
                                 const card32           flowLabel,
                                 const card8            trafficClass,
                                 const card32           streamID,                                 
                                 const cardinal         rawLength,
                                 const cardinal         payloadLength)
{
   if(Report == NULL)
      return;

   // ====== Update total values ============================================
   cardinal tosIndex = TrafficClassValues::getIndexForTrafficClass(trafficClass);
   TotalBytesRaw[tosIndex]     += rawLength;
   TotalBytesPayload[tosIndex] += payloadLength;
   TotalPackets[tosIndex]++;

   bool acceptedSource      = AcceptAllSources;
   bool acceptedDestination = AcceptAllDestinations;
   if((!acceptedSource) && (AcceptedSources != NULL) && (AcceptedSources->size() > 0)) {
      acceptedSource = isAccepted(AcceptedSources,source);
   }
   if((!acceptedDestination) && (AcceptedDestinations != NULL) && (AcceptedDestinations->size() > 0)) {
      acceptedDestination = isAccepted(AcceptedDestinations,destination);
   }

   if((acceptedSource == false) || (acceptedDestination == false)) {
      if(PrintRejected) {
         InternetAddress sa1(source);
         InternetAddress sa2(destination);
         char str[32];
         snprintf((char*)&str,sizeof(str),"$%05x/$%02x, sid=%d proto=$%02x",
                  flowLabel,trafficClass,streamID,protocol4);
         cout << "Rejected: " << sa1 << " / " << str << " -> " << sa2 << endl;
      }
      return;
   }

   // ====== Update report ==================================================
   for(cardinal i = 0;i < Count;i++) {
      if((Report[i]->Source      == source)      &&
         (Report[i]->Destination == destination) &&
         (Report[i]->FlowLabel   == flowLabel)   &&
         (Report[i]->StreamID    == streamID)    &&
         (Report[i]->Protocol3   == protocol3)   &&
         (Report[i]->Protocol4   == protocol4)) {
         Report[i]->BytesRaw[tosIndex]     += rawLength;
         Report[i]->BytesPayload[tosIndex] += payloadLength;
         Report[i]->Packets[tosIndex]++;
         return;
      }
   }
   if(Count < MaxCount) {
      Report[Count]->reset();
      Report[Count]->Source      = source;
      Report[Count]->Destination = destination;
      Report[Count]->FlowLabel   = flowLabel;
      Report[Count]->StreamID    = streamID;
      Report[Count]->Protocol3   = protocol3;
      Report[Count]->Protocol4   = protocol4;
      Report[Count]->BytesRaw[tosIndex]     += rawLength;
      Report[Count]->BytesPayload[tosIndex] += payloadLength;
      Report[Count]->Packets[tosIndex]++;
      Count++;

      if(PrintAccepted) {
         InternetAddress sa1(source);
         InternetAddress sa2(destination);
         char str[32];
         snprintf((char*)&str,sizeof(str),"$%05x/$%02x, sid=%d proto=$%02x",
                  flowLabel,trafficClass,streamID,protocol4);
         cout << "Accepted #" << Count << ": "
              << sa1 << " / " << str << " -> " << sa2 << endl;
      }

   }
   else {
      cerr << "WARNING: StreamMonitorReport::update() - MaxCount setting is too low!" << endl;
   }
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, StreamMonitorReport& report)
{
   if(!report.ready())
      return(os);

   // ====== Print total sums ===============================================
   os << "TotalBytesRaw     = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalBytesRaw[i] != 0) {
         os << report.TotalBytesRaw[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;
   os << "TotalBytesPayload = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalBytesPayload[i] != 0) {
         os << report.TotalBytesPayload[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;
   os << "TotalPackets      = ";
   for(cardinal i = 0;i < TrafficClassValues::MaxValues;i++) {
      if(report.TotalPackets[i] != 0) {
         os << report.TotalPackets[i] << "(" << TrafficClassValues::getNameForIndex(i) << ") ";
      }
   }
   os << endl;

   // ====== Print streams ==================================================
   for(cardinal i = 0;i < report.getCount();i++) {
      os << "Stream #" << i << " = " << *(report.getReport(i));
   }

   return(os);
}


}
