/*
 *  $Id: testreceiver.cc,v 1.4 2001/11/07 11:41:54 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Test Receiver
 *
 */


#include "tdsystem.h"
#include "thread.h"
#include "tdsocket.h"
#include "tdmessage.h"
#include "randomizer.h"
#include "breakdetector.h"
#include "ext_socket.h"
#include "testpacket.h"



using namespace Coral;


class ReceiverThread : public Thread
{
   // ====== Constructor/Destructor =========================================
   public:
   ReceiverThread(Socket*        socket,
                  const cardinal maxIncoming,
                  const bool     useSCTP);
   ReceiverThread::~ReceiverThread();

   // ====== Access methods =================================================
   card64 getBytesReceived(const cardinal stream);
   card64 getPacketsReceived(const cardinal stream);
   card64 getPacketsLost(const cardinal stream);
   double getFractionLost(const cardinal stream);
   double getJitter(const cardinal stream);


   // ====== Private data ===================================================
   private:
   void run();

   Socket*  ReceiverSocket;
   cardinal Incoming;
   card64   TotalReceived;
   struct ReceiverInfo {
      card64   Received;
      card64   PacketsReceived;
      card64   PacketsLost;
      cardinal IntervalPacketsReceived;
      cardinal IntervalPacketsLost;
      card64   Bandwidth;
      card64   PrevPacketTransmitTimeStamp;
      card64   PrevPacketArrivalTimeStamp;
      card64   PrevSeqNumber;
      double   Jitter;
   };
   ReceiverInfo* Received;
   bool          UseSCTP;
};


// ###### Constructor #######################################################
ReceiverThread::ReceiverThread(Socket*        socket,
                               const cardinal maxIncoming,
                               const bool     useSCTP)
   : Thread("ReceiverThread")
{
   UseSCTP        = useSCTP;
   ReceiverSocket = socket;
   TotalReceived  = 0;
   Incoming       = maxIncoming;
   Received = new ReceiverInfo[maxIncoming];
   if(Received == NULL) {
      cerr << "ERROR: ReceiverThread::ReceiverThread() - Out of memory!" << endl;
      ::exit(1);
   }
   for(cardinal i = 0;i < Incoming;i++) {
      Received[i].Received                    = 0;
      Received[i].PacketsReceived             = 0;
      Received[i].PacketsLost                 = 0;
      Received[i].IntervalPacketsReceived     = 0;
      Received[i].IntervalPacketsLost         = 0;
      Received[i].Bandwidth                   = 0;
      Received[i].Jitter                      = 0.0;
      Received[i].PrevPacketTransmitTimeStamp = 0;
      Received[i].PrevPacketArrivalTimeStamp  = 0;
      Received[i].PrevSeqNumber               = (card64)-1;
   }
   start();
}


// ###### Destructor ########################################################
ReceiverThread::~ReceiverThread()
{
   stop();
   delete Received;
   Received = NULL;
}


// ###### Get number of bytes received ######################################
card64 ReceiverThread::getBytesReceived(const cardinal stream)
{
   synchronized();
   card64 received;
   if(stream < Incoming) {
      received = Received[stream].Received;
   }
   else {
      received = TotalReceived;
   }
   unsynchronized();
   return(received);
}


// ###### Get jitter ########################################################
double ReceiverThread::getJitter(const cardinal stream)
{
   synchronized();
   double jitter;
   if(stream < Incoming) {
      jitter = Received[stream].Jitter;
   }
   else {
      jitter = -1.0;
   }
   unsynchronized();
   return(jitter);
}


// ###### Get packets ########################################################
card64 ReceiverThread::getPacketsReceived(const cardinal stream)
{
   synchronized();
   card64 packets;
   if(stream < Incoming) {
      packets = Received[stream].PacketsReceived;
   }
   else {
      packets = 0;
   }
   unsynchronized();
   return(packets);
}


// ###### Get packets ########################################################
card64 ReceiverThread::getPacketsLost(const cardinal stream)
{
   synchronized();
   card64 packets;
   if(stream < Incoming) {
      packets = Received[stream].PacketsLost;
   }
   else {
      packets = 0;
   }
   unsynchronized();
   return(packets);
}


// ###### Get loss rate #####################################################
double ReceiverThread::getFractionLost(const cardinal stream)
{
   synchronized();
   double lossRate;
   if(stream < Incoming) {
      const cardinal total = Received[stream].IntervalPacketsLost +
                                Received[stream].IntervalPacketsReceived;
      if(total != 0) {
         lossRate = (double)Received[stream].IntervalPacketsLost / (double)total;
      }
      else {
         lossRate = 0.0;
      }
      Received[stream].IntervalPacketsLost     = 0;
      Received[stream].IntervalPacketsReceived = 0;
   }
   else {
      lossRate = -1.0;
   }
   unsynchronized();
   return(lossRate);
}


// ###### timerEvent() implementation #######################################
void ReceiverThread::run()
{
   const cardinal packetSize = 65536;
   char           dataBuffer[packetSize];

   for(;;) {

      // ====== Receive string from association =============================
      card16  streamID = 0;
      ssize_t result;
      if(UseSCTP) {
         SocketMessage<CSpace(sizeof(sctp_sndrcvinfo))> message;
         message.setBuffer((char*)&dataBuffer,sizeof(dataBuffer));
         result = ReceiverSocket->receiveMsg(&message.Header,0);
         if(result > 0) {
            cmsghdr* cmsg = message.getFirstHeader();
            if(cmsg != NULL) {
               while(cmsg != NULL) {
                  if((cmsg->cmsg_level == IPPROTO_SCTP) &&
                     (cmsg->cmsg_type  == SCTP_SNDRCV)) {
                     sctp_sndrcvinfo* info = (sctp_sndrcvinfo*)CData(cmsg);
                     streamID = info->sinfo_stream;
                     if(streamID >= Incoming) {
                        streamID = 0;
                     }
                     break;
                  }
                  cmsg = message.getNextHeader(cmsg);
               }
            }
         }
      }
      else {
         result = ReceiverSocket->receive((void*)&dataBuffer,packetSize);
      }

      if(result >= 0) {
         const card64 now = getMicroTime();
         synchronized();
         TotalReceived               += result;
         Received[streamID].Received += result;
         Received[streamID].PacketsReceived++;
         Received[streamID].IntervalPacketsReceived++;
         if(result >= (ssize_t)sizeof(TestPacket)) {
            TestPacket* packet = (TestPacket*)&dataBuffer;
            packet->ID = translate32(packet->ID);
            if(packet->ID == packet->TestPacketID) {
               packet->SeqNumber = translate32(packet->SeqNumber);
               packet->Bandwidth = translate64(packet->Bandwidth);
               packet->TimeStamp = translate64(packet->TimeStamp);

               const double diff = fabs(
                  ((double)now - (double)packet->TimeStamp) -
                  ((double)Received[streamID].PrevPacketArrivalTimeStamp -
                   (double)Received[streamID].PrevPacketTransmitTimeStamp));

               Received[streamID].Jitter += (1.0/16.0) * (diff - Received[streamID].Jitter);

               double a = (double)packet->SeqNumber;
               double b = (double)Received[streamID].PrevSeqNumber;
               card64 d = (card64)rint(fabs(a - b));
               if((d != 1) && (d < ((card64)1 << 14))) {
                  Received[streamID].PacketsLost         += d;
                  Received[streamID].IntervalPacketsLost += d;
               }

               Received[streamID].PrevSeqNumber               = (card64)packet->SeqNumber;
               Received[streamID].PrevPacketArrivalTimeStamp  = now;
               Received[streamID].PrevPacketTransmitTimeStamp = packet->TimeStamp;
            }
         }
         unsynchronized();
      }
   }
}



// ###### Main program ######################################################
int main(int argc, char** argv)
{
   SocketAddress** localAddressArray = NULL;
   cardinal        localAddresses    = 0;
   cardinal        maxIncoming       = 128;
   bool            optForceIPv4      = false;
   bool            useSCTP           = true;
   bool            useDatagramMode   = true;
   bool            colorMode         = true;


   // ====== Get arguments ==================================================
   if(argc < 2) {
      cerr << "Usage: " << argv[0] << " "
              "[Port number] {-local=address1} ... {-local=addressN} {-sctp|-nosctp} {-dgram|-nodgram} {-nocolor} {-color} {-in=Max incoming streams} {-force-ipv4|-use-ipv6}"
           << endl;
      exit(1);
   }
   for(unsigned int i = 2;i < (cardinal)argc;i++) {
      if(!(strcasecmp(argv[i],"-force-ipv4")))    optForceIPv4 = true;
      else if(!(strcasecmp(argv[i],"-use-ipv6"))) optForceIPv4 = false;
      else if(!(strncasecmp(argv[i],"-in=",4))) {
         maxIncoming = atol(&argv[i][4]);
         if(maxIncoming < 1) {
            maxIncoming = 1;
         }
         else if(maxIncoming > 65535) {
            maxIncoming = 65535;
         }
      }
      else if(!(strncasecmp(argv[i],"-local=",7))) {
         if(localAddresses < SCTP_MAXADDRESSES) {
            if(localAddressArray == NULL) {
               localAddressArray = SocketAddress::newAddressList(SCTP_MAXADDRESSES);
               if(localAddressArray == NULL) {
                  cerr << "ERROR: Out of memory!" << endl;
                  exit(1);
               }
            }
            localAddressArray[localAddresses] = SocketAddress::createSocketAddress(
                                                   SocketAddress::PF_HidePort,
                                                   &argv[i][7]);
            if(localAddressArray[localAddresses] == NULL) {
               cerr << "ERROR: Argument #" << i << " is an invalid address!" << endl;
               exit(1);
            }
            localAddresses++;
         }
         else {
            cerr << "ERROR: Too many local addresses!" << endl;
            exit(1);
         }
      }
      else if(!(strcasecmp(argv[i],"-color"))) {
         colorMode = true;
      }
      else if(!(strcasecmp(argv[i],"-nocolor"))) {
         colorMode = false;
      }
      else if(!(strcmp(argv[i],"-sctp"))) {
         useSCTP = true;
      }
      else if(!(strcmp(argv[i],"-nosctp"))) {
         useSCTP = false;
      }
      else if(!(strcmp(argv[i],"-dgram"))) {
         useDatagramMode = true;
      }
      else if(!(strcmp(argv[i],"-nodgram"))) {
         useDatagramMode = false;
      }
      else {
         cerr << "ERROR: Bad parameter " << argv[i] << "!" << endl;
         exit(1);
      }
   }
   if(optForceIPv4) {
      InternetAddress::UseIPv6 = false;
   }
   if(localAddressArray == NULL) {
      if(!useSCTP) {
         localAddressArray = SocketAddress::newAddressList(1);
         if(localAddressArray == NULL) {
            cerr << "ERROR: Out of memory" << endl;
            exit(1);
         }
         localAddressArray[0] = new InternetAddress(atol(argv[1]));
         if(localAddressArray[0] == NULL) {
            cerr << "ERROR: Out of memory" << endl;
            exit(1);
         }
         localAddresses = 1;
      }
      else {
         if(!Socket::getLocalAddressList(
               localAddressArray,
               localAddresses,
               Socket::GLAF_HideLoopback|Socket::GLAF_HideLinkLocal|Socket::GLAF_HideBroadcast|Socket::GLAF_HideMulticast|Socket::GLAF_HideAnycast)) {
            cerr << "ERROR: Cannot obtain local addresses!" << endl;
            exit(1);
         }
         if(localAddresses < 1) {
            cerr << "ERROR: No valid local addresses have been found?!" << endl
                 << "       Check your network interface configuration!" << endl;
            exit(1);
         }
      }
   }


   // ====== Create socket and connect ======================================
   Socket receiverSocket(Socket::IP,
                         (useDatagramMode == true) ? SOCK_DGRAM : SOCK_STREAM,
                         (useSCTP == true) ? Socket::SCTP : 0);
   if(useSCTP) {
      sctp_initmsg init;
      init.sinit_num_ostreams   = 1;
      init.sinit_max_instreams  = maxIncoming;
      init.sinit_max_attempts   = 0;
      init.sinit_max_init_timeo = 60;
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_INITMSG,(char*)&init,sizeof(init)) < 0) {
         cerr << "ERROR: Unable to set SCTP_INITMSG parameters!" << endl;
         exit(1);
      }
   }
   for(cardinal i = 0;i < localAddresses;i++) {
      localAddressArray[i]->setPort(atol(argv[1]));
   }
   if(receiverSocket.bindx((const SocketAddress**)localAddressArray,
                           localAddresses,
                           SCTP_BINDX_ADD_ADDR) == false) {
      cerr << "ERROR: Unable to bind socket!" << endl;
      exit(1);
   }
   if(useSCTP) {
      int on = 1;
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVDATAIOEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVDATAIOEVNT failed!" << endl;
         exit(1);
      }
      on = 0;
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVASSOCEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVASSOCEVNT failed!" << endl;
         exit(1);
      }
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVPADDREVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVPADDREVNT failed!" << endl;
         exit(1);
      }
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVSENDFAILEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVSENDFAILEVNT failed!" << endl;
         exit(1);
      }
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVPEERERR,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVPEERERR failed!" << endl;
         exit(1);
      }
      if(receiverSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVSHUTDOWNEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVSHUTDOWNEVNT failed!" << endl;
         exit(1);
      }
   }


   // ====== Print information ==============================================
   cout << "Test Receiver - Copyright (C) 1999-2001 Thomas Dreibholz" << endl;
   cout << "--------------------------------------------------------" << endl;
   cout << "Version:           " << __DATE__ << ", " << __TIME__ << endl;
   cout << "SCTP:              " << ((useSCTP == true) ? "on" : "off") << endl;
   cout << "Datagram Mode:     " << ((useDatagramMode == true) ? "on" : "off") << endl;
   cout << "Local Addresses:   " << *(localAddressArray[0]) << endl;
   for(cardinal i = 1;i < localAddresses;i++) {
      cout << "                   " << *(localAddressArray[i]) << endl;
   }
   cout << "Max Incoming:      " << maxIncoming << endl;
   cout << endl << endl;


   // ====== Start receiver thread ==========================================
   Socket* socket;
   if(useDatagramMode) {
      socket = &receiverSocket;
   }
   else {
      cout << "Waiting for incoming connection... ";
      cout.flush();
      receiverSocket.listen();
      socket = receiverSocket.accept();
      if(socket == NULL) {
         cerr << endl << "ERROR: Connection establishment failed!" << endl;
         exit(1);
      }
      cout << "done." << endl << endl;
   }
   ReceiverThread receiverThread(socket,maxIncoming,useSCTP);


   // ====== Main loop ======================================================
   installBreakDetector();

   receiverThread.synchronized();
   double jitter[maxIncoming];
   double lossRate[maxIncoming];
   card64 packets[maxIncoming];
   card64 packetsLost[maxIncoming];
   card64 totalReceived;
   card64 received[maxIncoming];
   card64 oldReceived[maxIncoming];
   card64 oldTotalReceived = receiverThread.getBytesReceived((cardinal)-1);
   for(cardinal i = 0;i < maxIncoming;i++) {
      oldReceived[i] = receiverThread.getBytesReceived(i);
   }
   card64 oldTimeStamp = getMicroTime();
   receiverThread.unsynchronized();

   while(!breakDetected()) {
      Thread::delay(2500000);

      // ====== Get current values ==========================================
      receiverThread.synchronized();
      totalReceived = receiverThread.getBytesReceived((cardinal)-1);
      for(cardinal i = 0;i < maxIncoming;i++) {
         received[i]    = receiverThread.getBytesReceived(i);
         jitter[i]      = receiverThread.getJitter(i);
         lossRate[i]    = receiverThread.getFractionLost(i);
         packets[i]     = receiverThread.getPacketsReceived(i);
         packetsLost[i] = receiverThread.getPacketsLost(i);
      }
      const card64 now = getMicroTime();
      receiverThread.unsynchronized();


      // ====== Print status ================================================
      char str[512];
      snprintf((char*)&str,sizeof(str),
              "   Total      [Recv: %12Ld Bytes = %12Ld Bits]  [Rate: %9Ld Bytes/s = %9Ld Bits/s]",
              totalReceived, totalReceived * 8,
              (totalReceived - oldTotalReceived) * 1000000 / (now - oldTimeStamp),
              (totalReceived - oldTotalReceived) * 8 * 1000000 / (now - oldTimeStamp));
      if(colorMode) {
         cout << "\x1b[31m";
      }
      printTimeStamp(cout);
      if(colorMode) {
         cout << "\x1b[33m";
      }
      cout << endl << str << endl;

      bool color = false;
      for(cardinal i = 0;i < maxIncoming;i++) {
         if(received[i] > 0) {
            if(colorMode) {
               color = !color;
               if(color) {
                  cout << "\x1b[34m";
               }
               else {
                  cout << "\x1b[36m";
               }
            }
            snprintf((char*)&str,sizeof(str),
                    "   Stream #%02d [Recv: %12Ld Bytes = %12Ld Bits]  [Rate: %9Ld Bytes/s = %9Ld Bits/s]",
                    i + 1,
                    received[i], received[i] * 8,
                    (received[i] - oldReceived[i]) * 1000000 / (now - oldTimeStamp),
                    (received[i] - oldReceived[i]) * 8 * 1000000 / (now - oldTimeStamp));
            cout << str << endl;
            snprintf((char*)&str,sizeof(str),
                    "              [%10Ld Packets, Lost: %8Ld (%1.2f %%)]  [Jitter = %1.1f s]",
                    packets[i], packetsLost[i],
                    lossRate[i] * 100.0, jitter[i]);
            cout << str << endl;
         }
      }
      if(colorMode) {
         cout << "\x1b[30m";
         cout.flush();
      }


      // ====== Backup values ===============================================
      for(cardinal i = 0;i < maxIncoming;i++) {
         oldReceived[i] = received[i];
      }
      oldTotalReceived = totalReceived;
      oldTimeStamp     = now;
   }


   // ====== Stop thread ====================================================
   receiverThread.stop();
   if(!useDatagramMode) {
      delete socket;
   }
   SocketAddress::deleteAddressList(localAddressArray);
   cout << "Terminated!" << endl;
   return 0;
}
