/*
 *  $Id: testsender.cc,v 1.6 2001/11/20 12:24:41 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Test Sender
 *
 */


#include "tdsystem.h"
#include "tdsocket.h"
#include "tdmessage.h"
#include "internetaddress.h"
#include "timedthread.h"
#include "breakdetector.h"
#include "trafficclassvalues.h"
#include "randomizer.h"
#include "testpacket.h"



using namespace Coral;



const cardinal MaxStreams = 128;

struct StreamInfo
{
   card64 Bandwidth;
   card32 SeqNumber;
   card8  TrafficClass;
};
StreamInfo Streams[MaxStreams];



class SenderInterface
{
   public:
   virtual ~SenderInterface();
   virtual card64 getBytesSent(const cardinal stream) = 0;
};


// ###### Destructor ########################################################
SenderInterface::~SenderInterface()
{
}




class CBRSenderThread : public TimedThread,
                        public SenderInterface
{
   // ====== Constructor ====================================================
   public:
   CBRSenderThread(Socket*        socket,
                   StreamInfo*    streamArray,
                   const cardinal streamCount,
                   const cardinal framesPerSecond,
                   const cardinal packetSize,
                   const bool     useSCTP);

   card64 getBytesSent(const cardinal stream);


   // ====== Private data ===================================================
   private:
   void timerEvent();


   Socket*     SenderSocket;
   StreamInfo* StreamArray;
   cardinal    StreamCount;
   cardinal    FramesPerSecond;
   cardinal    PacketSize;
   card64      TotalSent;
   card64      Sent[MaxStreams];
   bool        UseSCTP;
};


// ###### Constructor #######################################################
CBRSenderThread::CBRSenderThread(Socket*        socket,
                                 StreamInfo*    streamArray,
                                 const cardinal streamCount,
                                 const cardinal framesPerSecond,
                                 const cardinal packetSize,
                                 const bool     useSCTP)
   : TimedThread(1000000)
{
   setTimerCorrection(0);
   StreamCount      = streamCount;
   StreamArray      = streamArray;
   PacketSize       = packetSize;
   FramesPerSecond  = framesPerSecond;
   SenderSocket     = socket;
   UseSCTP          = useSCTP;
   TotalSent        = 0;
   for(cardinal i = 0;i < StreamCount;i++) {
      Sent[i] = 0x00;
   }
   setInterval(1000000 / FramesPerSecond);
   start();
}


// ###### Get number of bytes sent ##########################################
card64 CBRSenderThread::getBytesSent(const cardinal stream)
{
   synchronized();
   card64 sent;
   if(stream < StreamCount) {
      sent = Sent[stream];
   }
   else {
      sent = TotalSent;
   }
   unsynchronized();
   return(sent);
}


// ###### timerEvent() implementation #######################################
void CBRSenderThread::timerEvent()
{
   char data[PacketSize];
   synchronized();
   for(cardinal i = 0;i < StreamCount;i++) {
      // ====== Send width bandwidth limit ==================================
      if(StreamArray[i].Bandwidth != (card64)-1) {
         card64 total = (card64)rint((double)StreamArray[i].Bandwidth /
                                        (double)FramesPerSecond);
         do {
            // ====== Calculate packet size =================================
            int64 size = (int64)total;
            if(size > (int64)PacketSize) {
               size = (int64)PacketSize;
            }
            if(size < sizeof(TestPacket)) {
               size = sizeof(TestPacket);
            }


            // ====== Create packet =========================================
            TestPacket* packet = (TestPacket*)&data;
            packet->ID        = translate32(TestPacket::TestPacketID);
            packet->SeqNumber = translate32(StreamArray[i].SeqNumber);
            packet->Bandwidth = translate32(StreamArray[i].Bandwidth);
            packet->TimeStamp = translate64(getMicroTime());


            // ====== Send a packet =========================================
            ssize_t result;
            if(UseSCTP) {
               SocketMessage<CSpace(sizeof(sctp_sndrcvinfo))> message;
               message.setBuffer((char*)&data,(size_t)size);
               sctp_sndrcvinfo* info = (sctp_sndrcvinfo*)message.addHeader(sizeof(sctp_sndrcvinfo),IPPROTO_SCTP,SCTP_SNDRCV);
               info->sinfo_assoc_id = 0;
               info->sinfo_stream   = (unsigned short)i;
               info->sinfo_flags    = 0;
               info->sinfo_ppid     = (unsigned int)getpid();
               result = SenderSocket->sendMsg(&message.Header,0,StreamArray[i].TrafficClass);
            }
            else {
               result = SenderSocket->send((char*)&data,size,0,StreamArray[i].TrafficClass);
            }


            // ====== Update size to send and sequence number ===============
            StreamArray[i].SeqNumber++;
            if(result == size) {
               Sent[i]   += (card64)size;
               TotalSent += (card64)size;
            }

            if(total >= (card64)size) {
               total -= (card64)size;
            }
            else {
               total = 0;
            }
         } while(total > 0);
      }

      // ====== Send as much as possible ====================================
      else {
         ssize_t result;
         do {
            // ====== Create packet =========================================
            TestPacket* packet = (TestPacket*)&data;
            packet->ID        = translate32(TestPacket::TestPacketID);
            packet->SeqNumber = translate32(StreamArray[i].SeqNumber);
            packet->Bandwidth = translate32(StreamArray[i].Bandwidth);
            packet->TimeStamp = translate64(getMicroTime());

            // ====== Send packet ===========================================
            if(UseSCTP) {
               SocketMessage<CSpace(sizeof(sctp_sndrcvinfo))> message;
               message.setBuffer((char*)&data,(size_t)PacketSize);
               sctp_sndrcvinfo* info = (sctp_sndrcvinfo*)message.addHeader(sizeof(sctp_sndrcvinfo),IPPROTO_SCTP,SCTP_SNDRCV);
               info->sinfo_assoc_id = 0;
               info->sinfo_stream   = (unsigned short)i;
               info->sinfo_flags    = 0;
               info->sinfo_ppid     = (unsigned int)getpid();
               result = SenderSocket->sendMsg(&message.Header,0,StreamArray[i].TrafficClass);
            }
            else {
               result = SenderSocket->send((char*)&data,PacketSize,0,StreamArray[i].TrafficClass);
            }

            // ====== Update counter ========================================
            if(result == (ssize_t)PacketSize) {
               Sent[i]   += (card64)PacketSize;
               TotalSent += (card64)PacketSize;
               StreamArray[i].SeqNumber++;
            }
         } while(result >= 0);
      }
   }
   unsynchronized();
}


// ###### Main program ######################################################
int main(int argc, char** argv)
{
   Randomizer      random;
   bool            optForceIPv4      = false;
   bool            useSCTP           = true;
   bool            useDatagramMode   = true;
   bool            useBits           = false;
   bool            colorMode         = true;
   cardinal        streams           = 0;
   cardinal        framesPerSecond   = 25;
   cardinal        packetSize        = 1400;
   cardinal        localAddresses    = 0;
   SocketAddress** localAddressArray = SocketAddress::newAddressList(SCTP_MAXADDRESSES);
   if(localAddressArray == NULL) {
      cerr << "ERROR: Out of memory!" << endl;
      exit(1);
   }


   // ====== Get arguments ==================================================
   if(argc < 2) {
      cerr << "Usage: " << argv[0] << " "
           << "[Remote address] [Bandwidth1{,Traffic Class1}] ... [BandwidthN{,Traffic ClassN}] {-frames=Frames/sec.} {-pktsize=Packet size} {-local=address1} ... {-local=addressN} {-sctp|-nosctp} {-dgram|-nodgram} {-nocolor} {-color} {-force-ipv4|-use-ipv6}"
           << endl;
      exit(1);
   }
   for(unsigned int i = 2;i < (cardinal)argc;i++) {
      card64 bandwidth;
      char   tcString[32];
      const integer args =
         sscanf(argv[i],"%Ld,%8[0-z]",&bandwidth,(char*)&tcString);
      if(args > 0) {
         if(streams >= MaxStreams - 1) {
            cerr << "ERROR: Too many streams!" << endl;
            exit(1);
         }
         if(useBits) {
            bandwidth /= 8;
         }
         if(bandwidth < 16) {
            bandwidth = 16;
         }
         Streams[streams].Bandwidth = bandwidth;
         Streams[streams].SeqNumber = random.random32();
         if(args > 1) {
            card16 tc;
            tc = TrafficClassValues::getTrafficClassForName((char*)&tcString);
            if(tc == 0xffff) {
               Streams[streams].TrafficClass = 0;
            }
            else {
               Streams[streams].TrafficClass = (card8)tc;
            }
         }
         else {
            Streams[streams].TrafficClass = 0;
         }
         streams++;
      }
      else if(!(strncasecmp(argv[i],"-frames=",8))) {
         framesPerSecond = atol(&argv[i][8]);
      }
      else if(!(strncasecmp(argv[i],"-pktsize=",9))) {
         packetSize = atol(&argv[i][9]);
      }
      else if(!(strcmp(argv[i],"-sctp"))) {
         useSCTP = true;
      }
      else if(!(strcmp(argv[i],"-nosctp"))) {
         useSCTP = false;
      }
      else if(!(strcmp(argv[i],"-dgram"))) {
         useDatagramMode = true;
      }
      else if(!(strcmp(argv[i],"-nodgram"))) {
         useDatagramMode = false;
      }
      else if(!(strcmp(argv[i],"-color"))) {
         colorMode = true;
      }
      else if(!(strcmp(argv[i],"-nocolor"))) {
         colorMode = false;
      }
      else if(!(strcasecmp(argv[i],"-force-ipv4"))) optForceIPv4 = true;
      else if(!(strcasecmp(argv[i],"-use-ipv6")))   optForceIPv4 = false;
      else if(!(strncasecmp(argv[i],"-local=",7))) {
         if(localAddresses < SCTP_MAXADDRESSES) {
            localAddressArray[localAddresses] =
               SocketAddress::createSocketAddress(SocketAddress::PF_HidePort,
                                                  &argv[i][7]);
            if(localAddressArray[localAddresses] == NULL) {
               cerr << "ERROR: Argument #" << i << " is an invalid address!" << endl;
               exit(1);
            }
            localAddresses++;
         }
         else {
            cerr << "ERROR: Too many local addresses!" << endl;
            exit(1);
         }
      }
      else {
         cerr << "ERROR: Bad parameter " << argv[i] << "!" << endl;
         exit(1);
      }
   }
   if(streams < 1) {
      Streams[0].Bandwidth    = 100000;
      Streams[0].SeqNumber    = random.random32();
      Streams[0].TrafficClass = 0x00;
      streams = 1;
   }
   if(optForceIPv4) {
      InternetAddress::UseIPv6 = false;
   }
   SocketAddress* remoteAddress = SocketAddress::createSocketAddress(0,argv[1]);
   if(remoteAddress == NULL) {
      cerr << "ERROR: Bad remote address! Use <address>:<port> format." << endl;
      exit(1);
   }
   if(localAddresses < 1) {
      localAddressArray[0] = SocketAddress::getLocalAddress(*remoteAddress);
      if(localAddressArray[0] == NULL) {
         cerr << "ERROR: Out of memory!" << endl;
         exit(1);
      }
      localAddresses = 1;
   }


   // ====== Print information ==============================================
   cout << "Test Sender - Copyright (C) 1999-2001 Thomas Dreibholz" << endl;
   cout << "------------------------------------------------------" << endl;
   cout << "Version:               " << __DATE__ << ", " << __TIME__ << endl;
   cout << "SCTP:                  " << ((useSCTP == true) ? "on" : "off") << endl;
   cout << "Datagram Mode:         " << ((useDatagramMode == true) ? "on" : "off") << endl;
   cout << "Local Addresses:       " << *(localAddressArray[0]) << endl;
   for(cardinal i = 1;i < localAddresses;i++) {
      cout << "                       " << *(localAddressArray[i]) << endl;
   }
   cout << "Remote Address:        " << *remoteAddress  << endl;
   cout << "Frames Per Second      " << framesPerSecond << endl;
   cout << "Outgoing Streams:" << endl;
   card64 bufferSize = 0;
   for(cardinal i = 0;i < streams;i++) {
      char str[256];
      if(Streams[i].Bandwidth != (card64)-1) {
         snprintf((char*)&str,sizeof(str),"Stream #%02d:  %12Ld Bytes/s = %12Ld Bits/s, %s",
                  i + 1,
                  Streams[i].Bandwidth, Streams[i].Bandwidth * 8,
                  TrafficClassValues::getNameForTrafficClass(Streams[i].TrafficClass));
         bufferSize += Streams[i].Bandwidth;
      }
      else {
         snprintf((char*)&str,sizeof(str),"Stream #%02d:  as much as possible, %s",
                  i + 1,
                  TrafficClassValues::getNameForTrafficClass(Streams[i].TrafficClass));
         bufferSize += 100000;
      }
      cout << "   " << str << endl;
   }
   cout << endl << endl;


   // ====== Create socket and connect ======================================
   Socket senderSocket(Socket::IP,
                       (useDatagramMode == true) ? SOCK_DGRAM : SOCK_STREAM,
                       (useSCTP == true) ? Socket::SCTP : 0);
   if(useSCTP) {
      sctp_initmsg init;
      init.sinit_num_ostreams   = streams;
      init.sinit_max_instreams  = 1;
      init.sinit_max_attempts   = 0;
      init.sinit_max_init_timeo = 60;
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_INITMSG,(char*)&init,sizeof(init)) < 0) {
         cerr << "ERROR: Unable to set SCTP_INITMSG parameters!" << endl;
         exit(1);
      }
   }
   if(senderSocket.bindx((const SocketAddress**)localAddressArray,
                         localAddresses,
                         SCTP_BINDX_ADD_ADDR) == false) {
      cerr << "ERROR: Unable to bind socket!" << endl;
      exit(1);
   }
   if(useSCTP) {
      int on = 0;
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVDATAIOEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVDATAIOEVNT failed!" << endl;
         exit(1);
      }
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVASSOCEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVASSOCEVNT failed!" << endl;
         exit(1);
      }
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVPADDREVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVPADDREVNT failed!" << endl;
         exit(1);
      }
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVSENDFAILEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVSENDFAILEVNT failed!" << endl;
         exit(1);
      }
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVPEERERR,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVPEERERR failed!" << endl;
         exit(1);
      }
      if(senderSocket.setSocketOption(IPPROTO_SCTP,SCTP_RECVSHUTDOWNEVNT,&on,sizeof(on)) < 0) {
         cerr << "ERROR: SCTP_RECVSHUTDOWNEVNT failed!" << endl;
         exit(1);
      }
   }
   else {
      senderSocket.setSoBroadcast(true);
      unsigned int size = (unsigned int)((2 * bufferSize) / framesPerSecond);
      if(senderSocket.setSocketOption(SOL_SOCKET,SO_SNDBUF,(void*)&size,sizeof(size)) < 0) {
         cerr << "WARNING: Unable to enlarge send buffer!" << endl;
         cerr << "err=" << senderSocket.getLastError() << endl;
      }
   }
   cout << "Connecting... ";
   cout.flush();
   if(senderSocket.connect(*remoteAddress) == false) {
      cout << "failed!" << endl;
      cerr << "ERROR: Unable to connect to remote address!" << endl;
      exit(1);
   }
   cout << "done." << endl << endl;


   // ====== Start threads ==================================================
   installBreakDetector();
   senderSocket.setBlockingMode(false);
   CBRSenderThread sender(&senderSocket,
                          (StreamInfo*)&Streams, streams,
                          framesPerSecond, packetSize,
                          useSCTP);


   // ====== Main loop ======================================================
   card64 oldTotalSent = 0;
   card64 oldSent[streams];
   for(cardinal i = 0;i < streams;i++) {
      oldSent[i] = 0;
   }
   card64 oldTimeStamp = getMicroTime();
   while(!breakDetected()) {
      Thread::delay(2500000,true);

      // ====== Get current values ==========================================
      sender.synchronized();
      const card64 totalSent = sender.getBytesSent((cardinal)-1);
      card64 sent[streams];
      for(cardinal i = 0;i < streams;i++) {
         sent[i] = sender.getBytesSent(i);
      }
      const card64 now = getMicroTime();
      sender.unsynchronized();


      // ====== Print status ================================================
      char str[512];
      snprintf((char*)&str,sizeof(str),
              "   Total      [Sent: %12Ld Bytes = %12Ld Bits]  [Rate: %9Ld Bytes/s = %9Ld Bits/s]",
              totalSent, totalSent * 8,
              (totalSent - oldTotalSent) * 1000000 / (now - oldTimeStamp),
              (totalSent - oldTotalSent) * 8 * 1000000 / (now - oldTimeStamp));
      if(colorMode) {
         cout << "\x1b[31m";
      }
      printTimeStamp(cout);
      if(colorMode) {
         cout << "\x1b[33m";
      }
      cout << endl << str << endl;

      if(colorMode) {
         cout << "\x1b[34m";
      }
      for(cardinal i = 0;i < streams;i++) {
         snprintf((char*)&str,sizeof(str),
                 "   Stream #%02d [Sent: %12Ld Bytes = %12Ld Bits]  [Rate: %9Ld Bytes/s = %9Ld Bits/s %s]",
                 i + 1,
                 sent[i], sent[i] * 8,
                 (sent[i] - oldSent[i]) * 1000000 / (now - oldTimeStamp),
                 (sent[i] - oldSent[i]) * 8 * 1000000 / (now - oldTimeStamp),
                 TrafficClassValues::getNameForTrafficClass(Streams[i].TrafficClass));
         cout << str << endl;
      }
      if(colorMode) {
         cout << "\x1b[30m";
         cout.flush();
      }


      // ====== Backup Sents ===============================================
      oldTotalSent = totalSent;
      for(cardinal i = 0;i < streams;i++) {
         oldSent[i] = sent[i];
      }
      oldTimeStamp = now;
   }


   // ====== Stop thread ====================================================
   sender.stop();
   SocketAddress::deleteAddressList(localAddressArray);
   cout << "Terminated!" << endl;
   return 0;
}
