/*
 *  $Id: traffic_tool1.c,v 1.13 2001/11/29 17:59:39 ajung Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * traffic_tool1.c  - program for generating a traffic profile
 *
 */

#include "sctp_params.h"



static struct ulp_data ulpData[MAXIMUM_NUMBER_OF_ASSOCIATIONS];
static unsigned char localAddressList[MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES][SCTP_MAX_IP_LEN];
static unsigned char destAddress[SCTP_MAX_IP_LEN];

static unsigned short noOfLocalAddresses      = 0;
static unsigned char  tosByte                 = 0x10;  /* IPTOS_LOWDELAY */
static unsigned int doMeasurements            = 0;
static unsigned int doAllMeasurements         = 0;
static int unknownCommand                     = 0;
static unsigned int deltaT                    = 1000;
static int sender                             = 0;
static int receiver                           = 1;
static int sendUnordered                      = 0;
static int assoc1Active                       = 0;
static int path1Active                        = 0;
static int path2Active                        = 0;
static unsigned int tid                       =0;
static int useDumpFile                        = 0;



int distribution(double lambda, double min, double max);


void printUsage(void)
{
    printf("usage:   traffic_tool1 [options] -s source_addr_1 -s source_addr_2 -d dest_addr\n");
    printf("         SENDER starts exp traffic sources to destinations [dest_addr1/2], after CommUp.\n");
    printf("usage:   traffic_tool1 [options] -s source_addr_1 -s source_addr_2  --  RECEIVER \n");
    printf("options:\n");
    printf("-l length           number of bytes of the payload when generating traffic (default 100)\n");
    printf("-i param1           parameter 1 of chosen distribution\n");
    printf("-j param2           (optional) parameter 2 (min) of chosen distribution\n");
    printf("-k param3           (optional) parameter 3 (max) of chosen distribution\n");
    printf("-m                  print number of received bytes and chunks per period (see -p)\n");
    printf("-n chunks           number of chunks sent for each timer event \n");
    printf("-M                  print number of received bytes, chunks per period (see -p) and flow control info\n");
    printf("-p period           period for the measurements in milliseconds (default 1000)\n");
    printf("-t byte             TOS byte used by all assiciations (default 0x10)\n");
    printf("-u                  inject the packets unordered\n");
    printf("-v                  verbose mode\n");   
    printf("-V                  very verbose mode\n");   
}

void getArgs(int argc, char **argv)
{
    int c;
    extern char *optarg;
    extern int optind;

    fptr = stdout;

    while ((c = getopt(argc, argv, "s:d:n:i:j:k:l:p:t:mMuvVf")) != -1)
    {
        switch (c) {
        case 'd':
            if (strlen(optarg) < SCTP_MAX_IP_LEN) {
                strcpy(destAddress, optarg);
                sender = 1;
                receiver = 0;
            }
            break;
        case 'l':
            chunkLength = min(atoi(optarg),MAXIMUM_PAYLOAD_LENGTH);
          break;
       case 'f':
            useDumpFile = 1;
            fptr = fopen("./l.out","w+");
            break;
        case 'm':
            doMeasurements = 1;
            break;
        case 'n':
            numOfChunks = atoi(optarg);
            break;
        case 'M':
            doMeasurements = 1;
            doAllMeasurements = 1;
            break;
        case 'i':
            param1 = atof(optarg);
            break;
        case 'j':
            param2 = atof(optarg);
            break;
        case 'k':
            param3 = atof(optarg);
            break;
        case 'p':
            deltaT = atoi(optarg);
            break;
        case 's':
            if ((noOfLocalAddresses < MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES) &&
                (strlen(optarg) < SCTP_MAX_IP_LEN  )) {
                strcpy(localAddressList[noOfLocalAddresses], optarg);
                noOfLocalAddresses++;
            }
            break;  
            break;
        case 't':
            tosByte = (unsigned char) atoi(optarg);
            break;
        case 'u':
            sendUnordered = 1;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'V':
            verbose = 1;
            vverbose = 1;
            break;
        default:
            unknownCommand = 1;
            break;
        }
    }
}
void checkArgs(void)
{
    int abortProgram;
    int printUsageInfo;
    
    abortProgram = 0;
    printUsageInfo = 0;

    if ((noOfLocalAddresses != 2) ) {
        printf("Error:   Two sourceaddress must be specified.\n");
        abortProgram = 1;
        printUsageInfo = 1;
    }
    if (unknownCommand ==1) {
         printf("Error:   Unkown options in command.\n");
         printUsageInfo = 1;
    }

    if ((param2 != 0.0 && param3 == 0.0)||(param2==0.0 && param3 !=0.0)) {
        printf("Error:  Set BOTH min AND max Parameters !\n");
        abortProgram = 1;
    }

    if (printUsageInfo == 1)
        printUsage();
    if (abortProgram == 1)
        exit(-1);
}

#define dtotimeval(d, tvp) \
        { \
        (tvp)->tv_sec = floor(d); \
        (tvp)->tv_usec = ((d) - (tvp)->tv_sec) * 1000000; \
        }

/* Convert the double d to timer tvp, offsetting from current time.     */

#define dtotimevalfromnow(d, tvp) \
        { \
        unsigned long   seconds; \
        \
        if (gettimeofday(tvp, (struct timezone *)NULL) == -1) \
                { \
                (void)perror("gettimeofday"); \
                exit(-1); \
                }\
        seconds = floor(d); \
        (tvp)->tv_sec += seconds; \
        (tvp)->tv_usec += ((d) - seconds) * 1000000; \
        if ((tvp)->tv_usec >= 1000000) \
                { \
                (tvp)->tv_usec -= 1000000; \
                (tvp)->tv_sec++; \
                } \
        }

/* Convert the double d to timer tvp, offsetting from specified time.   */

#define dtotimevalfromthen(then, d, tvp) \
        { \
        unsigned long   seconds; \
        \
        seconds = floor(d); \
        (tvp)->tv_sec = (then)->tv_sec + seconds; \
        (tvp)->tv_usec = (then)->tv_usec + ((d) - seconds) * 1000000; \
        if ((tvp)->tv_usec >= 1000000) \
                { \
                (tvp)->tv_usec -= 1000000; \
                (tvp)->tv_sec++; \
                } \
        }



/* ------------------------------------------------------------------------------------------ */
int distribution(double lambda, double min, double max)
{
    double value, l;
    int result;
    unsigned int msecs, secs, usecs;
    unsigned long int maxrand =  MAX_RANDOM;
    maxrand++;
    l = lambda - min;
    /* in general, a function of param1 and param2 */
    /* we send chunkLength bytes each time, and want param1 bytes per second */
    /* so we need */
    /* this is to be CBR-like */
    do {
        value = - l * log( ((double)(lrand48() + 1))/
                                    (double)(maxrand));
    } while ((value < min) || (value > max));

    secs =  floor(value);
    usecs = (value - secs) * 1000000;
    msecs = secs * 1000 + usecs / 1000;

    //   value = ((double)(((double)chunkLength)/((double)param1)))*((double)1000);
    result = (int)msecs;
//    if (result < 20) printf("Distribution result is small: (len: %u, lambda: %f) is: %d msecs \n", chunkLength, param1, result);
//    printf("Distro (len: %u, lambda: %f, min: %f, max: %f) is: %d msecs \n", chunkLength, lambda, min, max,result);
    return result;
}

void
shutdownFunction(unsigned int timerID, void *parameter1, void *parameter2)
{
    printf("--------> Shutting down interface eth0...... <---------------------------\n");
    system("/sbin/iptables -A OUTPUT -d 132.252.151.222 -j DROP --verbose");
    system("/sbin/iptables -A INPUT  -s 132.252.151.222 -j DROP --verbose");
    printf("--------> ...................... DONE !!!!   <---------------------------\n");
}



void
trafficGenerationTimer(unsigned int timerID, void *parameter1, void *parameter2)
{
    unsigned char chunk[MAXIMUM_PAYLOAD_LENGTH];
    int result, counter;
    struct c_data* cd = (struct c_data*) chunk;

    int delta = distribution(0.1, 0.02, 100.0);

    /* chose assoc1, path1, else assoc1 path2, else assoc2 (path21) else report LINK FAILURE ?! */

    struct ulp_data* ud = &ulpData[0];

    /* generate a packet, send it */
    memset(chunk, 0xE1, sizeof(chunk));


    for (counter=0; counter < numOfChunks; counter++) {
        /* insert send time into chunk */
        gettimeofday(&(cd->send_time), NULL);
        cd->send_number = ud->nrOfSentChunksTotal;

        result = SCTP_send(ud->assocID,
                       0, /* stream id */
                      chunk, chunkLength,
                      SCTP_GENERIC_PAYLOAD_PROTOCOL_ID,
                      SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, SCTP_INFINITE_LIFETIME,
                      (sendUnordered)?SCTP_UNORDERED_DELIVERY:SCTP_ORDERED_DELIVERY,
                      SCTP_BUNDLING_DISABLED);
        if (vverbose) {
            fprintf(fptr, "%-8x: Data sent (%u bytes on stream %u, %s) Result: %d\n",
                        ud->assocID, chunkLength, 0,
                        (sendUnordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered", result);
            fflush(fptr);
        }
        ud->nrOfSentChunks++;
        ud->nrOfSentChunksTotal++;
        ud->nrOfSentBytes += chunkLength;
        ud->nrOfSentBytesPerStream[0] += chunkLength;
        ud->nrOfSentChunksPerStream[0]++;
    }

    /* start new timer */
    tid = SCTP_startTimer(delta, &trafficGenerationTimer, parameter1, parameter2);
}

/* ------------------------------------------------------------------------------------------ */




void sendFailureNotif(unsigned int assocID,
                      unsigned char *unsent_data, unsigned int dataLength, unsigned int *context, void* dummy)
{
  if (verbose) {  
    fprintf(fptr, "%-8x: Send failure\n", assocID);
    fflush(fptr);
  }
}

void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, unsigned short newState, void* ulpDataPtr)
{

    SCTP_AssociationStatus assocStatus;
    SCTP_PathStatus pathStatus;
    unsigned short pathID;

    if (verbose) {
        SCTP_getPathStatus(assocID, affectedPathID, &pathStatus);
        fprintf(fptr, "%-8x: Network status change: path %u (towards %s) is now %s\n",
                        assocID, affectedPathID,
                        pathStatus.destinationAddress,
                        ((newState == SCTP_PATH_OK) ? "ACTIVE" : "INACTIVE"));
        fflush(fptr);
    }

    if (affectedPathID == 0) {
       if (newState == SCTP_PATH_OK) {
           path1Active = 1;
        } else {
           path1Active = 0;
        }
    } else if (affectedPathID == 1) {
        if (newState == SCTP_PATH_OK) {
            path2Active = 1;
        } else {
            path2Active = 0;
        }
    } else {
        fprintf(fptr, "%-8x: Error in Setup - more than 2 PATHs ????\n",assocID);
        fflush(fptr);
        exit (-1);
    }

    /* if the primary path has become inactive */
    if ((newState == SCTP_PATH_UNREACHABLE) && (affectedPathID == SCTP_getPrimary(assocID))) {
        /* select a new one */
        SCTP_getAssocStatus(assocID, &assocStatus);
        for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
            SCTP_getPathStatus(assocID, pathID, &pathStatus);
            if (pathStatus.state == SCTP_PATH_OK)
                break;
        }
        
        /* and use it */
        if (pathID < assocStatus.numberOfAddresses) {
            SCTP_setPrimary(assocID, pathID);
            if (verbose) {
                fprintf(fptr, "%-8x: Set new Primary : Path %u \n", assocID, pathID);
                fflush(fptr);
            }
        }
    }
}

void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,
                           unsigned short noOfInStreams, unsigned short noOfOutStreams, void* dummy)
{	
    unsigned int index, packetNumber;
    int delta = distribution(param1, param2, param3);
    SCTP_PathStatus pathStatus;
    unsigned short pathID;
     
    if (verbose) {  
        fprintf(fptr, "%-8x: Communication up (%u paths:", assocID, noOfPaths);
        for (pathID=0; pathID < noOfPaths; pathID++){
            SCTP_getPathStatus(assocID, pathID, &pathStatus);
            fprintf(fptr, " %s", pathStatus.destinationAddress);
        }
        fprintf(fptr, ")\n");
        fprintf(fptr, "%-8x:                  %u incoming, %u outgoing streams.\n",
                assocID, noOfInStreams, noOfOutStreams);
        if (sender)   fprintf(fptr, "I am a SENDER !\n");
        if (receiver)   fprintf(fptr, "I am a RECEIVER !\n");
        fflush(fptr);
    }
  
    /* look for a free ULP data */
    for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
        if (ulpData[index].maximumStreamID == -1)
            break;
    }
   
    /* if found */
    if (index == 0) {
        /* use it */
        ulpData[index].maximumStreamID = noOfOutStreams - 1;
        ulpData[index].assocID = assocID;
        if (index==0) assoc1Active = 1;

        for (pathID=0; pathID < noOfPaths; pathID++){
            SCTP_getPathStatus(assocID, pathID, &pathStatus);
            if (verbose) {
                fprintf(fptr, "%-8x: Path %u (status %s)\n", assocID, pathID,
                                (pathStatus.state==SCTP_PATH_OK)?"ACTIVE":"INACTIVE");
                fflush(fptr);
            }
            if (pathID==0) {
                if(pathStatus.state==SCTP_PATH_OK) path1Active = 1;
                else path1Active = 0;
            }
            if (pathID==1) {
                if(pathStatus.state==SCTP_PATH_OK) path2Active = 1;
                else path2Active = 0;
            }
        }

        if (sender) {
              tid = SCTP_startTimer(delta, &trafficGenerationTimer, &(ulpData[index]), NULL);
              gettimeofday(&shutdown_time, NULL);
              shutdown_time.tv_sec += 10;
              SCTP_startTimer(10000, &shutdownFunction, NULL, NULL);
        }
        return &ulpData[index];
    } else {
        if (verbose) {
           fprintf(fptr, "%-8x: CommUP : index != 0 --> aborting !\n", assocID);
           fflush(fptr);
        }
        /* abort assoc due to lack of resources */
        SCTP_abort(assocID);
        return NULL;
    }
}

void communicationLostNotif(unsigned int assocID, unsigned short status, void* ulpDataPtr)
{	
    unsigned char buffer[MAXIMUM_PAYLOAD_LENGTH];
    unsigned int bufferLength;
    unsigned short streamID, streamSN;
    unsigned int protoID;
    unsigned int tsn;
    int index;
    
    if (verbose) {
        fprintf(fptr, "%-8x: Communication lost (status %u)\n", assocID, status);
        fflush(fptr);
    }

    assoc1Active = 0;
    path1Active = 0;
    path2Active = 0;

    /* retrieve data */
    bufferLength = sizeof(buffer);
    while (SCTP_receiveUnsent(assocID, buffer, &bufferLength, &tsn,
                              &streamID, &streamSN, &protoID) > 0){
        /* do something with the retrieved data */
        /* after that, reset bufferLength */
        bufferLength = sizeof(buffer);
    }
    
    bufferLength = sizeof(buffer);
    while (SCTP_receiveUnacked(assocID, buffer, &bufferLength, &tsn,
                                &streamID, &streamSN, &protoID) > 0){
        /* do something with the retrieved data */
        /* after that, reset bufferLength */
        bufferLength = sizeof(buffer);
    }

    if (tid != 0 && assoc1Active==0) {
        sctp_stopTimer(tid);
    }
    /* free ULP data */
    ((struct ulp_data *) ulpDataPtr)->maximumStreamID = -1;
    
    /* delete the association */
    SCTP_deleteAssociation(assocID);
}

void communicationErrorNotif(unsigned int assocID, unsigned short status, void* dummy)
{
  if (verbose) {  
    fprintf(fptr, "%-8x: Communication error (status %u)\n", assocID, status);
    fflush(fptr);
  }
}

void restartNotif(unsigned int assocID, void* ulpDataPtr)
{
    SCTP_AssociationStatus assocStatus;
    
    if (verbose) {  
        fprintf(fptr, "%-8x: Restart\n", assocID);
        fflush(fptr);
    }
    
    SCTP_getAssocStatus(assocID, &assocStatus);
    
    /* update ULP data */
    ((struct ulp_data *) ulpDataPtr)->maximumStreamID = assocStatus.outStreams - 1;
    ((struct ulp_data *) ulpDataPtr)->assocID = assocID;

}

void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr)
{
    int index;

    if (verbose) {
        fprintf(fptr, "%-8x: Shutdown complete\n", assocID);
        fflush(fptr);
    }
    assoc1Active = 0;
    path1Active = 0;
    path2Active = 0;

    if (tid != 0 && assoc1Active==0) {
        sctp_stopTimer(tid);
    }

    /* free ULP data */
    ((struct ulp_data *) ulpDataPtr)->maximumStreamID = -1;
    SCTP_deleteAssociation(assocID);

}
/* ------------------ CHANGE -------------------------------------------------------------*/

void
measurementTimerRunOffFunction(unsigned int timerID, void *parameter1, void *parameter2)
{
    int index;
    SCTP_AssociationStatus assocStatus;
    SCTP_PathStatus pathStatus;
    unsigned short pathID;

    for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
        if (ulpData[index].maximumStreamID >= 0){
            if (doAllMeasurements) {
                SCTP_getAssocStatus(ulpData[index].assocID, &assocStatus);
                for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
                    SCTP_getPathStatus(ulpData[index].assocID, pathID, &pathStatus);
                    if (receiver) {
                        fprintf(fptr, "A%-4x P%-5u Ch:%-6lu By:%-6lu rw:%-6u RQ:%-6u cw:%-6u ssth:%-6u\n",
                            ulpData[index].assocID,
                            pathID,
                            ulpData[index].nrOfReceivedChunks,
                            ulpData[index].nrOfReceivedBytes,
                            assocStatus.currentReceiverWindowSize,
                            assocStatus.noOfChunksInReceptionQueue,
                            pathStatus.cwnd,
                            pathStatus.ssthresh);
                    } else {
                        fprintf(fptr, "A%-4x P%-5u Ch:%-6lu By:%-6lu rw:%-6u SQ:%-6u OBC:%-6u cw:%-6u ssth:%-6u\n",
                            ulpData[index].assocID,
                            pathID,
                            ulpData[index].nrOfSentChunks,
                            ulpData[index].nrOfSentBytes,
                            assocStatus.currentReceiverWindowSize,
                            assocStatus.noOfChunksInSendQueue,
                            assocStatus.outstandingBytes,
                            pathStatus.cwnd,
                            pathStatus.ssthresh);
                    }
                }
            } else {
                if (receiver) {
                    fprintf(fptr, "%-8x: %-6lu Chunks, %-8lu Bytes received\n",
                            ulpData[index].assocID, ulpData[index].nrOfReceivedChunks, ulpData[index].nrOfReceivedBytes);
                }else{
                    fprintf(fptr, "%-8x: %-6lu Chunks, %-8lu Bytes sent\n",
                            ulpData[index].assocID, ulpData[index].nrOfSentChunks, ulpData[index].nrOfSentBytes);
                }

            }
            ulpData[index].nrOfReceivedChunks = 0;
            ulpData[index].nrOfReceivedBytes  = 0;
            ulpData[index].nrOfSentChunks = 0;
            ulpData[index].nrOfSentBytes  = 0;
        }
    }
    SCTP_startTimer(deltaT, measurementTimerRunOffFunction, NULL, NULL);
}
/* ------------------ CHANG UP TO HERE --------------------------------------------------------*/

int main(int argc, char **argv)
{
    unsigned short sctpInst=0;
    SCTP_ulpCallbacks trafficUlp;
    SCTP_InstanceParameters instanceParameters;
    unsigned int index, count;
    unsigned long int sum = 0;
    double mean;

    /* initialize ULP data */
    for (index=0; index < MAXIMUM_NUMBER_OF_ASSOCIATIONS; index++) {
        ulpData[index].maximumStreamID    = -1;
        ulpData[index].nrOfReceivedChunks = 0;
        ulpData[index].nrOfReceivedBytes  = 0;
        ulpData[index].nrOfSentChunks = 0;
        ulpData[index].nrOfSentChunksTotal = 0;
        ulpData[index].nrOfSentBytes  = 0;
        for (count = 0; count < MAXIMUM_NUMBER_OF_STREAMS; count++) {
             ulpData[index].nrOfReceivedBytesPerStream[count] = 0;
             ulpData[index].nrOfReceivedChunksPerStream[count] = 0;
        }
        /* struct timeval start_time;  gets set when COMM_UP comes */
    }
    tid = 0;

    /* initialize the echo_ulp variable */
    trafficUlp.dataArriveNotif          = &dataArriveNotif;
    trafficUlp.sendFailureNotif         = &sendFailureNotif;
    trafficUlp.networkStatusChangeNotif = &networkStatusChangeNotif;
    trafficUlp.communicationUpNotif     = &communicationUpNotif;
    trafficUlp.communicationLostNotif   = &communicationLostNotif;
    trafficUlp.communicationErrorNotif  = &communicationErrorNotif;
    trafficUlp.restartNotif             = &restartNotif;
    trafficUlp.shutdownCompleteNotif    = &shutdownCompleteNotif;

    getArgs(argc, argv);
    checkArgs();

    SCTP_initLibrary(1);

    sctpInst = SCTP_registerInstance(TRAFFIC_PORT,
                                     MAXIMUM_NUMBER_OF_STREAMS, MAXIMUM_NUMBER_OF_STREAMS,
                                     noOfLocalAddresses, localAddressList,
                                     trafficUlp);
    /* set the TOS field */

    SCTP_getAssocDefaults(sctpInst, &instanceParameters);
    instanceParameters.rtoInitial = InitialRTO;
    /** the minimum timeout value */
    instanceParameters.rtoMin = MinRTO;
    /** the maximum timeout value */
    instanceParameters.rtoMax = MaxRTO;
    /** maximum retransmissions per association */
    instanceParameters.assocMaxRetransmits = AssocRetransmitLimit;
    /** maximum retransmissions per path */
    instanceParameters.pathMaxRetransmits = PathRetransmitLimit;
    /** maximum initial retransmissions */
    instanceParameters.maxInitRetransmits = InitialRetransmitLimit;
    /** recvcontrol: delay for delayed ACK in msecs */
    instanceParameters.delay = SackDelay;
    /* let me have an unlimited queue */
    instanceParameters.maxSendQueue = UnlimitedQueueLength;
    instanceParameters.ipTos=tosByte;
    SCTP_setAssocDefaults(sctpInst, &instanceParameters);
    if (sender) {
       SCTP_associate(sctpInst, 1, destAddress, TRAFFIC_PORT, &ulpData[0]);
    }

    if (doMeasurements) {
        SCTP_startTimer(deltaT, &measurementTimerRunOffFunction, NULL, NULL);
    }



    /* run the event handler forever */
    while (1) {
        SCTP_eventLoop();
    };

#if defined (BIENE)
    for (count = 0; count < 100000; count++) {
        sum += distribution(0.2, 0.02, 100.0);
    }
    mean = sum / 100000;
    printf ("Sum : %ld, Mean : %f\n", sum, mean);
#endif
    /* this will never be reached */
    exit(0);
}

