/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * SCTPd program. A UDP based Demon for handling SCTP associations.
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 *
 */


#include "sctp.h"

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>         /* for atoi() under Linux */

#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <fcntl.h>

/* Modules from SCTPd package*/
#include "../sctpdlib/sctpudputils.h"
#include "sctpdutils.h"
#include "sctpd_state.h"
#include "logger.h"
#include "configuration.h"
#include "pacemaker.h"

#include "glib.h"

#define MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES    10
#define MAXIMUM_NUMBER_OF_ASSOCIATIONS       20    
#define MAXIMUM_NUMBER_OF_IN_STREAMS         17
#define MAXIMUM_NUMBER_OF_OUT_STREAMS        17
#define min(x,y)            (x)<(y)?(x):(y)

/* Temp string shared by functions */
static unsigned char tempstr[255];

/* Level of logging required */
static short logLevel = 6;        /* Default to most verbose */

/* File descriptor to the UDP socket created 
   for listening to the UAs by the SCTP library */
static int udpid;
static int selfListeningFD;
static unsigned char *loopbackaddr      = "127.0.0.1";

/* Parameters to control the state of the daemon */
static unsigned int maxNoUA = 5;
static unsigned int maxNoAssoc = 50;

/* Counter of the number of instances currently 
 * registered with the sctp library 
 */
static int hasInstance=0;

/* SCTP Call back handlers for the daemon */
SCTP_ulpCallbacks ulp;


void printUsage(void) {
    printf("usage:   echo_tool [options] -s source_addr_1 -s source_addr_2 ...\n");
    printf("options:\n");
    printf("-L                  Log level\n");   
}

void getArgs(int argc, char **argv) {

  int c;
  extern char *optarg;
  extern int optind;

  while ((c = getopt(argc, argv, "bd:l:mn:p:s:t:L:")) != -1){
    switch (c) {
    case 'L':
      logLevel = atoi(optarg);
      break;

    }
  }
}


void configureDameon() {

  configParam params[10];
  int noOfParams,i;
  
  /* Check for LogLevel */
  findValue("LOG_LEVEL",params,&noOfParams);

  if (noOfParams) {
    /* Run through all instances, the last one being the effective setting. */
    for (i=0;i<noOfParams;i++) {
      logLevel = atoi(params[i].value);     
    }
  }

  /* Check for maximum number of UA supported */
  findValue("MAX_UA",params,&noOfParams);

  if (noOfParams) {
    /* Run through all instances, the last one being the effective setting. */
    for (i=0;i<noOfParams;i++) {
      maxNoUA = atoi(params[i].value);     
    }
  }

  /* Check for maximum number of Associations supported */
  findValue("MAX_ASSOC",params,&noOfParams);

  if (noOfParams) {
    /* Run through all instances, the last one being the effective setting. */
    for (i=0;i<noOfParams;i++) {
      maxNoAssoc = atoi(params[i].value);     
    }
  }

  /* Check for maximum number of Associations supported */
  findValue("HEARTBEAT_INT",params,&noOfParams);

  if (noOfParams) {
    /* Run through all instances, the last one being the effective setting. */
    for (i=0;i<noOfParams;i++) {
      setHeartbeatInt((unsigned)atoi(params[i].value));     
    }
  }

  /* Check for maximum number of Associations supported */
  findValue("HEARTBEAT_TIMEOUT",params,&noOfParams);

  if (noOfParams) {
    /* Run through all instances, the last one being the effective setting. */
    for (i=0;i<noOfParams;i++) {
      setHeartbeatTimeout((unsigned)atoi(params[i].value));     
    }
  }

}


/* Callback functions given to the SCTP library 
   to be called when UDP data arrived */
void recvUAMesg(int sockid,
		unsigned char * buf,
		int buflen,
		unsigned char srcAddr[],
		unsigned short srcPort) {

  if (buflen>0) {

    /* Call the function provided in the SCTPd programming utilities to 
       examine the message and dispatch the message to the respective
       registered UDP callback functions */
    if (!dispatchSCTPDMesg(buf)) {
      logMsg("Error in dispatching messages\n",monStatus);
    }
  }

}

void stopHeartbeats(unsigned short ua_id) {

  int heartbeat_id = 0;

#ifdef SHOW_HEARTBEATS
  fprintf(stdout,"Stopping heartbeats\n");
#endif

  /* Cancel all heartbeats for this UA */
  while((heartbeat_id=findHeartbeatForUA(ua_id))!=0) {            
    removeHeartbeat(heartbeat_id);
  }
  
  /* Cancel All Timers */
  cancelTimersForUA(ua_id);

}


int terminateAssociations(unsigned short ua_id,int residue) {

  int assocID=0;
  GSList * assocList=NULL;
	      
  if (residue==1) {
    /* There are outstanding associations */
    sctpd_state_uaID_to_assoc_list (ua_id, &assocList);

    /* Terminate all outstanding associations */
    while ((assocID=sctpd_state_get_next_assocID(assocList))!=-1 ) {
      
      sprintf(tempstr,"Sending SHUTDOWN to association ID = %d\n",assocID);
      logMsg(tempstr,monSCTPStatus);
      
      /* Call the sctp_shutdown for that association */
      if (sctp_shutdown(assocID)!=0) {          
	logMsg("Error in shutting down the association\n",monSCTPStatus);
      }      
    }
  }
  else if (residue==0) {   
    

  }

  return 0;

}


/* Test sanity */
/* Test to see if the State Tables are same as the actual situation */
int testSanity(){

  int i=0;
  int result=0;
  SCTP_InstanceParameters sctpInstanceParams;

  printf("*** Sanity Check ***\n");

  for (i=1;i<3;i++) {
    if ((sctp_getAssocDefaults(i,&sctpInstanceParams))==-1) {
      printf("Instance ID %d does not exists\n",i);
      result =-1;
    }
    else {
      printf("Instance ID %d exists\n",i);
    }
  }

  return result;

}


/* Handlers for SCTP notifications from the SCTP library ------------------------------------- */
void dataArriveNotif(unsigned int assocID, unsigned int streamID, unsigned int len,
                     unsigned int protoID, unsigned int unordered, void* ulpDataPtr) { 

  sprintf(tempstr, "%-4x: Data arrived (%u bytes on stream %u, %s)\n",
	  assocID, len, streamID, (unordered==SCTP_ORDERED_DELIVERY)?"ordered":"unordered");
  logMsg(tempstr,monSCTPStatus);

  /* Notify respective UA */
  notify_UA_dataArrive (assocID,streamID,len,protoID,unordered,ulpDataPtr);

}

void sendFailureNotif(unsigned int assocID,
                      unsigned char *unsent_data, unsigned int dataLength, 
		      unsigned int *context, void* dummy) {

  sprintf(tempstr, "%-4x: Send failure\n", assocID);
  logMsg(tempstr,monSCTPStatus);

  /* Notify respective UA */
  notify_UA_sendFailure(assocID,unsent_data,dataLength,context,dummy);


}

void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, 
			      unsigned short newState, void* ulpDataPtr) {

  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;
  unsigned short pathID;
    
  sctp_getPathStatus(assocID, affectedPathID, &pathStatus);
  sprintf(tempstr, "%-4x: Network status change: path %u (towards %s) is now %s\n", 
	  assocID, affectedPathID,
	  pathStatus.destinationAddress,
	  ((newState == SCTP_PATH_OK) ? "ACTIVE" : "INACTIVE"));
  logMsg(tempstr,monSCTPStatus);
      
  /* if the primary path has become inactive */
  if ((newState == SCTP_PATH_UNREACHABLE) &&
      (affectedPathID == sctp_getPrimary(assocID))) {
        
    /* select a new one */
    sctp_getAssocStatus(assocID, &assocStatus);
    for (pathID=0; pathID < assocStatus.numberOfAddresses; pathID++){
      sctp_getPathStatus(assocID, pathID, &pathStatus);
      if (pathStatus.state == SCTP_PATH_OK)
	break;
    }
        
    /* and use it */
    if (pathID < assocStatus.numberOfAddresses) {
      sctp_setPrimary(assocID, pathID);
    }
  }

  /* Notify respective UA */
  logMsg("Sending notification of networkStatusChange to UA\n",monStatus);
  notify_UA_networkStatusChange (assocID,affectedPathID ,newState,ulpDataPtr);


}

void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,unsigned short noOfInStreams, 
			   unsigned short noOfOutStreams, void* dummy) {	

  SCTP_PathStatus pathStatus;
  SCTP_AssociationStatus assocStatus;
  
  unsigned short pathID;
  unsigned char tstr[255];
  unsigned short instanceID;
  unsigned short sctp_port;
  unsigned short ua_id;

  /* Generate log message */
  sprintf(tempstr, "%-4x: Communication up (%u paths:", assocID, noOfPaths);
  for (pathID=0; pathID < noOfPaths; pathID++) {
    sctp_getPathStatus(assocID, pathID, &pathStatus);
    sprintf (tstr," %s", pathStatus.destinationAddress);
    strcat(tempstr, tstr);
  }
  sprintf(tstr, ")\n");
  strcat(tempstr,tstr);
  logMsg(tempstr,monSCTPStatus);

  /* Retrieve association status information */
  sctp_getAssocStatus(assocID,&assocStatus);
        
  /* get the instance id */
  if (sctp_getInstanceID(assocID, &instanceID)!=0) {
    logMsg("Error in getting instance ID for this association\n",monStatus);
    sctp_abort(assocID);
    return NULL;
  }

  /* Get the UA_ID for this instance */
  if (sctpd_state_instanceID_to_ua_id(instanceID,&ua_id)!=0) {
    sprintf(tempstr,"UA not found for this instance. ID=%u\n",instanceID);
    logMsg(tempstr,monStatus);
    sctp_abort(assocID);
    return NULL;
  }

  /* Get the SCTP Port number */
  if (sctpd_state_get_sctp_port(instanceID,&sctp_port)!=0) {
    sprintf(tempstr,"Unable to find the SCTP Port number for this Instance (id=%u)\n",
	    instanceID);
    logMsg(tempstr,monStatus);
    sctp_abort(assocID);
    return NULL;
  
  }

  /* Check the sctp_port to see if = 0 -> client */
  if (sctp_port == 0) {
    /* Assign the port used by the library for this association */
    sctp_port = assocStatus.sourcePort ;

    /* Update SCTP port number */
    if (sctpd_state_set_sctp_port(instanceID,sctp_port)!=0) {      
	logMsg("Unable to update SCTP Port number\n",monStatus);
	sctp_abort(assocID);
	return NULL;      
    }
  }

  /* Add the new association to internal state tables */
  if (sctpd_state_add_assoc(assocID,ua_id,instanceID)==0) {
   
    logMsg("Association information added into internal list \n",monStatus);
    
    /* Notify respective UA */
    sprintf(tempstr,"Sending notification of communicationUp (assocID=%d) to UA\n",assocID);
    logMsg(tempstr,monStatus);
    notify_UA_communicationUp (assocID,status,noOfPaths,noOfInStreams,noOfOutStreams,dummy);    

  }
  else {

    /* abort assoc */
    sctp_abort(assocID);
    logMsg("Unable to create a new association.",monStatus);
    
  }
  return NULL;
}



void communicationLostNotif(unsigned int assocID, unsigned short status, void* ulpDataPtr) {

  unsigned short ua_id;
  int res=0;

  sprintf(tempstr, "%-4x: Communication lost (status %u)\n", assocID, status);
  logMsg(tempstr,monSCTPStatus);      

  /* Ask the sctplib to remove this association */
  //sctp_deleteAssociation(assocID);

  /* Get the ua_id for this assoc */
  sctpd_state_assocID_to_ua_id(assocID,&ua_id); 

  /* Remove this association from State Tables */
  res = sctpd_state_remove_assoc(assocID);

  if (res>0) {

    /* A zombie instance has had its last assoc removed */
    /* Remove from SCTP library */
    sctp_unregisterInstance(res);

    /* Removal of instance will be done automatically in State */

    /* Send notification to UA */
    sprintf(tempstr,"Sending notification of communicationLost to UA (id=%d)\n",ua_id);
    logMsg(tempstr,monStatus);
    notify_UA_communicationLost(assocID,ua_id,status,ulpDataPtr);
    
  }
  else if (res ==0 ){

    /* Notify respective UA */
    logMsg("Sending notification of communicationLost to UA\n",monStatus);
    notify_UA_communicationLost(assocID,ua_id,status,ulpDataPtr);    
    
  }

}


void communicationErrorNotif(unsigned int assocID, unsigned short status, void* dummy) {
  
  sprintf(tempstr, "%-4x: Communication error (status %u)\n", assocID, status);
  logMsg(tempstr,monSCTPStatus);
  
 /* Notify respective UA */
  logMsg("Sending notification of communicationError to UA\n",monStatus);
  notify_UA_communicationError(assocID,status,dummy);


}

void restartNotif(unsigned int assocID, void* ulpDataPtr) {

  SCTP_AssociationStatus assocStatus;
      
  sprintf(tempstr, "%-4x: Restart\n", assocID);
  logMsg(tempstr,monSCTPStatus);
    
  sctp_getAssocStatus(assocID, &assocStatus);   

  /* Notify respective UA */
  logMsg("Sending notification of restart to UA\n",monStatus);
  notify_UA_restart(assocID,ulpDataPtr);

}


void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr) {

  unsigned short ua_id;
  int res=0;

  sprintf(tempstr, "%-4x: Shutdown complete\n", assocID);
  logMsg(tempstr,monSCTPStatus); 

  /* Get the ua_id for this assoc */
  sctpd_state_assocID_to_ua_id(assocID,&ua_id); 

  /* Remove this association from State Tables */
  res = sctpd_state_remove_assoc(assocID);

  if (res>0) {

    /* A zombie instance has had its last assoc removed */
    /* Remove from SCTP library */
    sctp_unregisterInstance(res);

    /* Removal of instance will be done automatically in State */

    /* Send notification to UA */
    sprintf(tempstr,"Sending notification of shutdownComplete to UA (id=%d)\n",ua_id);
    logMsg(tempstr,monStatus);
    notify_UA_shutdownComplete(assocID,ua_id,ulpDataPtr);
    
  }
  else if (res ==0 ){

    /* Send notification to UA */
    sprintf(tempstr,"Sending notification of shutdownComplete to UA (id=%d)\n",ua_id);
    logMsg(tempstr,monStatus);
    notify_UA_shutdownComplete(assocID,ua_id,ulpDataPtr);

  }
  else {
    logMsg("Error in removing State information about associuation\n",monStatus);
  }

  /* Ask the sctplib to remove this association */
  //sctp_deleteAssociation(assocID);

}



void queueStatusChangeNotif (unsigned int assocID, 
			     int queueType,
			     int queueID, 
			     int queueLength, 
			     void* ulpDataPtr) { 

  /* Notify respective UA */
  logMsg("Sending notification of queueStatusChange to UA\n",monStatus);
  notify_queueStatusChange (assocID,queueType,queueID,queueLength,ulpDataPtr);

}


void asconfStatusNotif (unsigned int assocID, 
			unsigned int correlationID, 
			int result, 
			void* tempstruct, 
			void* ulpDataPtr){


  /* Notify respective UA */
  logMsg("Sending notification of asconfStatus to UA\n",monStatus);
  notify_asconfStatus (assocID,correlationID,result,tempstruct,ulpDataPtr);


}




/* define some handlers for the func calls from the UA -------------------------------- */

/* Sends a 1 to the UA if successful and a 0 if failed */
void handle_sctpd_initLibrary(unsigned short pid,int udp_port,int udp_resp_port) {

  int t=0;
  int len=0;
  unsigned short res=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;    

  sprintf(tempstr,"SCTPD_INITLIBRARY message received\nPID=%d, udp_port=%d, udp_resp_port=%d\n",
	  pid,udp_port,udp_resp_port);
  logMsg(tempstr,monUA);
  
  if ((t=sctpd_state_register_ua(pid,udp_port,udp_resp_port,0))==0) {  

    sprintf(tempstr,"UA (id=%d, udp_port=%d, udp_resp_port=%d) Registered\n",
	    pid,udp_port,udp_resp_port);
    logMsg(tempstr,monStatus);

    /* Send SUCCESS response to the UA */
    res = 1;
    create_func_call (&fc,RESP_SCTPD_INITLIBRARY);
    add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(res),&res);    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRawData(udp_resp_port,udpdataarr,len);       

    /* Start the heartbeat mechanism */
    startPacemaker(pid);

  }
  else {

    sprintf(tempstr,"Unable to add UA info into State\nError = %d\n",t);
    logMsg(tempstr,monStatus);

    /* Send back a return value of 0 to signify a failure */
    res = 0;
    create_func_call (&fc,RESP_SCTPD_INITLIBRARY);
    add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(res),&res);    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRawData(udp_resp_port,udpdataarr,len);        
                  
  }

}


void handle_sctpd_stopLibrary(unsigned short ua_id) {

  int res=0;
  short instanceID=0;

  GSList *instanceID_list=NULL;

  logMsg("SCTPD_STOP_LIBRARY message received\n",monUA);

  /* Stop heartbeats */
  stopHeartbeats(ua_id);

  /* Remove UA from State */
  res = sctpd_state_unregister_ua(ua_id,1);

  switch (res) {
  case 0:              /* Sucessful normal unregistering */
    sprintf(tempstr,"Successful in unregistering UA %u from State.\n",
	    ua_id);    
    logMsg(tempstr,monStatus);

    break;

  case 1:              /* Outstanding associations exists */
    sprintf(tempstr,"Oustanding Associations exists for UA %u from State.\n",
	    ua_id);    
    logMsg(tempstr,monStatus);
    terminateAssociations(ua_id,1);
    break;

  case 2:              /* Outstanding instances ONLY exists */    

#ifdef DEBUG
    printf("Outstanding Instances exist\n");
#endif

    /* Get the list of oustanding instances */
    if (sctpd_state_uaID_to_instance_list (ua_id, &instanceID_list) != -1) {
      /* Loop */
      while ( (instanceID = sctpd_state_get_next_instanceID(instanceID_list))!=-1) {
	/* Remove from sctp library */
	sctp_unregisterInstance(instanceID);
	/* Remove from State */
	sctpd_state_remove_instance(instanceID);
      }	      
    }       
    break;
  default:             /* Error */

  } 
}


void handle_sctp_registerInstance(int ua_id,
				  unsigned short localPort,
				  unsigned short noOfInStreams,
				  unsigned short noOfOutStreams,
				  unsigned int   noOfLocalAddresses,
				  unsigned char  localAddressList[][SCTP_MAX_IP_LEN]){	      

  unsigned short sctpInstance=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;
  int len=0;

  logMsg("SCTPD_REGISTER_INSTANCE message received\n",monUA);

  /* Register this instance with the SCTP library */
  /* We do this before the registration with the State Tables
     becoz we need the instance id */
  sctpInstance = sctp_registerInstance(localPort,
				       noOfInStreams,
				       noOfOutStreams,
				       noOfLocalAddresses, 
				       localAddressList,
				       ulp);

#ifdef DEBUG
  fprintf(stdout,"New Instance created =%d\n",sctpInstance);
  fflush(stdout);
#endif

  if (sctpInstance!=0) {
    /* Increase the number of existing instances */
    hasInstance++;

    /* Update State tables */
    if (sctpd_state_add_instance (sctpInstance, localPort, ua_id) !=0 ) {
      /* Returning a zero instance name means an error */
      sctpInstance = 0;
      logMsg("Unsuccessful in adding instance info into State.\n",monStatus);

      /* Remove sctpInstance */
      sctp_unregisterInstance(sctpInstance);

    }        
    else { /* Success */      
      sprintf(tempstr,"New instance %d belonging to UA (%d) created\n",sctpInstance,ua_id);
      logMsg(tempstr,monStatus);
    }
  }
  else {
    logMsg("Unsuccessful in registering an instance with the SCTP library.\n",monStatus);
  }

  /* Send response to the UA */
  create_func_call (&fc,RESP_SCTP_REGISTER_INSTANCE);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(sctpInstance),&sctpInstance);
  
  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);
  sendnRespToUA(ua_id,udpdataarr,len);          

}


void handle_sctp_unregisterInstance(unsigned short instance_name) {

  unsigned short ua_id=0;
  int res=0;
  int noOfAssoc = 0;
  int assocID=0;
  int result=0;
  int len=0;
  GSList *assocID_list = NULL;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  logMsg("SCTPD_UNREGISTER_INSTANCE message received\n",monUA);

  /* Get the UA ID for this instance */
  sctpd_state_instanceID_to_ua_id(instance_name,&ua_id);

  /* Remove the Instance from State  */
  res = sctpd_state_remove_instance(instance_name);

  switch (res) {
  case 0:   /* Sucess */
    sprintf(tempstr,"Instance %d removed from State.\n",
	    instance_name);
    logMsg (tempstr,monStatus);

    /* Unregister with sctplib */
    sctp_unregisterInstance(instance_name);

    result = 0;
    break;

  case 1:   /* Oustanding assoc exists, marked as zombie */
    sprintf(tempstr,"Outstanding associations exists for Instance %d.\n",
	    instance_name);
    logMsg (tempstr,monStatus);

    /* Get the list of oustanding associations */
    if ((noOfAssoc = sctpd_state_instanceID_to_assoc_list (instance_name,
							   &assocID_list)) !=-1 ){   
      while ( (assocID = sctpd_state_get_next_assocID (assocID_list)) != -1) {
	/* Call sctp_shutdown for this association */
	sctp_shutdown (assocID);
      }
    }
    result = 0;

    break;

  case -1:  /* Error */

    sprintf(tempstr,"Unable to remove instance %d from State.\n",
	    instance_name);
    logMsg (tempstr,monStatus);
    result = -1;
    break;

  }   /* End switch */

  /* Send response to the UA */
  if (ua_id!=0) {  
    create_func_call (&fc,RESP_SCTP_UNREGISTER_INSTANCE);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(result),&result);   
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUA(ua_id,udpdataarr,len);          
  }
  else {
    sprintf(tempstr,"Error in retrieveing UA ID for instance %u\n",instance_name);
    logMsg(tempstr,monStatus);
  }

}


void handle_sctp_getAssocStatus(unsigned int assocID) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  SCTP_AssociationStatus assocStatus;
  fnCall fc;

  if (assocID>0){

    sprintf(tempstr,"sctp_getAssocStatus message : assocID=%d\n",assocID);
    logMsg(tempstr,monUA);
  
    if (!((retValue=sctp_getAssocStatus(assocID,&assocStatus))==-1)) {
    
      create_func_call (&fc,RESP_SCTP_GET_ASSOC_STATUS);
      add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
      add_func_arg(&fc,ARG_TYPE_ASSOC_STATUS,sizeof(assocStatus),&assocStatus);  

      /* pack and send the notifcation */
      len=pack_function_call(&fc,udpdataarr);
      sendnRespToUAAssocID(assocID,udpdataarr,len);    
      logMsg("Response to sctp_getAssocStatus message sent\n",monSCTPD);
    }
  }
}


void handle_sctp_setAssocDefaults(unsigned short SCTP_InstanceName, 
				  SCTP_InstanceParameters* params){

  int retValue;
  int len=0;
  unsigned short ua_id;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (SCTP_InstanceName>0) {
  
    sprintf(tempstr,"sctp_setAssocDefaults message : Instance=%d\n",SCTP_InstanceName);
    logMsg(tempstr,monUA);

    if ((retValue=sctp_setAssocDefaults(SCTP_InstanceName,params))!=-1) {
    
      create_func_call (&fc,RESP_SCTP_SET_ASSOC_DEFAULTS);
      add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

      /* pack and send the notifcation */
      len=pack_function_call(&fc,udpdataarr);

      /* Get the UA ID for this instance */
      sctpd_state_instanceID_to_ua_id(SCTP_InstanceName,&ua_id);

      sendnRespToUA(ua_id,udpdataarr,len); 
									  
      logMsg("Sending response to sctp_setAssocDefaults\n",monSCTPD);

    }
    else {
      sprintf(tempstr,"Instance %d does not exists.\n",
	      SCTP_InstanceName);
      logMsg(tempstr,monStatus);
    }

  }

}


void handle_sctp_setAssocStatus(unsigned int associationID, 
				SCTP_AssociationStatus* new_status) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_setAssocStatus message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    if ((retValue=sctp_setAssocStatus(associationID,new_status))!=-1){
    
      create_func_call (&fc,RESP_SCTP_SET_ASSOC_STATUS);
      add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

      /* pack and send the notifcation */
      len=pack_function_call(&fc,udpdataarr);
      sendnRespToUAAssocID(associationID,udpdataarr,len);        
									  
      logMsg("Sending response to sctp_setAssocStatus\n",monSCTPD);

    }
    else {
      sprintf(tempstr,"Unable to get association information for assocID = %d\n",
	      associationID);
      logMsg(tempstr,monStatus);
    }
  }
}


void handle_sctp_setPathStatus(unsigned int associationID, short path_id, 
			       SCTP_PathStatus *new_status) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0 && path_id>=0){
  
    sprintf(tempstr,"sctp_setPathStatus message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    if (!((retValue=sctp_setPathStatus(associationID,path_id,new_status))==-1)){
    
      create_func_call (&fc,RESP_SCTP_SET_PATH_STATUS);
      add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

      /* pack and send the notifcation */
      len=pack_function_call(&fc,udpdataarr);
      sendnRespToUAAssocID(associationID,udpdataarr,len);        
								     
      logMsg("Sending response to sctp_setPathStatus\n",monSCTPD);

    }
  }

}


void handle_sctp_getPathStatus(unsigned int assocID,short pathID) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  SCTP_PathStatus pathStatus;
  fnCall fc;

  if (assocID>0 && pathID>=0){
  
    sprintf(tempstr,"sctp_getPathStatus message : assocID=%d\n",assocID);
    logMsg(tempstr,monUA);

    retValue=sctp_getPathStatus(assocID,pathID,&pathStatus);
    create_func_call (&fc,RESP_SCTP_GET_PATH_STATUS);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

    if (!((retValue)==-1)){    
      add_func_arg(&fc,ARG_TYPE_PATH_STATUS,sizeof(pathStatus),&pathStatus);     
    }

    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(assocID,udpdataarr,len);
    logMsg("Sending response to sctp_getPathStatus\n",monSCTPD);

  }

}


void handle_sctp_getAssocDefaults(unsigned short instanceID) {

  int retValue;
  int len=0;
  unsigned short ua_id;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  SCTP_InstanceParameters params;
  fnCall fc;
 
  if (instanceID>0) {

    sprintf(tempstr,"sctp_getAssocDefaults message : instanceID=%d\n",instanceID);
    logMsg(tempstr,monUA);

    retValue=sctp_getAssocDefaults(instanceID,&params);

    create_func_call (&fc,RESP_SCTP_GET_ASSOC_DEFAULTS);  
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

    if (!((retValue)==-1)){
      add_func_arg(&fc,ARG_TYPE_INSTANCE_PARAMS,sizeof(params),&params);  
    }

    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);

    /* Get the UA ID for this instance */
    sctpd_state_instanceID_to_ua_id(instanceID,&ua_id);

    sendnRespToUA(ua_id,udpdataarr,len); 
    logMsg("Sending response to sctp_getAssocDefaults\n",monSCTPD);
  }

}


void handle_sctp_setPrimary(unsigned int associationID, short path_id) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0 && path_id>=0){
  
    sprintf(tempstr,"sctp_setPrimary message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_setPrimary(associationID,path_id);

    create_func_call (&fc,RESP_SCTP_SET_PRIMARY);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len);  
    logMsg("Sending response to sctp_setPrimary\n",monSCTPD);
    
  }

}

void handle_sctp_getPrimary(unsigned int associationID) {

  short retValue=-1;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_getPrimary message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_getPrimary(associationID);

    create_func_call (&fc,RESP_SCTP_GET_PRIMARY);
    add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len); 
    sprintf(tempstr,"Sending response to sctp_getPrimary (%d)\n",retValue);
    logMsg(tempstr,monSCTPD);
    
  }

}

void handle_sctp_getSrttReport(unsigned int associationID, short path_id) {

  unsigned int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_getSrttReport message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_getSrttReport(associationID,path_id);

    create_func_call (&fc,RESP_SCTP_GET_PRIMARY);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len); 
    logMsg("Sending response to sctp_getSrttReport\n",monSCTPD);
    
  }


}

void handle_sctp_changeHeartBeat(unsigned int associationID, 
				 short path_id, int heartbeatON, unsigned int timeIntervall) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_changeHeartBeat message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_changeHeartBeat(associationID,path_id,heartbeatON,timeIntervall);

    create_func_call (&fc,RESP_SCTP_CHANGE_HEARTBEAT);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len); 
    logMsg("Sending response to sctp_changeHeartbeat\n",monUA);
    
  }


}


void handle_sctp_requestHeartbeat(unsigned int associationID, short path_id) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_requestHeartbeat message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_requestHeartbeat(associationID,path_id);

    create_func_call (&fc,RESP_SCTP_REQUEST_HEARTBEAT);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len); 
    logMsg("Sending response to sctp_requestHeartbeat\n",monSCTPD);
    
  }

}

void handle_sctp_setFailureThreshold(unsigned int associationID,
				     unsigned short pathMaxRetransmissions) {
  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  if (associationID>0){
  
    sprintf(tempstr,"sctp_setFailureThreshold message : assocID=%d\n",associationID);
    logMsg(tempstr,monUA);

    retValue = sctp_setFailureThreshold(associationID,pathMaxRetransmissions);

    create_func_call (&fc,RESP_SCTP_SET_FAILURE_THRES);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
    
    /* pack and send the notifcation */
    len=pack_function_call(&fc,udpdataarr);
    sendnRespToUAAssocID(associationID,udpdataarr,len); 
    logMsg("Sending response to sctp_setFailureThreshold\n",monSCTPD);
    
  }

}


void handle_sctp_associate(unsigned short SCTP_InstanceName,
			   unsigned short noOfOutStreams,
			   unsigned char  *destinationAddress,
			   unsigned short destinationPort,
			   void* ulp_data) {

  int retValue;
  int len=0;
  unsigned short ua_id;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_associate message : instanceID=%d\n",SCTP_InstanceName);
  logMsg(tempstr,monUA);

#ifdef DEBUG
  printf("Connecting to %s on port %u\n",destinationAddress,destinationPort);
#endif

  retValue=sctp_associate(SCTP_InstanceName,noOfOutStreams,destinationAddress,
			  destinationPort,ulp_data);

  /* Send result to UA */
  create_func_call (&fc,RESP_SCTP_ASSOCIATE);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);

  /* Get the UA ID for this instance */
  sctpd_state_instanceID_to_ua_id(SCTP_InstanceName,&ua_id);

  sendnRespToUA(ua_id,udpdataarr,len); 

  sprintf(tempstr,"Sending response to sctp_associate (%d)\n",retValue);
  logMsg(tempstr,monSCTPD);

}



void handle_sctp_send (unsigned int associationID, unsigned short streamID,unsigned char *buffer, 
		       unsigned int length, unsigned int protocolId, short path_id,void * context, 
		       unsigned int lifetime, int unorderedDelivery, int dontBundle) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_send message (length=%d): instanceID=%d\n",length,associationID);
  logMsg(tempstr,monUA);

  /* Just pass the call on to the sctp lib */
  retValue = sctp_send(associationID,streamID,buffer,length,protocolId,path_id,context,lifetime,
		       unorderedDelivery,dontBundle);

  /* need to send return value back to the UA */
  create_func_call (&fc,RESP_SCTP_SEND);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_send\n",monSCTPD);

}


void handle_sctp_receive(unsigned int   associationID, unsigned short streamID,
			 unsigned char  *buffer, unsigned int   *length,
			 unsigned int flags) {

  int retValue;
  int len;
  int slength=0;
  unsigned char buf[SCTP_MAXIMUM_DATA_LENGTH]; /* TO DO: use a dynamically allocated
						  buffer with size given by length */
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_receive message : \nassocID = %d,streamID = %d\n",
	  associationID,streamID);
  logMsg(tempstr,monUA);

  len = sizeof(buf);
  retValue = sctp_receive(associationID,streamID,buf,&len,flags);

  if (retValue==0) {
#ifdef DEBUG
    fprintf(stdout,"Buffer read : length = %d\n", len);
    fflush(stdout);
#endif
  }
  else {
#ifdef DEBUG
    fprintf(stdout,"Invalid assocID\n");
    fflush(stdout);
#endif
  }

  create_func_call (&fc,RESP_SCTP_RECEIVE);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(streamID),&streamID);
  if (retValue==0) {
    add_func_arg(&fc,ARG_TYPE_CHAR_N,len,&buf);
    add_func_arg(&fc,ARG_TYPE_INT,sizeof(len),&len);
  }

  /* pack and send the notifcation */
  slength=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,slength); 
  logMsg("Sending response to sctp_receive\n",monSCTPD);

}


void handle_sctp_shutdown(unsigned int assocID) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_shutdown message : association ID=%d\n",assocID);
  logMsg(tempstr,monUA);

  /* Send return value back to the UA */
  retValue = sctp_shutdown(assocID);
  create_func_call (&fc,RESP_SCTP_SHUTDOWN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(assocID,udpdataarr,len); 

  logMsg("Sending response to sctp_shutdown\n",monSCTPD);

}


void handle_sctp_abort(unsigned int assocID) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_abort message : association ID=%d\n",assocID);
  logMsg(tempstr,monUA);

  /* need to send return value back to the UA */
  retValue = sctp_abort(assocID);

  create_func_call (&fc,RESP_SCTP_ABORT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(assocID,udpdataarr,len); 

  logMsg("Sending response to sctp_abort\n",monSCTPD);

}


void handle_sctp_deleteAssociation(unsigned int assocID) {

  int retValue;
  int len=0;
  unsigned short instanceID=0;
  unsigned short ua_id=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_deleteAssociation message : association ID=%d\n",assocID);
  logMsg(tempstr,monUA);

  /* Get the instance ID of this association */
  sctp_getInstanceID(assocID,&instanceID);
 
  fprintf(stdout,"instanceID %d\n",instanceID);
  fflush(stdout);

  /* Send return value back to the UA */
  retValue = sctp_deleteAssociation(assocID);
  create_func_call (&fc,RESP_SCTP_DELETE_ASSOC);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  

  /* Get the UA ID for this instance */
  sctpd_state_instanceID_to_ua_id(instanceID,&ua_id);

  sendnRespToUA(ua_id,udpdataarr,len); 

  logMsg("Sending response to sctp_deleteAssociation\n",monSCTPD);

}


void handle_sctp_getInstanceID(unsigned int assocID, unsigned short* instanceID) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  unsigned short id=0;
  unsigned short ua_id=0;
  fnCall fc;

  sprintf(tempstr,"sctp_getInstanceID message : association ID=%d\n",assocID);
  logMsg(tempstr,monUA);

  /* Get the instance ID of this association */
  retValue = sctp_getInstanceID(assocID,&id);
  
  /* Send return value back to the UA */
  create_func_call (&fc,RESP_SCTP_GET_INSTANCE_ID);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(id),&id);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  

  /* Get the UA_ID just in case that the association have already been removed from
     the state tables */ 
  sctpd_state_instanceID_to_ua_id(id,&ua_id);

  sendnRespToUA(ua_id,udpdataarr,len); 

  logMsg("Sending response to sctp_getInstanceID\n",monSCTPD);


}

void handle_sctp_receiveUnsent(unsigned int associationID, unsigned char *buffer, 
			       unsigned int *length,unsigned short *streamID, 
			       unsigned short *streamSN,unsigned int* protocolId) {

  int retValue,len;
  int slength=0;
  unsigned char buf[SCTP_MAXIMUM_DATA_LENGTH];  
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];

  unsigned char flags;
  unsigned int tsn;
  void* context_pointer;

  fnCall fc;

  sprintf(tempstr,"sctp_receiveUnsent message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  len = sizeof(buf);

  retValue = sctp_receiveUnsent(associationID, buf, length,
                                &tsn, streamID, streamSN, protocolId,
                                &flags, &context_pointer);

  create_func_call (&fc,RESP_SCTP_RECEIVE_UNSENT);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,len,&buf);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*length),length);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamID),streamID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamSN),streamSN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*protocolId),protocolId);

  /* pack and send the notifcation */
  slength = pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,slength); 
  logMsg("Sending response to sctp_receiveUnsent\n",monSCTPD);


}

void handle_sctp_receiveUnacked(unsigned int associationID, unsigned char *buffer, 
				unsigned int *length, unsigned short *streamID, 
				unsigned short *streamSN, unsigned int* protocolId){


  int retValue,len;
  int slength=0;
  unsigned char buf[SCTP_MAXIMUM_DATA_LENGTH];  
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];

  unsigned int tsn;
  unsigned char flags;
  void* ctx;

  fnCall fc;

  sprintf(tempstr,"sctp_receiveUnack message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  len = sizeof(buf);
  retValue = sctp_receiveUnacked(associationID,buf,length,&tsn,streamID,streamSN,protocolId,&flags,&ctx);

  create_func_call (&fc,RESP_SCTP_RECEIVE_UNACK);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(associationID),&associationID);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,len,&buf);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*length),length);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamID),streamID);
  add_func_arg(&fc,ARG_TYPE_SHORT,sizeof(*streamSN),streamSN);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*protocolId),protocolId);

  /* pack and send the notifcation */
  slength=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,slength); 
  logMsg("Sending response to sctp_receiveUnack\n",monSCTPD);


}


void handle_sctp_addIPAddress(unsigned int associationID,
			      unsigned char address[SCTP_MAX_IP_LEN],
			      unsigned int *correlationId) {

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_addIPAddress message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  //retValue = sctp_addIPAddress(associationID,address,correlationId);

  create_func_call (&fc,RESP_SCTP_ADD_IP_ADDRESS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* Don't return the IP addr */
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*correlationId),correlationId);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_addIPAddress\n",monSCTPD);


}


void handle_sctp_deleteIPAddress(unsigned int associationID, 
				 unsigned char address[SCTP_MAX_IP_LEN],
				 unsigned int* correlationId) {


  int retValue;
  int len=0;  
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_deleteIPAddress message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  //  retValue = sctp_deleteIPAddress(associationID,address,correlationId);

  create_func_call (&fc,RESP_SCTP_DELETE_IP_ADDRESS);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);
  /* Don't return the IP addr */
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(*correlationId),correlationId);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_deleteIPAddress\n",monSCTPD);


}


void handle_sctp_setRemotePrimary(unsigned int associationID, 
				  unsigned char address[SCTP_MAX_IP_LEN]) {
  
  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_setRemotePrimary message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  //retValue = sctp_setRemotePrimary(associationID,address);

  create_func_call (&fc,RESP_SCTP_SET_REMOTE_PRIMARY);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_setRemotePrimary\n",monSCTPD);

}


void handle_sctp_getNumberOfLocalAddresses(unsigned int associationID){

  int retValue;
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_getNumberOfLocalAddresses message : \nassocID = %d\n",
	  associationID);
  logMsg(tempstr,monUA);

  //retValue = sctp_getNumberOfLocalAddresses(associationID);

  create_func_call (&fc,RESP_SCTP_GET_NO_LOCAL_ADDR);
  add_func_arg(&fc,ARG_TYPE_INT,sizeof(retValue),&retValue);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_setRemotePrimary\n",monSCTPD);

}


void handle_sctp_getLocalAddress(unsigned int associationID, int index) {

  char retValue[SCTP_MAX_IP_LEN];
  int len=0;
  unsigned char udpdataarr[MAX_UDP_DATA_LENGTH];
  fnCall fc;

  sprintf(tempstr,"sctp_getLocalAddresses message : \nassocID = %d index=%d\n",
	  associationID,index);
  logMsg(tempstr,monUA);

  //strcpy(retValue,sctp_getLocalAddress(associationID,index));

  create_func_call (&fc,RESP_SCTP_GET_NO_LOCAL_ADDR);
  add_func_arg(&fc,ARG_TYPE_CHAR_N,SCTP_MAX_IP_LEN,retValue);

  /* pack and send the notifcation */
  len=pack_function_call(&fc,udpdataarr);  
  sendnRespToUAAssocID(associationID,udpdataarr,len); 
  logMsg("Sending response to sctp_getLocalAddress\n",monSCTPD);


}


/* Creates a socket to listen of the well-known daemon port */
int initSelfListening () {

  struct sockaddr_in listenaddr;
  
  selfListeningFD = socket(AF_INET, SOCK_DGRAM,0);
  bzero (&listenaddr,sizeof(listenaddr));

  listenaddr.sin_family = AF_INET;
  listenaddr.sin_port = htons(sctpdPort);

  if (inet_aton(loopbackaddr, &listenaddr.sin_addr)) {
    if ((bind(selfListeningFD, (struct sockaddr *) &listenaddr, sizeof(listenaddr)))==0) {
      sprintf(tempstr,"Listening on port=%d\n",sctpdPort);    
      logMsg(tempstr,monStatus);
      return 1;
    }
    else {
      logMsg("Unable to bind to UDP port\n",monStatus);
      return 0;
    }
  }
  else {
    logMsg("Unable to convert loopback address\n",monStatus);  
    return 0;
  }
      
}


int stopSelfListening() {

  close (selfListeningFD);

  return 1;

}


/* Wait on the well-known sctpd UDP port for incoming message.
 * This function will be used until an instance has been registered
 * and the monitoring of the UDP port can be passed to the sctp event loop 
*/
int wait_for_udp_mesg() {

  int len,n;
  struct sockaddr_in srcaddr;
  unsigned char temparr[MAX_UDP_DATA_LENGTH];

  bzero (&srcaddr,sizeof(srcaddr));
         
  len = sizeof(srcaddr);
  n=recvfrom(selfListeningFD,temparr,MAX_UDP_DATA_LENGTH,0, (struct sockaddr *) &srcaddr, &len);
  
  if (n>0){
    if (!dispatchSCTPDMesg(temparr)) {
      logMsg("Error in dispatching messages\n",monStatus);
      return 0;
    }      
  }

  return -1;

}      


int main(int argc, char **argv){

  UDP_SCTPD_CB udp_callbacks;

  /* Initialize the internal state tables */
  sctpd_state_set_max_no_of_ua (maxNoUA);
  sctpd_state_set_max_no_of_assoc (maxNoAssoc);
  
  /* Retrieve and check the arguements against what should be */
  // getArgs(argc, argv);

  /* Read configuration file */
  readConfig("sctpd.conf");
  
  /* Debug : Print out the configuration values */
#ifdef DEBUG
  printConfig();
#endif

  /* Configure daemon accordingly */
  configureDameon();

  /* Start the logging */
  startLogger("sctpd.log",logLevel);

  logMsg("SCTP daemon started\n",monStatus);   

  sctp_initLibrary();

  /* Initialize the SCTP callback functions variable */
  ulp.dataArriveNotif          = &dataArriveNotif;
  ulp.sendFailureNotif         = &sendFailureNotif;
  ulp.networkStatusChangeNotif = &networkStatusChangeNotif;
  ulp.communicationUpNotif     = &communicationUpNotif;
  ulp.communicationLostNotif   = &communicationLostNotif;
  ulp.communicationErrorNotif  = &communicationErrorNotif;
  ulp.restartNotif             = &restartNotif;
  ulp.shutdownCompleteNotif    = &shutdownCompleteNotif;
  ulp.queueStatusChangeNotif   = &queueStatusChangeNotif;
  ulp.asconfStatusNotif        = &asconfStatusNotif;

  /* Initialize the Callback functions for UDP Events */
  udp_callbacks.handle_sctp_getAssocStatus = &handle_sctp_getAssocStatus;
  udp_callbacks.handle_sctp_getPathStatus = &handle_sctp_getPathStatus;
  udp_callbacks.handle_sctp_setAssocStatus = &handle_sctp_setAssocStatus;
  udp_callbacks.handle_sctp_setPathStatus = &handle_sctp_setPathStatus;
  udp_callbacks.handle_sctp_send = &handle_sctp_send;
  udp_callbacks.handle_sctp_receive = &handle_sctp_receive;
  udp_callbacks.handle_sctp_shutdown = &handle_sctp_shutdown;
  udp_callbacks.handle_sctp_abort = &handle_sctp_abort;
  udp_callbacks.handle_sctp_getAssocDefaults = &handle_sctp_getAssocDefaults;
  udp_callbacks.handle_sctp_setAssocDefaults = &handle_sctp_setAssocDefaults;
  udp_callbacks.handle_sctp_setPrimary = &handle_sctp_setPrimary;
  udp_callbacks.handle_sctp_getPrimary = &handle_sctp_getPrimary;
  udp_callbacks.handle_sctp_getSrttReport = &handle_sctp_getSrttReport;
  udp_callbacks.handle_sctp_changeHeartBeat = &handle_sctp_changeHeartBeat;
  udp_callbacks.handle_sctp_requestHeartbeat = &handle_sctp_requestHeartbeat;
  udp_callbacks.handle_sctp_setFailureThreshold = &handle_sctp_setFailureThreshold;
  udp_callbacks.handle_sctp_deleteAssociation = &handle_sctp_deleteAssociation;
  udp_callbacks.handle_sctp_associate = &handle_sctp_associate;
  udp_callbacks.handle_sctp_registerInstance=&handle_sctp_registerInstance;
  udp_callbacks.handle_sctp_unregisterInstance=&handle_sctp_unregisterInstance;
  udp_callbacks.handle_sctpd_initLibrary=&handle_sctpd_initLibrary;
  udp_callbacks.handle_sctpd_stopLibrary=&handle_sctpd_stopLibrary;
  udp_callbacks.handle_sctp_getInstanceID=&handle_sctp_getInstanceID;
  udp_callbacks.handle_sctp_receiveUnsent=&handle_sctp_receiveUnsent;
  udp_callbacks.handle_sctp_receiveUnacked=&handle_sctp_receiveUnacked;
  udp_callbacks.handle_sctp_addIPAddress=&handle_sctp_addIPAddress;
  udp_callbacks.handle_sctp_deleteIPAddress=&handle_sctp_deleteIPAddress;
  udp_callbacks.handle_sctp_setRemotePrimary=&handle_sctp_setRemotePrimary;
  udp_callbacks.handle_sctp_getNumberOfLocalAddresses=&handle_sctp_getNumberOfLocalAddresses;
  udp_callbacks.handle_sctp_getLocalAddress=&handle_sctp_getLocalAddress;

  udp_callbacks.handle_heartbeat_ack = &handle_heartbeat_ack;
  udp_callbacks.handle_heartbeat_expired = &handle_heartbeat_expired;
  udp_callbacks.handle_heartbeat_int_expired = &handle_heartbeat_int_expired;
    
  register_SCTPD_UDP_Callbacks(udp_callbacks);

  /* Since no SCTP_Instance is created before any UA register with the daemon
     It is necessary to monitor the daemon's UDP port for registration before
     handing over the monitoring to the SCTP event loop. We cannot start the event
     loop before any instance is created.. Seg Fault. */

  if (initSelfListening()) {

    while (hasInstance==0) {    
      wait_for_udp_mesg();
    }

    stopSelfListening();

    /* First instance has registered, can hand over the monitoring of the
     * UDP to the sctp library 
     */
    logMsg("Registering with SCTP Event Loop for UDP ...\n",monStatus);      
    /* Register the UDP callback function with the SCTP event loop */
    udpid=sctp_registerUdpCallback("127.0.0.1",sctpdPort,&recvUAMesg);

    logMsg("done!\n",monStatus);

    /* Check if we got a valid file descriptor from the sctplib */
    if (udpid!=-1) {
      sprintf(tempstr,"Successful! socket file descriptor=%d\n",udpid);
      logMsg(tempstr,monStatus);
  
      /* run the sctp event handler forever */
      while (sctp_eventLoop() >= 0) { };      
    }
    else {
      logMsg("Error in registering sctpd well-know UDP port with the sctp library\n",
	     monStatus);

      /* We cannot continue, terminate daemon with error */
      /* Stop the logging */
      stopLogger();
      exit(1);	    
    }
  }
  /* this will never be reached */
  return 0;
}














