/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Purpose: Module header for administrating the state of several ULP tha
 *          may register with the local SCTP demon.
 *
 */

#ifndef SCTPD_STATE_H
#define SCTPD_STATE_H


#include "glib.h"

/***********************DEFINITION OF CONSTANTS**************************/

/* A UA is defined as a zombie UA when it has gone dead but there are still 
   associations belonging to it. This normally happens when the UA crashes, 
   therefore deregistration is not done */
#define UA_ZOMBIE_STATE          0
#define UA_ACTIVE_STATE          1

/* An instance is defined as a zombie when a UA do a "remove instance" but 
   there are still associations belonging to it. */
#define INSTANCE_ZOMBIE_STATE    0
#define INSTANCE_ACTIVE_STATE    1

/* Search Fields */
#define SEARCH_UA_ID             1     
#define SEARCH_ASSOC_ID          2
#define SEARCH_INSTANCE_ID       3
#define SEARCH_UDP_NOTIF_PORT    4
#define SEARCH_UDP_RESP_PORT     5
#define SEARCH_SCTP_PORT         6
#define SEARCH_STATE             7


/******************** Function Definitions *******************************/

/*  This function returns 0 if successful,
 *   1 if there is a duplicate ua
 *   2 if the Max no. of UA is reached   
 *  -1 if otherwise.
 *  does a force registration if force set to 1
 */
int sctpd_state_register_ua (unsigned short ua_id, 
			     unsigned short udp_port, 
			     unsigned short udp_resp_port, 
			     short force);


/*  This function returns 0 if successful,
 *   1 if there is outstanding associations,
 *   2 if there are outstanding instances, 
 *   3 if UA not found
 *  -1 if an error occurred.
 *  does a force registration if force set to 1, ie
 */
int sctpd_state_unregister_ua (unsigned short ua_id, short force);


/*  This function returns 0 if successful,
 *   1 if UA does not exist,
 *   2 if the MAX no. of assoc is exceeded
 *   3 if the UA is a Zombie UA
 *   4 if the instance does not exist
 *   5 if the Instance is a Zombie Instance
 *  -1 if an error occurred.
 */
int sctpd_state_add_assoc (unsigned int assocID, unsigned short ua_id,
			   unsigned short instanceID);


/*  This function returns 0 if successful,
 *  -1 if an error occurred.
 *  A non-zero return value indicates the zombie Instance which
 *  has been removed from the state, since its last
 *  outstanding Instance has been removed. This non-zero return value
 *  is the instance ID of association that was removed.
 */
int sctpd_state_remove_assoc (unsigned int assocID);


/*  This function returns 0 if successful,
 *   1 if UA does not exist,
 *   2 if the MAX no. of instance is exceeded
 *   3 if the UA is a Zombie UA
 *   4 if the SCTP port is used
 *  -1 if an error occurred.
 */
int sctpd_state_add_instance (unsigned short instanceID, 
			      unsigned short sctp_port,
			      unsigned short ua_id);



/*  This function returns 0 if successful,
 *  -1 if an error occurred.
 *   1 if there are outstanding associations
 *   2 if the instance does not exist
 */
int sctpd_state_remove_instance (unsigned int instanceID);



/*  This function returns the no. of associations if successful,
 *  -1 if no association is found.
 */
int sctpd_state_uaID_to_assoc_list (unsigned short ua_id, 
				    GSList **assocID_list);


/*  This function returns the no. of associations if successful,
 *  -1 if no association is found.
 */
int sctpd_state_instanceID_to_assoc_list (unsigned short instance_id, 
					  GSList **assocID_list);


/*  This function returns the no. of instances if successful,
 *  -1 if no instance is found.
 */
int sctpd_state_uaID_to_instance_list (unsigned short ua_id, 
				       GSList **instanceID_list);


/*  This function is a helper function to retrieve the next remaining assoc ID
 *  belonging to a Zombie UA so as to call sctp_shutdown for all remaining 
 *  associations. 
 *  In making this function call, use a while loop so that the counter is reset
 *  when there are no more association ID entry in the list. 
 *  This function returns the next association ID in the GSList if it exists, 
 *  -1 if otherwise.
 */
int sctpd_state_get_next_assocID (GSList *assocID_list);


/*  This function is a helper function to retrieve the next remaining instance
 *  ID belonging to a Zombie UA so as to call sctp_UnregisterInstance for all  
 *  remaining associations. 
 *  In making this function call, use a while loop so that the counter is reset
 *  when there are no more instance ID entry in the list. 
 *  This function returns the next instance ID in the GSList if it exists, 
 *  -1 if otherwise.
 */
short sctpd_state_get_next_instanceID (GSList *instanceID_list);


/* This function returns the next UA id from a glist of uaID_list */
short sctpd_state_get_next_uaID (GSList *uaID_list);


/*  This function returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_assocID_to_ua_id (unsigned int assocID, unsigned short* ua_id);


/*  This function returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_assocID_to_instanceID (unsigned int assocID, 
				       unsigned short* instanceID);


/*  This function returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_instanceID_to_ua_id (unsigned short instanceID,
				     unsigned short* ua_id);


/*  This function returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_get_udp_notif_port (unsigned short ua_id, 
				    unsigned short* udp_notif_port);


/*  This function returns the 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_get_udp_resp_port (unsigned short ua_id, 
				   unsigned short* udp_resp_port);


/*  This function returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_get_sctp_port (unsigned short instance_id, 
			       unsigned short* sctp_port);


/*  This function sets the SCTP port,
 *  returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_set_sctp_port(unsigned short instance_id, 
			      unsigned short sctp_port);


/*  This function returns 0 if UA exists,
 *  -1 if otherwise.
 */
int sctpd_state_check_ua_exist (unsigned short ua_id);


/*  This function sets the maximum number of UA the daemon can handle,
 *  returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_set_max_no_of_ua (unsigned short max_no_of_ua);


/*  This function sets the MAX number of associations the daemon can handle,
 *  returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_set_max_no_of_assoc (unsigned short max_no_of_assoc);


/*  This function sets the MAX number of instances the daemon can handle,
 *  returns 0 if successful,
 *  -1 if otherwise.
 */
int sctpd_state_set_max_no_of_instance (unsigned short max_no_of_instance);



/* Utility functions to get port numbers */
unsigned short getUDPNotifPort(int assocID);
unsigned short getUDPRespPort(int assocID);
unsigned short getUDPRespPortForInstance(unsigned short instanceID);
unsigned short getUDPNotifPortForUA(unsigned short UA_ID);
unsigned short getUDPRespPortForUA(unsigned short UA_ID);

/* Print all Utilities */

void PrintAll_UAInfoList();
void PrintAll_AssocInfoList();
void PrintAll_InstanceInfoList();


/* Utility functions that return the list of IDs */
/* These functions return the number of records found if successful
   -1 if no records are found */
int GetAll_uaID(GSList **ua_ID);
int GetAll_assocID(GSList **assoc_ID);
int GetAll_instanceID(GSList **instance_ID);


#endif




















