/*
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 2000 by Siemens AG, Munich, Germany.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at http://www.sctp.de which should be
 * used for any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          Michael.Tuexen@icn.siemens.de
 *          ajung@exp-math.uni-essen.de
 *
 * Daemon-based daytime server
 * Author : Sim Woon Chiat Siemens ICM N PG U SE D2
 *
 */


#include <stdio.h>
#include <stdlib.h>         /* for atoi() under Linux */
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <glib.h>
#include "../sctpdlib/ua.h"


/* Constants declaration */
#define UDP_Notif_Port      5234
#define UDP_Resp_Port   5235
#define SCTP_App_Port     13                  /* Daytime service port */
#define MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES 5


/* Variables declaration */
static unsigned char localAddressList[MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES][SCTP_MAX_IP_LEN];
static unsigned short noOfLocalAddresses = 0;
static int unknownCommand = 0;
static unsigned short instanceID=0;


/* Exit function */
void cleanUpAndExit() {

  /* Unregister Instance with sctpd */
  sctpd_unregisterInstance(instanceID);

  /* Stop the sctpdlib */
  sctpd_stopLibrary();
 
  exit(0);
}


/* Output status / error message */
void displayStatus(char *mesg) {

  fprintf(stdout,mesg);
  fflush(stdout);

}

/* Arguments Functions ------------------------------------------------------- */
void printUsage(void) {

  printf("Usage:   daytime_server [options] -s source_addr_1 -s source_addr_2 ...\n");

}

void getArgs(int argc, char **argv) {
  
  int c;
  extern char *optarg;
  extern int optind;

  while ((c = getopt(argc, argv, "s:")) != -1) {
    switch (c) {
    case 's':
      if ((noOfLocalAddresses < MAXIMUM_NUMBER_OF_LOCAL_ADDRESSES) &&
	  (strlen(optarg) < SCTP_MAX_IP_LEN  )) {
	strcpy(localAddressList[noOfLocalAddresses], optarg);
	noOfLocalAddresses++;
      }
      break;
    default:
      unknownCommand = 1;
      break;
    }
  }
}

void checkArgs(void) {

  int abortProgram;
  int printUsageInfo;
    
  abortProgram = 0;
  printUsageInfo = 0;
    
  if (noOfLocalAddresses == 0) {
    printf("Error:   At least one sourceaddress must be specified.\n");
    abortProgram = 1;
    printUsageInfo = 1;
  }
  if (unknownCommand ==1) {
    printf("Error:   Unkown options in command.\n");
    printUsageInfo = 1;
  }
  
  if (printUsageInfo == 1)
    printUsage();
  if (abortProgram == 1)
    exit(-1);
}



/* UDP Callback Handlers ----------------------------------------------- */
void* communicationUpNotif(unsigned int assocID, unsigned short status,
                           int noOfPaths,unsigned short noOfInStreams, 
			   unsigned short noOfOutStreams, void* dummy) {

  char *timeAsString;
  char tstr[255];
  time_t now;
  unsigned short sctpInstance=0;
  int res=0;
  int no_of_args,i;
  unsigned char *p;
  SCTP_AssociationStatus assocStatus;
  SCTP_PathStatus pathStatus;

  displayStatus("Reveived a CommunicationUpNotif from SCTP layer\n");
  
  /* get the current time and convert to string */
  time(&now);
  timeAsString = (char*)ctime(&now);
   
  sctpd_send(assocID,
	     0,
	     timeAsString, strlen(timeAsString),
	     SCTP_GENERIC_PAYLOAD_PROTOCOL_ID,
	     SCTP_USE_PRIMARY, SCTP_NO_CONTEXT, 
	     SCTP_INFINITE_LIFETIME, SCTP_ORDERED_DELIVERY, SCTP_BUNDLING_DISABLED);
    
  sctpd_shutdown(assocID);


}


void shutdownCompleteNotif(unsigned int assocID, void* ulpDataPtr) {

  displayStatus("Shutdown completed message received!\n");

}


void networkStatusChangeNotif(unsigned int assocID, short affectedPathID, 
			      unsigned short newState, void* ulpDataPtr) {

  displayStatus("Network status changed Notification message received!\n");

}


void communicationLostNotif(unsigned int assocID, unsigned short status, 
			    void* ulpDataPtr) {

  char tstr[255];

  sprintf(tstr,"Communication Lost for association (ID=%d)\n",assocID);
  displayStatus(tstr);

}



gint main (gint argc, gchar * argv[]) {

  UDP_UA_CB udp_callbacks;
  char tempstr[255];

  /* Trap Ctrl-C */
  signal(SIGINT, cleanUpAndExit);

  /* handle all command line options */
  getArgs(argc, argv);
  checkArgs();  

  /* Register the UDP callbacks with the dispatch mechanism */
  udp_callbacks.communicationUpNotif = &communicationUpNotif;
  udp_callbacks.shutdownCompleteNotif = &shutdownCompleteNotif;
  udp_callbacks.networkStatusChangeNotif = &networkStatusChangeNotif;

  /* Initialize sctplib */
  if (sctpd_initLibrary(UDP_Notif_Port,UDP_Resp_Port)==1) {

    displayStatus("Registering an Instance with the sctpd\n");
      
    /* Call sctpd_registerInstance */
    if ((instanceID = sctpd_registerInstance(SCTP_App_Port,10,10,noOfLocalAddresses,
					     localAddressList,udp_callbacks)) !=0 ) {
      
      displayStatus("Daytime server started ...\n");
	
      /* Start the eventloop */
      sctpd_eventLoop();
      
    }
      
  }
  else {
    displayStatus("Unable to init sctpdlib\n");
    exit (1);
  }
  
  /* Never reached */
  exit (0);

}




















