/*
 *  $Id: sctpassociation.cc,v 1.21 2001/12/06 12:04:06 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: SCTP Association
 *
 */


#include "tdsystem.h"
#include "sctpassociation.h"
#include "sctpsocketmaster.h"


namespace Coral {



// #define PRINT_SHUTDOWN



// ###### Constructor #######################################################
SCTPAssociation::SCTPAssociation(SCTPSocket*        socket,
                                 const unsigned int associationID,
                                 const unsigned int notificationFlags)
{
   Socket                        = socket;
   AssociationID                 = associationID;
   CommunicationUpNotification   = false;
   CommunicationLostNotification = false;
   ShutdownCompleteNotification  = false;
   IsShuttingDown                = false;
   UseCount                      = 0;
   LastUsage                     = getMicroTime();
   NotificationFlags             = notificationFlags;

   EstablishCondition.setName("SCTPAssociation::EstablishCondition");
   ShutdownCompleteCondition.setName("SCTPAssociation::ShutdownCompleteCondition");
   ReadyForTransmit.setName("SCTPAssociation::ReadyForTransmitCondition");
   InQueue.getUpdateCondition()->setName("SCTPAssociation::InQueue");
   ReadUpdateCondition.setName("SCTPAssociation::ReadUpdateCondition");
   WriteUpdateCondition.setName("SCTPAssociation::WriteUpdateCondition");
   ExceptUpdateCondition.setName("SCTPAssociation::ExceptUpdateCondition");

   InQueue.getUpdateCondition()->addParent(&ReadUpdateCondition);
   EstablishCondition.addParent(&ExceptUpdateCondition);
   ShutdownCompleteCondition.addParent(&ExceptUpdateCondition);
   ReadyForTransmit.addParent(&WriteUpdateCondition);

   SCTPSocketMaster::MasterInstance.lock();
   Socket->AssociationList.insert(pair<unsigned int, SCTPAssociation*>(AssociationID,this));
   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Destructor ########################################################
SCTPAssociation::~SCTPAssociation()
{
   SCTPSocketMaster::MasterInstance.lock();
   if(AssociationID == 0) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPAssociation::~SCTPAssociation() - "
              "AssociationID is 0! Destructor called twice?!" << endl;
#endif
      return;
   }

   // ====== Do shutdown ====================================================
   if(!ShutdownCompleteNotification) {
#ifdef PRINT_SHUTDOWN
      cout << "Active shutdown of association #" << AssociationID << " started..." << endl;
#endif
      SCTPSocketMaster::delayedDeleteAssociation(Socket->getID(),AssociationID);
      shutdown();
#ifdef PRINT_SHUTDOWN
      cout << "Active shutdown of association #" << AssociationID << " complete!" << endl;
#endif
   }
   else {
#ifdef PRINT_SHUTDOWN
      cout << "Passive shutdown of association #" << AssociationID << "!" << endl;
#endif
      sctp_deleteAssociation(AssociationID);
   }

   // ====== Remove association from list ================================
   multimap<unsigned int, SCTPAssociation*>::iterator iterator =
      Socket->AssociationList.find(AssociationID);
   if(iterator != Socket->AssociationList.end()) {
      Socket->AssociationList.erase(iterator);
   }
   else {
#ifndef DISABLE_WARNINGS
      cerr << "INTERNAL ERROR: SCTPAssociation::~SCTPAssociation() - "
              "Erase of association #" << AssociationID << " failed!" << endl;
      ::exit(1);
#endif
   }

   // ====== Clear ID, to make finding dangling references easier ===========
   AssociationID = 0;

   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Get local address #################################################
bool SCTPAssociation::getLocalAddresses(SocketAddress**& addressArray)
{
   return(Socket->getLocalAddresses(addressArray));
}


// ###### Get remote address ################################################
bool SCTPAssociation::getRemoteAddresses(SocketAddress**& addressArray)
{
   if(addressArray != NULL) {
      addressArray = NULL;
   }
   bool                    result = true;
   SCTP_Association_Status status;
   addressArray = NULL;

   SCTPSocketMaster::MasterInstance.lock();

   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      addressArray = SocketAddress::newAddressList(status.numberOfAddresses);
      if(addressArray == NULL) {
         return(false);
      }
      unsigned int i;
      for(i = 0;i < status.numberOfAddresses;i++) {
         SCTP_Path_Status pathStatus;
         if(sctp_getPathStatus(AssociationID,i,&pathStatus) != 0) {
#ifndef DISABLE_WARNINGS
            cerr << "INTERNAL ERROR: SCTPSocketMaster::getRemoteAddress() - sctp_getPathStatus() failure!"
                 << endl;
            ::exit(1);
#endif
         }
         else {
            addressArray[i] = SocketAddress::createSocketAddress(
                                 0,(char*)&pathStatus.destinationAddress,status.destPort);
            if(addressArray[i] == NULL) {
#ifndef DISABLE_WARNINGS
               cerr << "WARNING: SCTPAssociation::getRemoteAddresses() - Bad address "
                    << pathStatus.destinationAddress << ", port " << status.destPort << "!" << endl;
#endif
               SocketAddress::deleteAddressList(addressArray);
               result = false;
            }
         }
      }
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Check, if socket has data for given flags #########################
bool SCTPAssociation::hasData(const int flags)
{
   SCTPSocketMaster::MasterInstance.lock();
   const bool result = InQueue.hasData(flags, NotificationFlags);
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Receive ###########################################################
int SCTPAssociation::receive(char*           buffer,
                             size_t&         bufferSize,
                             int&            flags,
                             unsigned short& streamID,
                             unsigned int&   protoID)
{
   unsigned int assocID = AssociationID;
   streamID             = (unsigned short)-1;
   SCTPNotification notification;

   const int result = Socket->internalReceive(
                                 InQueue,
                                 buffer, bufferSize,
                                 flags,
                                 assocID, streamID, protoID,
                                 notification,
                                 NotificationFlags);
   return(result);
}


// ###### Receive ###########################################################
int SCTPAssociation::receiveFrom(char*             buffer,
                                 size_t&           bufferSize,
                                 int&              flags,
                                 unsigned short&   streamID,
                                 unsigned int&     protoID,
                                 SocketAddress***  addressArray,
                                 SCTPNotification& notification)
{
   // ====== Receive data ===================================================
   unsigned int assocID = AssociationID;
   streamID             = (unsigned short)-1;
   const int result = Socket->internalReceive(
                         InQueue,
                         buffer, bufferSize,
                         flags,
                         assocID, streamID, protoID,
                         notification,
                         NotificationFlags);

   // ====== Extract sender's addresses =====================================
   if((bufferSize > 0) && (addressArray != NULL)) {
      unsigned int i;
      *addressArray = SocketAddress::newAddressList(notification.RemoteAddresses);
      if(*addressArray != NULL) {
         for(i = 0;i < notification.RemoteAddresses;i++) {
            (*addressArray)[i] = SocketAddress::createSocketAddress(
                                    0,
                                    (char*)&notification.RemoteAddress[i],
                                    notification.RemotePort);
         }
      }
   }

   return(result);
}


// ###### Send ##############################################################
int SCTPAssociation::send(const char*          buffer,
                          const size_t         length,
                          const int            flags,
                          const unsigned short streamID,
                          const unsigned int   protoID)
{
   const int result = Socket->internalSend(
                         buffer, length,
                         flags,
                         AssociationID, streamID, protoID,
                         &ReadyForTransmit);
   return(result);
}


// ###### Shutdown ##########################################################
void SCTPAssociation::shutdown()
{
   SCTPSocketMaster::MasterInstance.lock();
   if(!IsShuttingDown) {
      IsShuttingDown = true;
      sctp_shutdown(AssociationID);
   }
   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Abort #############################################################
void SCTPAssociation::abort()
{
   SCTPSocketMaster::MasterInstance.lock();
   IsShuttingDown = true;
   sctp_abort(AssociationID);
   SCTPSocketMaster::MasterInstance.unlock();
}


// ###### Get association parameters ########################################
bool SCTPAssociation::getAssociationParameters(
                         SCTP_Association_Status& associationParameters)
{
   return(Socket->getAssociationParameters(AssociationID,associationParameters));
}


// ###### Set association parameters ########################################
bool SCTPAssociation::setAssociationParameters(
                         const SCTP_Association_Status& associationParameters)
{
   return(Socket->setAssociationParameters(AssociationID,associationParameters));
}


// ###### Get path parameters ###############################################
bool SCTPAssociation::getPathParameters(const int          pathIndex,
                                        SCTP_PathStatus&   pathParameters)
{
   return(Socket->getPathParameters(AssociationID,pathIndex,pathParameters));
}


// ###### Set path parameters ###############################################
bool SCTPAssociation::setPathParameters(const int              pathIndex,
                                        const SCTP_PathStatus& pathParameters)
{
   return(Socket->setPathParameters(AssociationID,pathIndex,pathParameters));
}


// ###### Get send buffer size ##############################################
ssize_t SCTPAssociation::getSendBuffer()
{
   ssize_t result = -1;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      result = status.maxSendQueue;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Set send buffer size ##############################################
bool SCTPAssociation::setSendBuffer(const size_t size)
{
   bool result = false;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      status.maxSendQueue = size;
      if(sctp_setAssocStatus(AssociationID,&status) == 0) {
         result = true;
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Get receive buffer size ###########################################
ssize_t SCTPAssociation::getReceiveBuffer()
{
   ssize_t result = -1;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      result = status.maxRecvQueue;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Set receive buffer size ###########################################
bool SCTPAssociation::setReceiveBuffer(const size_t size)
{
   bool result = false;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      status.maxRecvQueue = size;
      if(sctp_setAssocStatus(AssociationID,&status) == 0) {
         result = true;
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}

// ###### Set primary address ###############################################
SocketAddress* SCTPAssociation::getPrimaryAddress()
{
   return(Socket->getPrimaryAddress(AssociationID));
}


// ###### Set primary address ###############################################
bool SCTPAssociation::setPrimary(const SocketAddress& primary)
{
   return(Socket->setPrimary(AssociationID,primary));
}


// ###### Set peer primary address ##########################################
bool SCTPAssociation::setPeerPrimary(const SocketAddress& primary)
{
   return(Socket->setPeerPrimary(AssociationID,primary));
}


// ###### Add address #######################################################
bool SCTPAssociation::addAddress(const SocketAddress& addAddress)
{
   return(Socket->addAddress(AssociationID,addAddress));
}


// ###### Delete address ####################################################
bool SCTPAssociation::deleteAddress(const SocketAddress& delAddress)
{
   return(Socket->deleteAddress(AssociationID,delAddress));
}



// ###### Get traffic class #################################################
int SCTPAssociation::getTrafficClass(const int streamID)
{
   ssize_t result = -1;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      result = status.ipTos;
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Set traffic class #################################################
bool SCTPAssociation::setTrafficClass(const card8 trafficClass,
                                      const int   streamID)
{
   bool result = false;
   SCTPSocketMaster::MasterInstance.lock();
   SCTP_Association_Status status;
   if(sctp_getAssocStatus(AssociationID,&status) == 0) {
      status.ipTos = trafficClass;
      if(sctp_setAssocStatus(AssociationID,&status) == 0) {
         result = true;
      }
   }
   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


}
