/*
 *  $Id: sctpnotificationqueue.cc,v 1.5 2001/12/06 12:04:06 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Chunk Arrival Queue
 *
 */



#include "tdsystem.h"
#include "sctpnotificationqueue.h"
#include "sctp.h"


namespace Coral {


// ###### Constructor #######################################################
SCTPNotificationQueue::SCTPNotificationQueue()
{
   UpdateCondition.setName("SCTPNotificationQueue::UpdateCondition");
   First = NULL;
   Last  = NULL;
   Count = 0;
}


// ###### Destructor ########################################################
SCTPNotificationQueue::~SCTPNotificationQueue()
{
   flush();
}


// ###### Add notification to end of queue ##################################
bool SCTPNotificationQueue::addNotification(const SCTPNotification& notification)
{
   SCTPNotification* newNotification = new SCTPNotification;
   if(newNotification != NULL) {
      *newNotification = notification;
     newNotification->NextNotification = NULL;

      if(Last != NULL) {
         Last->NextNotification = newNotification;
      }
      Last = newNotification;
      if(First == NULL) {
         First = newNotification;
      }
      Count++;

      signal();
      return(true);
   }
#ifndef DISABLE_WARNINGS
   cerr << "ERROR: SCTPNotificationQueue::addNotification() - Out of memory!" << endl;
#endif
   return(false);
}


// ###### Get and remove chunk from queue ##################################
bool SCTPNotificationQueue::getNotification(SCTPNotification& notification)
{
   if(First != NULL) {
      notification = *First;
      return(true);
   }
   return(false);
}


// ###### Add notification to top of queue ##################################
void SCTPNotificationQueue::updateNotification(const SCTPNotification& notification)
{
   if(First != NULL) {
      SCTPNotification* next = First->NextNotification;
      *First = notification;
      First->NextNotification = next;
   }
   else {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SCTPNotificationQueue::updateHeadNotification() - Queue is empty!" << endl;
#endif
   }
}


// ###### Drop first notification ###########################################
void SCTPNotificationQueue::dropNotification()
{
   if(First != NULL) {
      SCTPNotification* next = First->NextNotification;
      if(Last == First) {
         Last = next;
      }
      delete First;
      First = next;
      Count--;
   }
}


// ###### Flush all chunks #################################################
void SCTPNotificationQueue::flush()
{
   SCTPNotification* notification = First;
   while(notification != NULL) {
      SCTPNotification* next = notification->NextNotification;
      delete notification;
      notification = next;
   }
   First = NULL;
   Last  = NULL;
   Count = 0;
}


// ###### Check, if queue has data to read for given flags ##################
bool SCTPNotificationQueue::hasData(const int          flags,
                                    const unsigned int notificationFlags)
{
   const bool receiveNotifications = (flags & MSG_NOTIFICATION);

   SCTPNotification* notification = First;
   while(notification != NULL) {
      const sctp_notification_header* header = &notification->Content.sn_notification_header;
      if(/* (receiveNotifications) && */
          ((header->snh_type == SCTP_DATA_ARRIVE)                                                      ||
          ((header->snh_type == SCTP_ASSOC_CHANGE)     && (notificationFlags & SCTP_RECVASSOCEVNT))    ||
          ((header->snh_type == SCTP_PEER_ADDR_CHANGE) && (notificationFlags & SCTP_RECVPADDREVNT))    ||
          ((header->snh_type == SCTP_REMOTE_ERROR)     && (notificationFlags & SCTP_RECVPEERERR))      ||
          ((header->snh_type == SCTP_SEND_FAILED)      && (notificationFlags & SCTP_RECVSENDFAILEVNT)) ||
          ((header->snh_type == SCTP_SHUTDOWN_EVENT)   && (notificationFlags & SCTP_RECVSHUTDOWNEVNT)))) {
puts("HAS_DATA!!");
          return(true);
      }
      notification = notification->NextNotification;
   }
   return(false);
}


}
