/*
 *  $Id: sctpsocketwrapper.cc,v 1.41 2002/01/07 15:22:43 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: SCTP Socket Wrapper
 *
 */


#include "tdsystem.h"
#include "tdmessage.h"
#include "sctpsocketwrapper.h"
#include "extsocketdescriptor.h"


#include <stdarg.h>

// Print note, if no SCTP is available in kernel.
// #define PRINT_NOSCTP_NOTE



#ifndef HAVE_KERNEL_SCTP


#include "tdsocket.h"
#include "internetaddress.h"
#include "sctpsocketmaster.h"
#include "sctpsocket.h"
#include "sctpassociation.h"


#include <sys/time.h>
#include <sys/uio.h>
#include <sys/socket.h>



// Set errno variable and return.
inline static int getErrnoResult(const int result)
{
   if(result < 0) {
      errno = -result;
      return(-1);
   }
   errno = 0;
   return(result);
}
#define errno_return(x) return(getErrnoResult(x))



using namespace Coral;




// ###### Constructor #######################################################
ExtSocketDescriptorMaster::ExtSocketDescriptorMaster()
{
   for(unsigned int i = 0;i < MaxSockets;i++) {
      Sockets[i].Type = ExtSocketDescriptor::ESDT_Invalid;
   }
   if(STDIN_FILENO < MaxSockets) {
      Sockets[STDIN_FILENO].Type                   = ExtSocketDescriptor::ESDT_System;
      Sockets[STDIN_FILENO].Socket.SystemSocketID  = STDIN_FILENO;
   }
   if(STDOUT_FILENO < MaxSockets) {
      Sockets[STDOUT_FILENO].Type                  = ExtSocketDescriptor::ESDT_System;
      Sockets[STDOUT_FILENO].Socket.SystemSocketID = STDOUT_FILENO;
   }
   if(STDERR_FILENO < MaxSockets) {
      Sockets[STDERR_FILENO].Type                  = ExtSocketDescriptor::ESDT_System;
      Sockets[STDERR_FILENO].Socket.SystemSocketID = STDERR_FILENO;
   }
}


// ###### Destructor ########################################################
ExtSocketDescriptorMaster::~ExtSocketDescriptorMaster()
{
   for(unsigned int i = 0;i < MaxSockets;i++) {
      if(Sockets[i].Type != ExtSocketDescriptor::ESDT_Invalid) {
         ext_close(i);
      }
   }
}


// ###### Get ExtSocketDescriptor for given ID #########################################
inline ExtSocketDescriptor* ExtSocketDescriptorMaster::getSocket(const int id)
{
   if((id < (int)MaxSockets) && (id >= 0)) {
      return(&Sockets[id]);
   }
   return(NULL);
}


// ###### Set ExtSocketDescriptor of given ID ##########################################
int ExtSocketDescriptorMaster::setSocket(const ExtSocketDescriptor& newSocket)
{
   for(int i = (int)(MaxSockets - 1);i >= 0;i--) {
      if(Sockets[i].Type == ExtSocketDescriptor::ESDT_Invalid) {
         Sockets[i] = newSocket;
         return(i);
      }
   }
   return(-ENOMEM);
}


// ###### Get paddrinfo #####################################################
static int getPathAddrInfo(ExtSocketDescriptor* tdSocket,
                           unsigned int         assocID,
                           const int            pathIndex,
                           sctp_paddrinfo&      paddrinfo)
{
   int             result = -EBADF;
   SCTP_PathStatus parameters;

   SCTPSocketMaster::MasterInstance.lock();

   if(tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
            getPathParameters(pathIndex,parameters)) {
            assocID = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID();
            result = 0;
         }
      }
   }
   else {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
            getPathParameters(assocID,pathIndex,parameters)) {
            result = 0;
         }
      }
   }

   if(result == 0) {
      SocketAddress* address = SocketAddress::createSocketAddress(0,
                                  (char*)&parameters.destinationAddress);
      if(address != NULL) {
         if(address->getSystemAddress((sockaddr*)&paddrinfo.spinfo_address,
                                      sizeof(paddrinfo.spinfo_address),
                                      tdSocket->Socket.SCTPSocketDesc.Domain) > 0) {
            paddrinfo.spinfo_state    = parameters.state;
            paddrinfo.spinfo_cwnd     = parameters.cwnd;
            paddrinfo.spinfo_srtt     = parameters.srtt;
            paddrinfo.spinfo_rto      = parameters.rto;
            paddrinfo.spinfo_assoc_id = assocID;
         }
         else {
            result = -EIO;
         }
         delete address;
      }
      else {
         result = -EIO;
      }
   }

   SCTPSocketMaster::MasterInstance.unlock();

   return(result);
}


// ###### Get association parameters ########################################
static int getAssocParams(ExtSocketDescriptor*     tdSocket,
                          unsigned int&            assocID,
                          SCTP_Association_Status& parameters)
{
  if(tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
               getAssociationParameters(parameters)) {
            assocID = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID();
            return(0);
         }
      }
   }
   else {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
               getAssociationParameters(assocID,parameters)) {
            return(0);
         }
      }
   }
   return(-1);
}


// ###### Set association parameters ########################################
static int setAssocParams(ExtSocketDescriptor*           tdSocket,
                          const unsigned int             assocID,
                          const SCTP_Association_Status& parameters)
{
  if(tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
               setAssociationParameters(parameters)) {
            return(0);
         }
      }
   }
   else {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
               setAssociationParameters(assocID,parameters)) {
            return(0);
         }
      }
   }
   return(-1);
}


// ###### Get association parameters ########################################
static int getAddressParams(ExtSocketDescriptor* tdSocket,
                            unsigned int&        assocID,
                            sockaddr*            sockadr,
                            const socklen_t      socklen,
                            SCTP_Path_Status&    parameters)
{
   SocketAddress* sa = SocketAddress::createSocketAddress(0,sockadr,socklen);
   if(sa == NULL) {
      return(-EINVAL);
   }
   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            sa->getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy).getData());
   delete sa;

   SCTPSocketMaster::MasterInstance.lock();

   int index  = 0;
   int result = -EINVAL;
   for(;;) {
      // ====== Get path status =============================================
      if(tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) {
          if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
             if(!tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                    getPathParameters(index,parameters)) {
                break;
             }
             assocID = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID();
          }
       }
      else {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
            if(!tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                   getPathParameters(assocID,index,parameters)) {
               break;
            }
         }
      }

      // ====== Check, if path has searched address =========================
      if(!(strcasecmp((char*)&address,(char*)&parameters.destinationAddress))) {
         result = 0;
         break;
      }
      index++;
   }

   SCTPSocketMaster::MasterInstance.unlock();
   return(result);
}


// ###### Set association parameters ########################################
static int setAddressParams(ExtSocketDescriptor*    tdSocket,
                            unsigned int            assocID,
                            sockaddr*               sockadr,
                            const socklen_t         socklen,
                            const SCTP_Path_Status& newParameters)
{
   SocketAddress* sa = SocketAddress::createSocketAddress(0,sockadr,socklen);
   if(sa == NULL) {
      return(-EINVAL);
   }
   unsigned char address[SCTP_MAX_IP_LEN];
   snprintf((char*)&address,sizeof(address),"%s",
            sa->getAddressString(SocketAddress::PF_Address|SocketAddress::PF_HidePort|SocketAddress::PF_Legacy).getData());
   delete sa;
   int index = 0;
   for(;;) {
      SCTP_Path_Status parameters;

      // ====== Get path status =============================================
      if(tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) {
          if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
             if(!tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                    getPathParameters(index,parameters)) {
                return(-EIO);
             }
             assocID = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID();
          }
       }
      else {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
            if(!tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                   getPathParameters(assocID,index,parameters)) {
               return(-EIO);
            }
         }
      }

      // ====== Check, if path has searched address =========================
      if(!(strcasecmp((char*)&address,(char*)&parameters.destinationAddress))) {
          if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
             if(!tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                    setPathParameters(index,newParameters)) {
                return(-EIO);
             }
          }
          else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
             if(!tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                    setPathParameters(assocID,index,newParameters)) {
                return(-EIO);
             }
          }
      }
      index++;
   }
   return(-EINVAL);
}


// ###### socket() wrapper ##################################################
int ext_socket(int domain, int type, int protocol)
{
   ExtSocketDescriptor tdSocket;
   if(protocol == IPPROTO_SCTP) {
      // ====== Check parameters ============================================
      cardinal flags;
      if(type == SOCK_STREAM) {
         tdSocket.Socket.SCTPSocketDesc.ConnectionOriented = true;
         flags = 0;
      }
      else if((type == SOCK_DGRAM) || (type == SOCK_SEQPACKET)) {
         tdSocket.Socket.SCTPSocketDesc.ConnectionOriented = false;
         flags = SCTPSocket::SSF_GlobalQueue|SCTPSocket::SSF_AutoConnect;
      }
      else {
         errno_return(-EINVAL);
      }

      // ====== Create SCTP socket ==========================================
      tdSocket.Type = ExtSocketDescriptor::ESDT_SCTP;
      tdSocket.Socket.SCTPSocketDesc.Domain             = domain;
      tdSocket.Socket.SCTPSocketDesc.Type               = type;
      tdSocket.Socket.SCTPSocketDesc.Flags              = 0;
      tdSocket.Socket.SCTPSocketDesc.Parent             = 0;
      tdSocket.Socket.SCTPSocketDesc.InitMsg.sinit_num_ostreams   = 1;
      tdSocket.Socket.SCTPSocketDesc.InitMsg.sinit_max_instreams  = 1;
      tdSocket.Socket.SCTPSocketDesc.InitMsg.sinit_max_attempts   = 2;
      tdSocket.Socket.SCTPSocketDesc.InitMsg.sinit_max_init_timeo = 30;
      tdSocket.Socket.SCTPSocketDesc.Linger.l_onoff  = 1;
      tdSocket.Socket.SCTPSocketDesc.Linger.l_linger = 10;
      tdSocket.Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
      tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr = new SCTPSocket(flags);
      if(tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr == NULL) {
         errno_return(-ENOMEM);
      }

/*

   This has been turned off since it makes problems using ext_select()!

      tdSocket.Socket.SCTPSocketDesc.ConnectionOriented == false) {
         // Default notifications for UDP-like socket
         tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr->setNotificationFlags(
            SCTP_RECVDATAIOEVNT|SCTP_RECVASSOCEVNT);
      }
*/

      int result = ExtSocketDescriptorMaster::setSocket(tdSocket);
      if(result < 0) {
         delete tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr;
         tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr = NULL;
      }
      errno_return(result);
   }
   else {
      // ====== Create system socket ========================================
      tdSocket.Type = ExtSocketDescriptor::ESDT_System;
      tdSocket.Socket.SystemSocketID = socket(domain,type,protocol);
      if(tdSocket.Socket.SystemSocketID >= 0) {
         int result = ExtSocketDescriptorMaster::setSocket(tdSocket);
         if(result < 0) {
            delete tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr;
            tdSocket.Socket.SCTPSocketDesc.SCTPSocketPtr = NULL;
         }
         errno_return(result);
      }
      errno_return(tdSocket.Socket.SystemSocketID);
   }
}


// ###### open() wrapper ####################################################
int ext_open(const char* pathname, int flags, mode_t mode)
{
   ExtSocketDescriptor tdSocket;
   tdSocket.Type = ExtSocketDescriptor::ESDT_System;
   tdSocket.Socket.SystemSocketID = open(pathname,flags,mode);
   if(tdSocket.Socket.SystemSocketID >= 0) {
      int result = ExtSocketDescriptorMaster::setSocket(tdSocket);
      if(result < 0) {
         close(tdSocket.Socket.SystemSocketID);
      }
      errno_return(result);
   }
   errno_return(tdSocket.Socket.SystemSocketID);
}


// ###### creat() wrapper ###################################################
int ext_creat(const char* pathname, mode_t mode)
{
   ExtSocketDescriptor tdSocket;
   tdSocket.Type = ExtSocketDescriptor::ESDT_System;
   tdSocket.Socket.SystemSocketID = creat(pathname,mode);
   if(tdSocket.Socket.SystemSocketID >= 0) {
      int result = ExtSocketDescriptorMaster::setSocket(tdSocket);
      if(result < 0) {
         close(tdSocket.Socket.SystemSocketID);
      }
      errno_return(result);
   }
   errno_return(tdSocket.Socket.SystemSocketID);
}


// ###### Delete all childs of a parent socket ##############################
static void ext_closeChilds(const int parent)
{
   for(unsigned int i = 1;i <= ExtSocketDescriptorMaster::MaxSockets;i++) {
      ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(i);
      if((tdSocket != NULL) &&
         (tdSocket->Type == ExtSocketDescriptor::ESDT_SCTP) &&
         (tdSocket->Socket.SCTPSocketDesc.Parent == parent)) {
         ext_close(i);
      }
   }
}


// ###### close() wrapper ###################################################
int ext_close(int sockfd)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            ext_closeChilds(sockfd);
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               if(tdSocket->Socket.SCTPSocketDesc.Linger.l_onoff == 1) {
                  if(tdSocket->Socket.SCTPSocketDesc.Linger.l_linger > 0) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->shutdown();
                  }
                  else {
                     tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->abort();
                  }
               }
               delete tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
            }
            if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
               if(tdSocket->Socket.SCTPSocketDesc.Linger.l_onoff == 1) {
                  if(tdSocket->Socket.SCTPSocketDesc.Linger.l_linger > 0) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->unbind(false);
                  }
                  else {
                     tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->unbind(true);
                  }
               }
               delete tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr;
               tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr = NULL;
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            close(tdSocket->Socket.SystemSocketID);
            tdSocket->Socket.SystemSocketID = 0;
          break;
         default:
            errno_return(-ENXIO);
          break;
      }
      tdSocket->Type = ExtSocketDescriptor::ESDT_Invalid;
      errno_return(0);
   }
   errno_return(-EBADF);
}


// ###### bind() wrapper ####################################################
int ext_bind(int sockfd, struct sockaddr* my_addr, socklen_t addrlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
             return(ext_bindx(sockfd,(sockaddr_storage*)my_addr,1,SCTP_BINDX_ADD_ADDR));
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(bind(tdSocket->Socket.SystemSocketID,(sockaddr*)my_addr,addrlen));
          break;
         default:
            errno_return(-ENXIO);
          break;
      }
   }
   errno_return(-EBADF);
}


// ###### xbind() wrapper ###################################################
int ext_bindx(int                      sockfd,
              struct sockaddr_storage* addrs,
              int                      addrcnt,
              int                      flags)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP: {
               // ====== Check parameters and descriptor ====================
               if((addrcnt < 1) || (addrcnt > SCTP_MAX_NUM_ADDRESSES)) {
                  errno_return(-EINVAL);
               }
               else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr == NULL) {
                  errno_return(-EBADF);
               }

               // ====== Get addresses ======================================
               SocketAddress* addressArray[addrcnt + 1];
               sockaddr_storage* a = addrs;
               for(int i = 0;i < addrcnt;i++) {
                  addressArray[i] = SocketAddress::createSocketAddress(
                                       0,(sockaddr*)a,sizeof(sockaddr_storage));
                  if(addressArray[i] == NULL) {
                     for(int j = 0;j < i;j++) {
                        delete addressArray[j];
                     }
                     errno_return(-ENOMEM);
                  }
                  a = (sockaddr_storage*)((long)a + (long)sizeof(sockaddr_storage));
               }
               addressArray[addrcnt] = NULL;

               // ====== Bind socket ========================================
               int result = -EINVAL;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getID() == 0) {
                  if(flags == SCTP_BINDX_ADD_ADDR) {
                     result = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->bind(
                                 addressArray[0]->getPort(),
                                 tdSocket->Socket.SCTPSocketDesc.InitMsg.sinit_max_instreams,
                                 tdSocket->Socket.SCTPSocketDesc.InitMsg.sinit_num_ostreams,
                                 (const SocketAddress**)&addressArray);
                  }
               }
               else {
                  if(flags == SCTP_BINDX_ADD_ADDR) {
                     result = 0;
                     if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                        for(int i = 0;i < addrcnt;i++) {
                           const bool ok = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                                              addAddress(*(addressArray[i]));
                           if(!ok) {
                              result = -EINVAL;
                              break;
                           }
                        }
                     }
                     else {
                        for(int i = 0;i < addrcnt;i++) {
                           const bool ok = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                                              addAddress(0,*(addressArray[i]));
                           if(!ok) {
                              result = -EINVAL;
                              break;
                           }
                        }
                     }
                  }
                  else if(flags == SCTP_BINDX_REM_ADDR) {
                     result = 0;
                     if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                        for(int i = 0;i < addrcnt;i++) {
                           const bool ok = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                                              deleteAddress(*(addressArray[i]));
                           if(!ok) {
                              result = -EINVAL;
                              break;
                           }
                        }
                     }
                     else {
                        for(int i = 0;i < addrcnt;i++) {
                           const bool ok = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                                              deleteAddress(0,*(addressArray[i]));
                           if(!ok) {
                              result = -EINVAL;
                              break;
                           }
                        }
                     }
                  }
               }

               // ====== Remove addresses ===================================
               for(int i = 0;i < addrcnt;i++) {
                  delete addressArray[i];
               }
               errno_return(result);
            }
           break;
         case ExtSocketDescriptor::ESDT_System:
            return(bind(tdSocket->Socket.SystemSocketID,(sockaddr*)addrs,sizeof(sockaddr_storage)));
           break;
         default:
            errno_return(-ENXIO);
          break;
      }
   }
   errno_return(-EBADF);
}


// ###### listen() wrapper ####################################################
int ext_listen(int sockfd, int backlog)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
               tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->listen(backlog);
               errno_return(0);
            }
            errno_return(-EOPNOTSUPP);
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(listen(tdSocket->Socket.SystemSocketID,backlog));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### accept() wrapper ####################################################
int ext_accept(int sockfd, struct sockaddr* addr, socklen_t* addrlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                  SocketAddress** remoteAddressArray = NULL;
                  SCTPAssociation* association =
                     tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->accept(
                        &remoteAddressArray,
                        !(tdSocket->Socket.SCTPSocketDesc.Flags & O_NONBLOCK));
                  if(association != NULL) {
                     if((remoteAddressArray != NULL)    &&
                        (remoteAddressArray[0] != NULL) &&
                        (addr != NULL) && (addrlen != NULL)) {
                        *addrlen = remoteAddressArray[0]->getSystemAddress(
                                      addr, *addrlen,
                                      tdSocket->Socket.SCTPSocketDesc.Domain);
                     }
                     else {
                        addrlen = 0;
                     }

                     ExtSocketDescriptor newExtSocketDescriptor = *tdSocket;
                     newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPSocketPtr      = NULL;
                     newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr = association;
                     newExtSocketDescriptor.Socket.SCTPSocketDesc.Parent             = sockfd;
                     const int newFD = ExtSocketDescriptorMaster::setSocket(newExtSocketDescriptor);
                     SocketAddress::deleteAddressList(remoteAddressArray);
                     if(newFD < 0) {
                        delete newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr;
                        newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
                     }
                     errno_return(newFD);
                  }
                  else {
                     errno_return(0);
                  }
               }
               errno_return(-EOPNOTSUPP);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            {
               ExtSocketDescriptor newExtSocketDescriptor = *tdSocket;
               newExtSocketDescriptor.Socket.SystemSocketID = accept(tdSocket->Socket.SystemSocketID,addr,addrlen);
               if(newExtSocketDescriptor.Socket.SystemSocketID >= 0) {
                  int result = ExtSocketDescriptorMaster::setSocket(newExtSocketDescriptor);
                  if(result < 0) {
                     close(newExtSocketDescriptor.Socket.SystemSocketID);
                  }
                  errno_return(result);
               }
               errno_return(newExtSocketDescriptor.Socket.SystemSocketID);
            }
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### getsockname() wrapper #############################################
int ext_getsockname(int sockfd, struct sockaddr* name, socklen_t* namelen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               int result = -ENXIO;
               SocketAddress** localAddressArray = NULL;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getLocalAddresses(localAddressArray);
               }
               else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getLocalAddresses(localAddressArray);
               }
               else {
                  result = -EBADF;
               }

               if((localAddressArray != NULL)    &&
                  (localAddressArray[0] != NULL) &&
                  (name != NULL) && (namelen != NULL)) {
                  if(localAddressArray[0]->getSystemAddress(name,*namelen,
                                              tdSocket->Socket.SCTPSocketDesc.Domain)) {
                     result = 0;
                  }
                  else {
                     result = -ENAMETOOLONG;
                  }
               }

               SocketAddress::deleteAddressList(localAddressArray);
               errno_return(result);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(getsockname(tdSocket->Socket.SystemSocketID,name,namelen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### getpeername() wrapper #############################################
int ext_getpeername(int sockfd, struct sockaddr* name, socklen_t* namelen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               int result = -ENXIO;
               SocketAddress** remoteAddressArray = NULL;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getRemoteAddresses(remoteAddressArray);
               }
               else {
                  result = -EBADF;
               }

               if((remoteAddressArray != NULL)    &&
                  (remoteAddressArray[0] != NULL) &&
                  (name != NULL) && (namelen != NULL)) {
                  if(remoteAddressArray[0]->getSystemAddress(name,*namelen,tdSocket->Socket.SCTPSocketDesc.Domain) > 0) {
                     result = 0;
                  }
                  else {
                     result = -ENAMETOOLONG;
                  }
               }

               SocketAddress::deleteAddressList(remoteAddressArray);
               errno_return(result);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(getpeername(tdSocket->Socket.SystemSocketID,name,namelen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### fcntl() wrapper ###################################################
int ext_fcntl(int sockfd, int cmd, ...)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      va_list va;
      unsigned long int arg;
      va_start (va, cmd);
      arg = va_arg (va, unsigned long int);
      va_end (va);
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            switch(cmd) {
               case F_GETFL:
                  errno_return(tdSocket->Socket.SCTPSocketDesc.Flags);
                 break;
               case F_SETFL:
                   tdSocket->Socket.SCTPSocketDesc.Flags = arg;
                   errno_return(0);
                 break;
               default:
                  errno_return(-EOPNOTSUPP);
                break;
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(fcntl(tdSocket->Socket.SystemSocketID,cmd,arg));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### ioctl() wrapper ###################################################
int ext_ioctl(int sockfd, int request, const void* argp)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            errno_return(-EOPNOTSUPP);
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(ioctl(tdSocket->Socket.SystemSocketID,request,argp));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### Get option ########################################################
static int getFlag(ExtSocketDescriptor* tdSocket,
                   void* optval, socklen_t* optlen, const int flag)
{
   if((*optlen < (int)sizeof(int)) || (optval == NULL)) {
      errno_return(-EINVAL);
   }
   *optlen = sizeof(int);
   if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
      *((int*)optval) = (tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getNotificationFlags() & flag);
      errno_return(0);
   }
   else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
      *((int*)optval) = (tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getNotificationFlags() & flag);
      errno_return(0);
   }
   errno_return(-EBADF);
}


// ###### Get association status ############################################
static int getAssocStatus(ExtSocketDescriptor* tdSocket,
                          void* optval, socklen_t* optlen)
{
   if((*optlen < (int)sizeof(sctp_status)) || (optval == NULL)) {
      errno_return(-EINVAL);
   }
   sctp_status*            status  = (sctp_status*)optval;
   unsigned int            assocID = status->sstat_assoc_id;
   SCTP_Association_Status parameters;

   SCTPSocketMaster::MasterInstance.lock();
   int result = getAssocParams(tdSocket,assocID,parameters);
   if(result == 0) {
      status->sstat_state     = parameters.state;
      status->sstat_rwnd      = parameters.currentReceiverWindowSize;
      status->sstat_unackdata = parameters.noOfChunksInRetransmissionQueue;
      status->sstat_penddata  = parameters.noOfChunksInReceptionQueue;
      status->sstat_assoc_id  = assocID;
      result = getPathAddrInfo(tdSocket,assocID,-1,status->sstat_primary);
      *optlen = sizeof(sctp_status);
   }
   SCTPSocketMaster::MasterInstance.unlock();
   errno_return(result);
}


// ###### Get RTO info ######################################################
static int getRTOInfo(ExtSocketDescriptor* tdSocket,
                      void* optval, socklen_t* optlen)
{
   if((optval == NULL) || (*optlen < sizeof(sctp_rtoinfo))) {
      errno_return(-EINVAL);
   }
   sctp_rtoinfo*           rtoinfo = (sctp_rtoinfo*)optval;
   unsigned int            assocID = rtoinfo->srto_assoc_id;
   SCTP_Association_Status parameters;

   int result = getAssocParams(tdSocket,assocID,parameters);
   if(result == 0) {
      rtoinfo->srto_initial  = parameters.rtoInitial;
      rtoinfo->srto_max      = parameters.rtoMax;
      rtoinfo->srto_min      = parameters.rtoMin;
      rtoinfo->srto_assoc_id = assocID;
      *optlen = sizeof(sctp_rtoinfo);
   }
   errno_return(result);
}


// ###### Get RTO info ######################################################
static int getAssocMaxRxt(ExtSocketDescriptor* tdSocket,
                          void* optval, socklen_t* optlen)
{
   if((optval == NULL) || (*optlen < sizeof(sctp_assocparams))) {
      errno_return(-EINVAL);
   }
   sctp_assocparams*       assocparams = (sctp_assocparams*)optval;
   unsigned int            assocID     = assocparams->sasoc_assoc_id;
   SCTP_Association_Status parameters;

   int result = getAssocParams(tdSocket,assocID,parameters);
   if(result == 0) {
      assocparams->sasoc_assocmaxrxt = parameters.assocMaxRetransmits;
      assocparams->sasoc_assoc_id    = assocID;
      *optlen = sizeof(sctp_assocparams);
   }
   errno_return(result);
}


// ###### Get address information ###########################################
static int getPeerAddressInfo(ExtSocketDescriptor* tdSocket,
                              void* optval, socklen_t* optlen)
{
   if((optval == NULL) || (*optlen < sizeof(sctp_paddrinfo))) {
      errno_return(-EINVAL);
   }
   int              result    = -EBADF;
   sctp_paddrinfo*  paddrinfo = (sctp_paddrinfo*)optval;
   unsigned int     assocID   = paddrinfo->spinfo_assoc_id;
   SCTP_Path_Status parameters;

   result = getAddressParams(tdSocket,assocID,
                             (sockaddr*)&paddrinfo->spinfo_address,
                             sizeof(paddrinfo->spinfo_address),
                             parameters);
   if(result == 0) {
      paddrinfo->spinfo_state    = parameters.state;
      paddrinfo->spinfo_cwnd     = parameters.cwnd;
      paddrinfo->spinfo_srtt     = parameters.srtt;
      paddrinfo->spinfo_rto      = parameters.rto;
      paddrinfo->spinfo_assoc_id = assocID;
      *optlen = sizeof(sctp_paddrinfo);
   }
   errno_return(result);
}


// ###### Get address parameters ############################################
static int getPeerAddressParams(ExtSocketDescriptor* tdSocket,
                                void* optval, socklen_t* optlen)
{
   if((optval == NULL) || (*optlen < sizeof(sctp_paddrparams))) {
      errno_return(-EINVAL);
   }
   int               result      = -EBADF;
   sctp_paddrparams* paddrparams = (sctp_paddrparams*)optval;
   unsigned int      assocID     = paddrparams->spp_assoc_id;
   SCTP_Path_Status  parameters;

   result = getAddressParams(tdSocket,assocID,
                             (sockaddr*)&paddrparams->spp_address,
                             sizeof(paddrparams->spp_address),
                             parameters);
   if(result == 0) {
      paddrparams->spp_hbinterval = parameters.heartbeatIntervall;
      paddrparams->spp_pathmaxrxt = 0;
      paddrparams->spp_assoc_id = assocID;
      *optlen = sizeof(sctp_paddrparams);
   }
   errno_return(result);
}


// ###### getsockopt() wrapper ##############################################
int ext_getsockopt(int sockfd, int level, int optname, void* optval, socklen_t* optlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               switch(level) {
#if (SYSTEM == OS_Linux)
                  // ====== IPv6 ============================================
                  case SOL_IPV6:
                      switch(optname) {
                         case IPV6_FLOWINFO:
                            errno_return(0);
                          break;
                         case IPV6_FLOWINFO_SEND:
                            errno_return(0);
                          break;
                         default:
                            errno_return(-EOPNOTSUPP);
                          break;
                      }
                   break;

                  // ====== IPv4 ============================================
                  case SOL_IP:
                      switch(optname) {
                         case IP_TOS:
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               errno_return(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getTrafficClass());
                            }
                            errno_return(0);
                          break;
                         case IP_RECVTOS:
                            errno_return(0);
                          break;
                         default:
                            errno_return(-EOPNOTSUPP);
                          break;
                      }
                   break;
#endif

                  // ====== SCTP ============================================
                  case IPPROTO_SCTP:
                      switch(optname) {
                         case SCTP_STATUS:
                            return(getAssocStatus(tdSocket,optval,optlen));
                          break;
                         case SCTP_INITMSG: {
                               if((optval == NULL) || (*optlen < sizeof(sctp_initmsg))) {
                                  errno_return(-EINVAL);
                               }
                               sctp_initmsg* initmsg = (sctp_initmsg*)optval;
                               *initmsg = tdSocket->Socket.SCTPSocketDesc.InitMsg;
                               *optlen = sizeof(sctp_initmsg);
                               errno_return(0);
                            }
                          break;
                         case SCTP_GET_PEER_ADDR_INFO:
                             return(getPeerAddressInfo(tdSocket,optval,optlen));
                          break;
                         case SCTP_GET_PEER_ADDR_PARAMS:
                             return(getPeerAddressParams(tdSocket,optval,optlen));
                          break;
                         case SCTP_RTOINFO: {
                               return(getRTOInfo(tdSocket,optval,optlen));
                            }
                          break;
                         case SCTP_ASSOCRXTINFO:
                            return(getAssocMaxRxt(tdSocket,optval,optlen));
                          break;
                         case SCTP_RECVDATAIOEVNT:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVDATAIOEVNT));
                          break;
                         case SCTP_RECVASSOCEVNT:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVASSOCEVNT));
                          break;
                         case SCTP_RECVPADDREVNT:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVPADDREVNT));
                          break;
                         case SCTP_RECVSENDFAILEVNT:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVSENDFAILEVNT));
                          break;
                         case SCTP_RECVPEERERR:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVPEERERR));
                          break;
                         case SCTP_RECVSHUTDOWNEVNT:
                            return(getFlag(tdSocket,optval,optlen,SCTP_RECVSHUTDOWNEVNT));
                          break;
                         case SCTP_AUTOCLOSE:
                            if((optval == NULL) || (*optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                               *((unsigned int*)optval) =
                                  (unsigned int)(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getAutoClose() /
                                                    1000000);
                               *optlen = sizeof(unsigned int);
                               errno_return(0);
                            }
                            errno_return(-EBADF);
                          break;
                      }
                   break;

                  // ====== Socket ==========================================
                  case SOL_SOCKET:
                      switch(optname) {
                         case SO_SNDBUF:
                            if((optval == NULL) || (*optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               *((int*)optval) = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getSendBuffer();
                               *optlen = sizeof(unsigned int);
                               errno_return((*((int*)optval) >= 0) ? 0 : -EIO);
                            }
                            errno_return(-EBADF);
                          break;
                         case SO_RCVBUF:
                            if((optval == NULL) || (*optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               *((int*)optval) = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getReceiveBuffer();
                               *optlen = sizeof(unsigned int);
                               errno_return((*((int*)optval) >= 0) ? 0 : -EIO);
                            }
                            errno_return(-EBADF);
                          break;
                         case SO_LINGER:
                            if((optval == NULL) || (*optlen < sizeof(linger))) {
                               errno_return(-EINVAL);
                            }
                            *((linger*)optval) = tdSocket->Socket.SCTPSocketDesc.Linger;
                            *optlen = sizeof(linger);
                            errno_return(0);
                          break;
                      }
                   break;

                  // ====== Default =========================================
                  default:
                     errno_return(-EOPNOTSUPP);
                   break;
               }
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(getsockopt(tdSocket->Socket.SystemSocketID,level,optname,optval,optlen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### Set option ########################################################
static int setFlag(ExtSocketDescriptor* tdSocket,
                   const void* optval, const socklen_t optlen, const int flag)
{
   if((optlen != sizeof(int)) || (optval == NULL)) {
      errno_return(-EINVAL);
   }
   const bool on = (*((int*)optval) != 0);
   if(on) {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
         tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->setNotificationFlags(
            tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getNotificationFlags() |
            flag);
         errno_return(0);
      }
      else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
         tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setNotificationFlags(
            tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getNotificationFlags() |
            flag);
         errno_return(0);
      }
   }
   else {
      if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
         tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->setNotificationFlags(
            tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getNotificationFlags() &
            ~flag);
         errno_return(0);
      }
      else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
         tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setNotificationFlags(
            tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getNotificationFlags() &
            ~flag);
         errno_return(0);
      }
   }
   errno_return(-EBADF);
}


// ###### Set RTO info ######################################################
static int setRTOInfo(ExtSocketDescriptor* tdSocket,
                      const void* optval, const socklen_t optlen)
{
   if((optval == NULL) || (optlen < sizeof(sctp_rtoinfo))) {
      errno_return(-EINVAL);
   }
   sctp_rtoinfo*           rtoinfo = (sctp_rtoinfo*)optval;
   unsigned int            assocID = rtoinfo->srto_assoc_id;
   SCTP_Association_Status parameters;

   SCTPSocketMaster::MasterInstance.lock();
   int result = getAssocParams(tdSocket,assocID,parameters);
   if(result == 0) {
      if(rtoinfo->srto_initial != 0) {
         parameters.rtoInitial = rtoinfo->srto_initial;
      }
      if(rtoinfo->srto_min != 0) {
         parameters.rtoMin = rtoinfo->srto_min;
      }
      if(rtoinfo->srto_max != 0) {
         parameters.rtoMax = rtoinfo->srto_max;
      }
      result = setAssocParams(tdSocket,assocID,parameters);
   }
   SCTPSocketMaster::MasterInstance.unlock();
   errno_return(result);
}


// ###### Set association RTX info ##########################################
static int setAssocMaxRxt(ExtSocketDescriptor* tdSocket,
                          const void* optval, const socklen_t optlen)
{
   if((optval == NULL) || (optlen < sizeof(sctp_assocparams))) {
      errno_return(-EINVAL);
   }
   sctp_assocparams*       assocparams = (sctp_assocparams*)optval;
   unsigned int            assocID     = assocparams->sasoc_assoc_id;
   SCTP_Association_Status parameters;

   SCTPSocketMaster::MasterInstance.lock();
   int result = getAssocParams(tdSocket,assocID,parameters);
   if(result == 0) {
      parameters.assocMaxRetransmits = assocparams->sasoc_assocmaxrxt;
      result = setAssocParams(tdSocket,assocID,parameters);
   }
   SCTPSocketMaster::MasterInstance.unlock();
   errno_return(result);
}


// ###### Set primary address ###############################################
static int setPrimaryAddr(ExtSocketDescriptor* tdSocket,
                          const void* optval, const socklen_t optlen)
{
   if((optval == NULL) || (optlen < sizeof(sctp_setprim))) {
      errno_return(-EINVAL);
   }
   sctp_setprim* setprim = (sctp_setprim*)optval;
   SocketAddress* address = SocketAddress::createSocketAddress(0,
                               (sockaddr*)&setprim->ssp_addr,
                               sizeof(sockaddr_storage));
   if(address == NULL) {
      errno_return(-EINVAL);
   }

   int result = -EBADF;
   if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
      result = (tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                   setPrimary(*address) == true) ? 0 : -EIO;
   }
   else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
      result = (tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                   setPrimary(setprim->ssp_assoc_id,*address) == true) ? 0 : -EIO;
   }

   delete address;
   errno_return(result);
}


// ###### Set primary address ###############################################
static int setPeerPrimaryAddr(ExtSocketDescriptor* tdSocket,
                              const void* optval, const socklen_t optlen)
{
   if((optval == NULL) || (optlen < sizeof(sctp_setpeerprim))) {
      errno_return(-EINVAL);
   }
   sctp_setpeerprim* setpeerprim = (sctp_setpeerprim*)optval;
   SocketAddress* address = SocketAddress::createSocketAddress(0,
                               (sockaddr*)&setpeerprim->sspp_addr,
                               sizeof(sockaddr_storage));
   if(address == NULL) {
      errno_return(-EINVAL);
   }

   int result = -EBADF;
   if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
      result = (tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->
                   setPrimary(*address) == true) ? 0 : -EIO;
   }
   else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
      result = (tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->
                   setPrimary(setpeerprim->sspp_assoc_id,*address) == true) ? 0 : -EIO;
   }

   delete address;
   errno_return(result);
}


// ###### Get address parameters ############################################
static int setPeerAddressParams(ExtSocketDescriptor* tdSocket,
                                const void* optval, const socklen_t optlen)
{
   if((optval == NULL) || (optlen < sizeof(sctp_paddrparams))) {
      errno_return(-EINVAL);
   }
   int               result      = -EBADF;
   sctp_paddrparams* paddrparams = (sctp_paddrparams*)optval;
   unsigned int      assocID     = paddrparams->spp_assoc_id;
   SCTP_Path_Status  parameters;

   SCTPSocketMaster::MasterInstance.lock();
   result = getAddressParams(tdSocket,assocID,
                             (sockaddr*)&paddrparams->spp_address,
                             sizeof(paddrparams->spp_address),
                             parameters);
   if(result == 0) {
      parameters.heartbeatIntervall = paddrparams->spp_hbinterval;
      result = setAddressParams(tdSocket,assocID,
                                (sockaddr*)&paddrparams->spp_address,
                                sizeof(paddrparams->spp_address),
                                parameters);
   }
   SCTPSocketMaster::MasterInstance.unlock();

   errno_return(result);
}


// ###### setsockopt() wrapper ##############################################
int ext_setsockopt(int sockfd, int level, int optname, const void* optval, socklen_t optlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               switch(level) {
#if (SYSTEM == OS_Linux)
                  // ====== IPv6 ============================================
                  case SOL_IPV6:
                      switch(optname) {
                         case IPV6_FLOWINFO:
                            errno_return(0);
                          break;
                         case IPV6_FLOWINFO_SEND:
                            errno_return(0);
                          break;
                         default:
                            errno_return(-EOPNOTSUPP);
                          break;
                      }
                   break;

                  // ====== IPv4 ============================================
                  case SOL_IP:
                      switch(optname) {
                         case IP_TOS:
                            if((optval == NULL) || (optlen < sizeof(int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->setTrafficClass(*((int*)optval)) == true) ? 0 : -EIO);
                            }
                            else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setTrafficClass(*((int*)optval)) == true) ? 0 : -EIO);
                            }
                            errno_return(-EOPNOTSUPP);
                          break;
                         case IP_RECVTOS:
                            errno_return(0);
                          break;
                         default:
                            errno_return(-EOPNOTSUPP);
                          break;
                      }
                   break;
#endif

                  // ====== SCTP ============================================
                  case IPPROTO_SCTP:
                      switch(optname) {
                         case SCTP_INITMSG: {
                               if((optval == NULL) || (optlen < sizeof(sctp_initmsg))) {
                                  errno_return(-EINVAL);
                               }
                               sctp_initmsg* initmsg = (sctp_initmsg*)optval;
                               tdSocket->Socket.SCTPSocketDesc.InitMsg = *initmsg;
                               errno_return(0);
                            }
                          break;
                         case SCTP_SET_PRIMARY_ADDR:
                            return(setPrimaryAddr(tdSocket,optval,optlen));
                          break;
                         case SCTP_SET_PEER_PRIMARY_ADDR:
                            return(setPeerPrimaryAddr(tdSocket,optval,optlen));
                          break;
                         case SCTP_SET_PEER_ADDR_PARAMS:
                             return(setPeerAddressParams(tdSocket,optval,optlen));
                          break;
                         case SCTP_RTOINFO:
                            return(setRTOInfo(tdSocket,optval,optlen));
                          break;
                         case SCTP_ASSOCRXTINFO:
                            return(setAssocMaxRxt(tdSocket,optval,optlen));
                          break;
                         case SCTP_RECVDATAIOEVNT:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVDATAIOEVNT));
                          break;
                         case SCTP_RECVASSOCEVNT:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVASSOCEVNT));
                          break;
                         case SCTP_RECVPADDREVNT:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVPADDREVNT));
                          break;
                         case SCTP_RECVSENDFAILEVNT:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVSENDFAILEVNT));
                          break;
                         case SCTP_RECVPEERERR:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVPEERERR));
                          break;
                         case SCTP_RECVSHUTDOWNEVNT:
                            return(setFlag(tdSocket,optval,optlen,SCTP_RECVSHUTDOWNEVNT));
                          break;
                         case SCTP_AUTOCLOSE:
                            if((optval == NULL) || (optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr == NULL) {
                               errno_return(-EBADF);
                            }
                            tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setAutoClose(
                               (card64)1000000 * (card64)*((unsigned int*)optval));
                            errno_return(0);
                          break;
                         case SO_SNDBUF:
                            if((optval == NULL) || (optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->setSendBuffer(*((unsigned int*)optval)) == true) ? 0 : -EIO);
                            }
                            else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setSendBuffer(*((unsigned int*)optval)) == true) ? 0 : -EIO);
                            }
                            errno_return(-EBADF);
                          break;
                       }
                     break;

                  // ====== Socket ==========================================
                  case SOL_SOCKET:
                      switch(optname) {
                         case SO_RCVBUF:
                            if((optval == NULL) || (optlen < sizeof(unsigned int))) {
                               errno_return(-EINVAL);
                            }
                            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->setSendBuffer(*((unsigned int*)optval)) == true) ? 0 : -EIO);
                            }
                            else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                               errno_return((tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->setReceiveBuffer(*((unsigned int*)optval)) == true) ? 0 : -EIO);
                            }
                            errno_return(-EBADF);
                          break;
                         case SO_LINGER:
                            if((optval == NULL) || (optlen < sizeof(linger))) {
                               errno_return(-EINVAL);
                            }
                            if( (((linger*)optval)->l_linger < 0) ||
                                (((linger*)optval)->l_onoff < 0)  ||
                                (((linger*)optval)->l_onoff > 1) ) {
                               errno_return(-EINVAL);
                            }
                            tdSocket->Socket.SCTPSocketDesc.Linger = *((linger*)optval);
                            errno_return(0);
                          break;
                      }
                   break;

                  // ====== Default =========================================
                  default:
                     errno_return(-EOPNOTSUPP);
                   break;
               }
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(setsockopt(tdSocket->Socket.SystemSocketID,level,optname,optval,optlen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### shutdown() wrapper ################################################
int ext_shutdown(int sockfd, int how)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->shutdown();
               errno_return(0);
            }
            errno_return(-EOPNOTSUPP);
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(shutdown(tdSocket->Socket.SystemSocketID,how));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### connect() wrapper #################################################
int ext_connect(int sockfd, const struct sockaddr* serv_addr, socklen_t addrlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  errno_return(-EISCONN);
               }
               if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                  SocketAddress* address = SocketAddress::createSocketAddress(
                                              0,(sockaddr*)serv_addr,addrlen);
                  if(address != NULL) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr =
                        tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->associate(
                           tdSocket->Socket.SCTPSocketDesc.InitMsg.sinit_num_ostreams,
                           *address,
                           !(tdSocket->Socket.SCTPSocketDesc.Flags & O_NONBLOCK));
                     delete address;
                     if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr == NULL) {
                        errno_return(-EIO);
                     }
                     errno_return(0);
                  }
                  else {
                     errno_return(-EADDRNOTAVAIL);
                  }
               }
               errno_return(-EBADF);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(connect(tdSocket->Socket.SystemSocketID,serv_addr,addrlen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### recv() wrapper ####################################################
int ext_recv(int sockfd, void* buf, size_t len, int flags)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               socklen_t fromlen = 0;
               return(ext_recvfrom(sockfd,buf,len,flags,NULL,&fromlen));
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(recv(tdSocket->Socket.SystemSocketID,buf,len,flags));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### recvfrom() wrapper ################################################
int ext_recvfrom(int sockfd, void* buf, size_t len, int flags,
                struct sockaddr* from, socklen_t* fromlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               struct iovec  iov = { (char*)buf, len };
               char          cbuf[1024];
               struct msghdr msg = {
#if (SYSTEM == OS_Darwin)
                  (char*)from,
#else
                  from,
#endif
                  *fromlen,
                  &iov, 1,
                  cbuf, sizeof(cbuf),
                  flags
               };
               int cc = ext_recvmsg2(sockfd,&msg,flags,0);
               if(*fromlen != NULL) {
                  *fromlen = msg.msg_namelen;
               }
               return(cc);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(recvfrom(tdSocket->Socket.SystemSocketID,buf,len,flags,from,fromlen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### recvmsg() wrapper #################################################
int ext_recvmsg2(int sockfd, struct msghdr* msg, int flags,
                 const int receiveNotifications)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      unsigned int     assocID            = 0;
      unsigned short   streamID           = 0;
      unsigned int     protoID            = 0;
      unsigned int     notificationFlags  = 0;
      SocketAddress**  remoteAddressArray = NULL;
      SCTPNotification notification;
      int result = -EOPNOTSUPP;
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
               if((msg == NULL) || (msg->msg_iov == NULL)) {
                  errno_return(-EINVAL);
               }
               msg->msg_flags |= flags;
               if(receiveNotifications) {
                  msg->msg_flags |= MSG_NOTIFICATION;
               }
               else {
                  msg->msg_flags &= ~MSG_NOTIFICATION;
               }
               if(tdSocket->Socket.SCTPSocketDesc.Flags & O_NONBLOCK) {
                  msg->msg_flags |= MSG_DONTWAIT;
               }
               if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                  const size_t oldSize = msg->msg_iov->iov_len;
                  do {
                     msg->msg_iov->iov_len = oldSize;
                     result = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->receive(
                                 (char*)msg->msg_iov->iov_base,
                                 msg->msg_iov->iov_len,
                                 msg->msg_flags,
                                 streamID, protoID);
                  } while((result == -EAGAIN) && !(msg->msg_flags & MSG_DONTWAIT));
                  notificationFlags = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getNotificationFlags();
                  assocID = tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID();
               }
               else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                  const size_t oldSize = msg->msg_iov->iov_len;
                  do {
                     if(remoteAddressArray != NULL) {
                        SocketAddress::deleteAddressList(remoteAddressArray);
                     }
                     msg->msg_iov->iov_len = oldSize;
                     result = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->receiveFrom(
                                 (char*)msg->msg_iov->iov_base,
                                 msg->msg_iov->iov_len,
                                 msg->msg_flags,
                                 assocID, streamID, protoID,
                                 &remoteAddressArray,
                                 notification);
                  } while((result == -EAGAIN) && !(msg->msg_flags & MSG_DONTWAIT));
                  notificationFlags = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getNotificationFlags();
               }
               else {
                  result = -EBADF;
               }

               // ====== Set result to length in case of success ============
               if(result >= 0) {
                  result = msg->msg_iov->iov_len;
               }
           break;
         case ExtSocketDescriptor::ESDT_System: {
               const int result = recvmsg(tdSocket->Socket.SystemSocketID,msg,flags);
               return(result);
            }
           break;
         default:
            errno_return(-ENXIO);
          break;
      }

      if(result >= 0) {
         if((msg->msg_name != NULL)      &&
            (remoteAddressArray != NULL) &&
            (remoteAddressArray[0] != NULL)) {
            msg->msg_namelen = remoteAddressArray[0]->getSystemAddress(
               (sockaddr*)msg->msg_name,
               (socklen_t)msg->msg_namelen,
               tdSocket->Socket.SCTPSocketDesc.Domain);
         }
         else {
            msg->msg_namelen = 0;
         }
         if((notificationFlags & SCTP_RECVDATAIOEVNT) &&
            (msg->msg_control != NULL) &&
            (msg->msg_controllen >= CSpace(sizeof(sctp_sndrcvinfo)))) {
            cmsghdr* cmsg = (cmsghdr*)msg->msg_control;
            cmsg->cmsg_len   = CSpace(sizeof(sctp_sndrcvinfo));
            cmsg->cmsg_level = IPPROTO_SCTP;
            cmsg->cmsg_type  = SCTP_SNDRCV;
            sctp_sndrcvinfo* info = (sctp_sndrcvinfo*)((long)cmsg + (long)sizeof(cmsghdr));
            info->sinfo_stream     = streamID;
            info->sinfo_ssn        = 0;
            info->sinfo_flags      = flags;
            info->sinfo_ppid       = protoID;
            info->sinfo_timetolive = 0;
            info->sinfo_context    = 0;
            info->sinfo_assoc_id   = assocID;
            msg->msg_controllen = CSpace(sizeof(sctp_sndrcvinfo));
         }
         else {
            msg->msg_control    = NULL;
            msg->msg_controllen = 0;
         }
      }
      else {
         msg->msg_namelen    = 0;
         msg->msg_name       = NULL;
         msg->msg_controllen = 0;
         msg->msg_control    = NULL;
      }

      SocketAddress::deleteAddressList(remoteAddressArray);

      errno_return(result);
   }
   errno_return(-EBADF);
}


// ###### recvmsg() wrapper #################################################
int ext_recvmsg(int sockfd, struct msghdr* msg, int flags)
{
   return(ext_recvmsg2(sockfd,msg,flags,1));
}


// ###### send() wrapper ####################################################
int ext_send(int sockfd, const void* msg, size_t len, int flags)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            return(ext_sendto(sockfd,msg,len,flags,NULL,0));
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(send(tdSocket->Socket.SystemSocketID,msg,len,flags));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### sendto() wrapper ####################################################
int ext_sendto(int sockfd, const void* buf, size_t len, int flags,
               const struct sockaddr* to, socklen_t tolen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               struct iovec  iov = { (char*)buf, len };
               struct msghdr msg = {
#if (SYSTEM == OS_Darwin)
                  (char*)to,
#else
                  (void*)to,
#endif
                  tolen,
                  &iov, 1,
                  NULL, 0,
                  flags
               };
               return(ext_sendmsg(sockfd,&msg,flags));
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(sendto(tdSocket->Socket.SystemSocketID,buf,len,flags,to,tolen));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### sendmsg() wrapper #################################################
int ext_sendmsg(int sockfd, const struct msghdr* msg, int flags)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               sctp_sndrcvinfo* info = NULL;
               for(cmsghdr* cmsg = CFirst(msg);
                  cmsg != NULL;
                  cmsg = CNext(msg,cmsg)) {
                  if(cmsg->cmsg_level == IPPROTO_SCTP) {
                     if(cmsg->cmsg_type == SCTP_SNDRCV) {
                        info = (sctp_sndrcvinfo*)CData(cmsg);
                     }
                     else if(cmsg->cmsg_type == SCTP_INIT) {
                        const sctp_initmsg* initmsg = (sctp_initmsg*)CData(cmsg);
                        tdSocket->Socket.SCTPSocketDesc.InitMsg = *initmsg;
                        if((tdSocket->Socket.SCTPSocketDesc.ConnectionOriented) && (msg->msg_name != NULL)) {
                           if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                              errno_return(-EBUSY);
                           }
                           SocketAddress* address = SocketAddress::createSocketAddress(
                                                       0,(sockaddr*)msg->msg_name,msg->msg_namelen);
                           if(address != NULL) {
                              tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr =
                                 tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->associate(
                                    initmsg->sinit_num_ostreams,
                                    *address,
                                    !(tdSocket->Socket.SCTPSocketDesc.Flags & O_NONBLOCK));
                              delete address;
                              if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr == NULL) {
                                 errno_return(-ENOTCONN);
                              }
                           }
                           else {
                              errno_return(-EADDRNOTAVAIL);
                           }
                        }
                     }
                  }
               }
               if((msg != NULL) && (msg->msg_iov != NULL) && (msg->msg_iov->iov_base != NULL)) {
                  flags |= msg->msg_flags;
                  if(tdSocket->Socket.SCTPSocketDesc.Flags & O_NONBLOCK) {
                     flags |= MSG_DONTWAIT;
                  }
                  if(msg->msg_name != NULL) {
                     SocketAddress* destination = SocketAddress::createSocketAddress(
                                                     0,(sockaddr*)msg->msg_name,msg->msg_namelen);
                     if(destination == NULL) {
                        errno_return(-EADDRNOTAVAIL);
                     }
                     int result = -EBADF;
                     if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                        result = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->sendTo(
                                    (char*)msg->msg_iov->iov_base,
                                    msg->msg_iov->iov_len,
                                    flags,
                                    (info != NULL) ? info->sinfo_stream : 0x0000,
                                    (info != NULL) ? info->sinfo_ppid : 0x00000000,
                                    *destination,
                                    tdSocket->Socket.SCTPSocketDesc.InitMsg.sinit_num_ostreams);
                     }
                     delete destination;
                     errno_return(result);
                  }
                  else {
                     if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                        errno_return(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->send(
                                  (char*)msg->msg_iov->iov_base,
                                  msg->msg_iov->iov_len,
                                  flags,
                                  (info != NULL) ? info->sinfo_stream : 0x0000,
                                  (info != NULL) ? info->sinfo_ppid : 0x00000000));
                     }
                     errno_return(-EBADF);
                  }
               }
               errno_return(0);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            return(sendmsg(tdSocket->Socket.SystemSocketID,msg,flags));
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### read() wrapper ####################################################
ssize_t ext_read(int fd, void* buf, size_t count)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(fd);
   if(tdSocket != NULL) {
      if(tdSocket->Type == ExtSocketDescriptor::ESDT_System) {
         return(read(tdSocket->Socket.SystemSocketID,buf,count));
      }
      else {
         return(ext_recv(fd,buf,count,0));
      }
   }
   errno_return(-EBADF);
}


// ###### write() wrapper ####################################################
ssize_t ext_write(int fd, const void* buf, size_t count)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(fd);
   if(tdSocket != NULL) {
      if(tdSocket->Type == ExtSocketDescriptor::ESDT_System) {
         return(write(tdSocket->Socket.SystemSocketID,buf,count));
      }
      else {
         return(ext_send(fd,buf,count,0));
      }
   }
   errno_return(-EBADF);
}


// Internal structure for ext_select().
struct SelectData
{
   cardinal   Conditions;
   int        ConditionFD[FD_SETSIZE];
   cardinal   ConditionType[FD_SETSIZE];
   Condition* ConditionArray[FD_SETSIZE];
   Condition* ParentConditionArray[FD_SETSIZE];
   Condition  GlobalCondition;
   Condition  ReadCondition;
   Condition  WriteCondition;
   Condition  ExceptCondition;
   cardinal   UserCallbacks;
   int        UserCallbackFD[FD_SETSIZE];
   SCTPSocketMaster::UserSocketNotification* UserNotification[FD_SETSIZE];
};


// ###### Add file descriptor to SelectData structure #######################
static int addFD(SelectData&    sd,
                 const int      fd,
                 const cardinal type,
                 Condition&     condition)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(fd);
   if(tdSocket != NULL) {
      if(tdSocket->Type == ExtSocketDescriptor::ESDT_System) {
         for(cardinal i = 0;i < sd.UserCallbacks;i++) {
            if(sd.UserCallbackFD[i] == fd) {
               return(0);
            }
         }
         sd.UserCallbackFD[sd.UserCallbacks] = fd;
         sd.UserNotification[sd.UserCallbacks] = new SCTPSocketMaster::UserSocketNotification;
         if(sd.UserNotification[sd.UserCallbacks] != NULL) {
            sd.UserNotification[sd.UserCallbacks]->FileDescriptor = tdSocket->Socket.SystemSocketID;
            sd.UserNotification[sd.UserCallbacks]->UpdateCondition.addParent(&condition);
            SCTPSocketMaster::MasterInstance.addUserSocketNotification(sd.UserNotification[sd.UserCallbacks]);
            sd.UserCallbacks++;
            return(0);
         }
      }
      else {
         if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {

            sd.ConditionArray[sd.Conditions] =
               tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getUpdateCondition(type);
            sd.ParentConditionArray[sd.Conditions] = &condition;
            sd.ConditionArray[sd.Conditions]->addParent(sd.ParentConditionArray[sd.Conditions]);
            sd.ConditionFD[sd.Conditions]   = fd;
            sd.ConditionType[sd.Conditions] = type;
            sd.Conditions++;
            return(0);
         }
         else if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
            sd.ConditionArray[sd.Conditions] =
               tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getUpdateCondition(type);
            sd.ParentConditionArray[sd.Conditions] = &condition;
            sd.ConditionArray[sd.Conditions]->addParent(sd.ParentConditionArray[sd.Conditions]);
            sd.ConditionFD[sd.Conditions]   = fd;
            sd.ConditionType[sd.Conditions] = type;
            sd.Conditions++;
            return(0);
         }
         else {
            errno_return(-EBADF);
         }
      }
   }
   errno_return(-EBADF);
}


// ###### select() wrapper ##################################################
int ext_select(int             n,
               fd_set*         readfds,
               fd_set*         writefds,
               fd_set*         exceptfds,
               struct timeval* timeout)
{
   SelectData sd;
   sd.Conditions    = 0;
   sd.UserCallbacks = 0;
   sd.GlobalCondition.setName("ext_select::GlobalCondition");
   sd.ReadCondition.setName("ext_select::ReadCondition");
   sd.WriteCondition.setName("ext_select::WriteCondition");
   sd.ExceptCondition.setName("ext_select::ExceptCondition");
   sd.ReadCondition.addParent(&sd.GlobalCondition);
   sd.WriteCondition.addParent(&sd.GlobalCondition);
   sd.ExceptCondition.addParent(&sd.GlobalCondition);

   int result = 0;
   if(readfds != NULL) {
      for(int i = 0;i < min(n,FD_SETSIZE);i++) {
         if(FD_ISSET(i,readfds)) {
            result = addFD(sd,i,SCTPSocket::UCT_Read,sd.ReadCondition);
            if(result != 0) {
               break;
            }
         }
      }
   }
   if(writefds != NULL) {
      for(int i = 0;i < n;i++) {
         if(FD_ISSET(i,writefds)) {
            result = addFD(sd,i,SCTPSocket::UCT_Write,sd.WriteCondition);
            if(result != 0) {
               break;
            }
         }
      }
   }
   if(exceptfds != NULL) {
      for(int i = 0;i < n;i++) {
         if(FD_ISSET(i,exceptfds)) {
            result = addFD(sd,i,SCTPSocket::UCT_Except,sd.ExceptCondition);
            if(result != 0) {
               break;
            }
         }
      }
   }

   if(result == 0) {
      if((sd.Conditions > 0) || (sd.UserCallbacks > 0)) {
         if(timeout != NULL) {
            const card64 delay = ((card64)timeout->tv_sec * (card64)1000000) +
                                    (card64)timeout->tv_usec;
            sd.GlobalCondition.timedWait(delay);
         }
         else {
            sd.GlobalCondition.wait();
         }
      }
      else {
         select(0,NULL,NULL,NULL,timeout);
      }
   }

   if(readfds != NULL) {
      for(cardinal i = 0;i < sd.Conditions;i++) {
         FD_CLR(sd.ConditionFD[i],readfds);
      }
   }
   if(writefds != NULL) {
      for(cardinal i = 0;i < sd.Conditions;i++) {
         FD_CLR(sd.ConditionFD[i],writefds);
      }
   }
   if(exceptfds != NULL) {
      for(cardinal i = 0;i < sd.Conditions;i++) {
         FD_CLR(sd.ConditionFD[i],exceptfds);
      }
   }

   int changes = 0;
   for(cardinal i = 0;i < sd.Conditions;i++) {
      if(sd.ConditionArray[i]->fired()) {
         changes++;
         switch(sd.ConditionType[i]) {
            case SCTPSocket::UCT_Read:
               if(readfds != NULL) {
                  FD_SET(sd.ConditionFD[i],readfds);
               }
             break;
            case SCTPSocket::UCT_Write:
               if(writefds != NULL) {
                  FD_SET(sd.ConditionFD[i],writefds);
               }
             break;
            case SCTPSocket::UCT_Except:
               if(exceptfds != NULL) {
                  FD_SET(sd.ConditionFD[i],exceptfds);
               }
             break;
         }

      }
      sd.ConditionArray[i]->removeParent(sd.ParentConditionArray[i]);
   }
   if(readfds != NULL) {
      for(cardinal i = 0;i < sd.UserCallbacks;i++) {
         FD_CLR(sd.UserCallbackFD[i],readfds);
      }
   }
   if(writefds != NULL) {
      for(cardinal i = 0;i < sd.UserCallbacks;i++) {
         FD_CLR(sd.UserCallbackFD[i],writefds);
      }
   }
   if(exceptfds != NULL) {
      for(cardinal i = 0;i < sd.UserCallbacks;i++) {
         FD_CLR(sd.UserCallbackFD[i],exceptfds);
      }
   }
   for(cardinal i = 0;i < sd.UserCallbacks;i++) {
      bool changed = false;
      if((readfds != NULL) && (sd.UserNotification[i]->Mask & POLLIN)) {
         FD_SET(sd.UserCallbackFD[i],readfds);
         changed = true;
      }
      if((writefds != NULL) && (sd.UserNotification[i]->Mask & POLLOUT)) {
         FD_SET(sd.UserCallbackFD[i],writefds);
         changed = true;
      }
      if((exceptfds != NULL) && (sd.UserNotification[i]->Mask & (~(POLLIN|POLLOUT)))) {
         FD_SET(sd.UserCallbackFD[i],exceptfds);
         changed = true;
      }
      if(changed) {
         changes++;
      }
      SCTPSocketMaster::MasterInstance.deleteUserSocketNotification(sd.UserNotification[i]);
      delete sd.UserNotification[i];
   }
   errno_return((result < 0) ? result : changes);
}


// ###### An implementation of poll(), based on ext_select() ################
int ext_poll(struct pollfd* fdlist, long unsigned int count, int time)
{
   // ====== Prepare timeout setting ========================================
   timeval  timeout;
   timeval* to;
   if(time < 0)
      to = NULL;
   else {
      to = &timeout;
      timeout.tv_sec  = time / 1000;
      timeout.tv_usec = (time % 1000) * 1000;
   }

   // ====== Prepare FD settings ============================================
   int    fdcount = 0;
   fd_set readfdset;
   fd_set writefdset;
   fd_set exceptfdset;
   FD_ZERO(&readfdset);
   FD_ZERO(&writefdset);
   FD_ZERO(&exceptfdset);
   for(unsigned int i = 0; i < count; i++) {
      if(fdlist[i].fd < 0) {
         continue;
      }
      if(fdlist[i].events & POLLIN) {
         FD_SET(fdlist[i].fd,&readfdset);
      }
      if(fdlist[i].events & POLLOUT) {
         FD_SET(fdlist[i].fd,&writefdset);
      }
      FD_SET(fdlist[i].fd,&exceptfdset);
      fdcount++;
   }
   if(fdcount == 0) {
      return(0);
   }
   for(unsigned int i = 0;i < count;i++) {
      fdlist[i].revents = 0;
   }

   // ====== Do ext_select() ================================================
   const int tsize  = getdtablesize();
   int       result = ext_select(tsize,&readfdset,&writefdset,&exceptfdset,to);
   if(result < 0) {
      return(result);
   }

   // ====== Set result flags ===============================================
   for(unsigned int i = 0;i < count;i++) {
      if(FD_ISSET(fdlist[i].fd,&readfdset) && (fdlist[i].events & POLLIN)) {
         fdlist[i].revents |= POLLIN;
      }
      if(FD_ISSET(fdlist[i].fd,&writefdset) && (fdlist[i].events & POLLOUT)) {
         fdlist[i].revents |= POLLOUT;
      }
      if(FD_ISSET(fdlist[i].fd,&exceptfdset)) {
         fdlist[i].revents |= (POLLPRI|POLLERR|POLLHUP);
      }
   }
   return(result);
}


// ###### Check, if SCTP is available #######################################
int sctp_isavailable()
{
   return(SCTPSocketMaster::InitializationResult == 0);
}


// ###### Peel association off ##############################################
int sctp_peeloff(int              sockfd,
                 sctp_assoc_t*    id,
                 struct sockaddr* addr,
                 socklen_t*       addrlen)
{
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               SCTPAssociation* association = NULL;
               if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                  if(tdSocket->Socket.SCTPSocketDesc.Type == SOCK_DGRAM) {
                     if(id != NULL) {
                        association = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->peelOff(*id);
                     }
                     else if((addr != NULL) && (*addrlen != NULL)) {
                        SocketAddress* address = SocketAddress::createSocketAddress(
                                                    0,
                                                    (sockaddr*)addr,
                                                    (socklen_t)*addrlen);
                        if(address != NULL) {
                           association = tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->peelOff(*address);
                           delete address;
                        }
                        else {
                           errno_return(-EINVAL);
                        }
                     }
                  }
               }

               if(association != NULL) {
                  ExtSocketDescriptor newExtSocketDescriptor = *tdSocket;
                  newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPSocketPtr      = NULL;
                  newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr = association;
                  newExtSocketDescriptor.Socket.SCTPSocketDesc.Parent             = sockfd;
                  const int newFD = ExtSocketDescriptorMaster::setSocket(newExtSocketDescriptor);
                  if(newFD < 0) {
                     delete newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr;
                     newExtSocketDescriptor.Socket.SCTPSocketDesc.SCTPAssociationPtr = NULL;
                  }
                  errno_return(newFD);
               }
               errno_return(-EINVAL);
            }
         default:
            errno_return(-EOPNOTSUPP);
           break;
      }
   }
   return(-EBADF);
}


// ###### sctp_getpaddrs() implementation ###################################
int sctp_getlpaddrs(int                       sockfd,
                    sctp_assoc_t              id,
                    struct sockaddr_storage** addrs,
                    const bool                peerAddresses)
{
   *addrs = NULL;
   ExtSocketDescriptor* tdSocket = ExtSocketDescriptorMaster::getSocket(sockfd);
   if(tdSocket != NULL) {
      switch(tdSocket->Type) {
         case ExtSocketDescriptor::ESDT_SCTP:
            {
               int result = -ENXIO;
               SocketAddress** addressArray = NULL;

               // ====== Get local or peer addresses ========================
               if(peerAddresses) {
                  if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                     if((id != 0) && (tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getID() != id)) {
                        result = -EINVAL;
                     }
                     else {
                        tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getRemoteAddresses(addressArray);
                     }
                  }
                  else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getRemoteAddresses(addressArray,id);
                  }
                  else {
                     result = -EBADF;
                  }
               }
               else {
                  if(tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr != NULL) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPAssociationPtr->getLocalAddresses(addressArray);
                  }
                  else if(tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr != NULL) {
                     tdSocket->Socket.SCTPSocketDesc.SCTPSocketPtr->getLocalAddresses(addressArray);
                  }
                  else {
                     result = -EBADF;
                  }
               }

               // ====== Copy addresses into sockaddr_storage array =========
               if(addressArray != NULL) {
                  cardinal count = 0;
                  while(addressArray[count] != NULL) {
                     count++;
                  }
                  if(count > 0) {
                     result = (int)count;
                     *addrs = new sockaddr_storage[count];
                     if(*addrs != NULL) {
                        sockaddr* ptr = (sockaddr*)*addrs;
                        for(cardinal i = 0;i < count;i++) {
                           if(addressArray[i]->getSystemAddress(
                                 ptr,
                                 sizeof(sockaddr_storage),
                                 tdSocket->Socket.SCTPSocketDesc.Domain) <= 0) {
                              result = -ENAMETOOLONG;
                              delete *addrs;
                              *addrs = NULL;
                              break;
                           }
                           ptr = (sockaddr*)((long)ptr + (long)sizeof(sockaddr_storage));
                        }
                     }
                     else {
                        result = -ENOMEM;
                     }
                  }
               }

               SocketAddress::deleteAddressList(addressArray);
               errno_return(result);
            }
          break;
         case ExtSocketDescriptor::ESDT_System:
            errno_return(-EOPNOTSUPP);
          break;
      }
      errno_return(-ENXIO);
   }
   errno_return(-EBADF);
}


// ###### sctp_getpaddrs() implementation ###################################
int sctp_getpaddrs(int sockfd, sctp_assoc_t id, struct sockaddr_storage** addrs)
{
   errno_return(sctp_getlpaddrs(sockfd,id,addrs,true));
}


// ###### sctp_freepaddrs() #################################################
void sctp_freepaddrs(struct sockaddr_storage* addrs)
{
   delete [] addrs;
}


// ###### sctp_getladdrs() implementation ###################################
int sctp_getladdrs(int sockfd, sctp_assoc_t id, struct sockaddr_storage** addrs)
{
   errno_return(sctp_getlpaddrs(sockfd,id,addrs,false));
}


// ###### sctp_freeladdrs() #################################################
void sctp_freeladdrs(struct sockaddr_storage* addrs)
{
   delete [] addrs;
}


// ###### sctp_opt_info() implementation ####################################
int sctp_opt_info(int sd, sctp_assoc_t id, int opt, void* arg)
{
   switch(opt) {
      case SCTP_GET_PEER_ADDR_PARAMS: {
            socklen_t length;
            length = sizeof(sctp_paddrparams);
            ((sctp_paddrparams*)arg)->spp_assoc_id = id;
            return(ext_getsockopt(sd,IPPROTO_SCTP,opt,arg,&length));
         }
       break;
      case SCTP_SET_PEER_ADDR_PARAMS:
         ((sctp_paddrparams*)arg)->spp_assoc_id = id;
         return(ext_setsockopt(sd,IPPROTO_SCTP,opt,arg,sizeof(sctp_paddrparams)));
       break;
      case SCTP_GET_PEER_ADDR_INFO: {
            socklen_t length;
            length = sizeof(sctp_paddrinfo);
            ((sctp_paddrinfo*)arg)->spinfo_assoc_id = id;
            return(ext_getsockopt(sd,IPPROTO_SCTP,opt,arg,&length));
         }
       break;
   }
   errno_return(-EINVAL);
}


#else


// ###### Print error message for unimplemented bindx() call ################
int bindx(int                      sockfd,
          struct sockaddr_storage* addrs,
          int                      addrcnt,
          int                      flags)
{
#ifndef DISABLE_WARNINGS
   cerr << "ERROR: bindx() is currently not a libc function." << endl
        << "       This may be implemented in future releases..." << endl;
#endif
   errno_return(-EOPNOTSUPP);
}


// ###### Check, if SCTP is available #######################################
int sctp_isavailable()
{
   int result = socket(AF_INET,SOCK_SEQPACKET,IPPROTO_SCTP);
   if(result != -1) {
      close(result);
      errno_return(1);
   }
   else {
#ifdef PRINT_NOSCTP_NOTE
      cerr << "SCTP is unsupported on this host!" << endl;
#endif
      errno_return(0);
   }
}


#endif



// ###### Get list of local interface addresses  ############################
int sctp_getifaddrs(sockaddr_storage* addressArray,
                    unsigned int*     addressCount)
{
   SocketAddress** addresses;
   cardinal        count;
   if(Socket::getLocalAddressList(addresses,count,Socket::GLAF_Default)) {
      int result = -EINVAL;
      if(*addressCount >= count) {
         result = 0;
         *addressCount = 0;
         for(cardinal i = 0;i < count;i++) {
            if(addresses[i]->getSystemAddress(
                                (sockaddr*)addressArray,
                                sizeof(sockaddr_storage),
                                AF_UNSPEC)) {
               *addressCount = *addressCount + 1;
               addressArray = (sockaddr_storage*)((long)addressArray + (long)sizeof(sockaddr_storage));
            }
         }
      }
      SocketAddress::deleteAddressList(addresses);
      errno_return(result);
   }
   errno_return(-EIO);
}


// ###### Change OOTB handling ##############################################
void sctp_enableOOTBHandling(const unsigned int enable)
{
   SCTPSocketMaster::enableOOTBHandling(enable != 0);
}
