/*
 *  $Id: socketaddress.cc,v 1.3 2001/08/20 11:23:11 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Socket Address Implementation
 *
 */


#include "tdsystem.h"
#include "socketaddress.h"
#include "internetaddress.h"
#include "unixaddress.h"
#include "tdsocket.h"


namespace Coral {



// ###### Socket address destructor #######################################
SocketAddress::~SocketAddress()
{
   Format = PF_Default;
}


// ###### Delete list of addresses ##########################################
SocketAddress** SocketAddress::newAddressList(const cardinal entries)
{
   SocketAddress** list = new SocketAddress*[entries + 1];
   if(list == NULL) {
#ifndef DISABLE_WARNINGS
      cerr << "SocketAddress::newAddressList() - Out of memory!" << endl;
#endif
      return(NULL);
   }

   for(cardinal i = 0;i < entries + 1;i++) {
      list[i] = NULL;
   }
   return(list);
}


// ###### Delete list of addresses ##########################################
void SocketAddress::deleteAddressList(SocketAddress**& addressArray)
{
   if(addressArray != NULL) {
      cardinal i = 0;
      if(addressArray[i] != NULL) {
         delete addressArray[i];
         addressArray[i] = NULL;
         i++;
      }
      delete addressArray;
      addressArray = NULL;
   }
}


// ###### Get local address #################################################
SocketAddress* SocketAddress::getLocalAddress(const SocketAddress& peer)
{
   SocketAddress* address = createSocketAddress(peer.getFamily());
   if(address != NULL) {
      const int family = peer.getFamily();
      Socket socket(family,Socket::Datagram);
      if(socket.connect(peer)) {
         socket.getSocketAddress(*address);
      }
      address->setPort(0);
   }
   return(address);
}


// ###### Create SocketAddress object #######################################
SocketAddress* SocketAddress::createSocketAddress(const integer family)
{
   SocketAddress* address = NULL;
   switch(family) {
      case AF_INET:
      case AF_INET6:
         address = new InternetAddress();
       break;
      case AF_UNIX:
         address = new UnixAddress();
       break;
      default:
#ifndef DISABLE_WARNINGS
          cerr << "ERROR: SocketAddress::createSocketAddress(family) - "
                  "Unknown address family " << family << "!" << endl;
#endif
       break;
   }
   return(address);
}


// ###### Create SocketAddress object #######################################
SocketAddress* SocketAddress::createSocketAddress(const cardinal flags,
                                                  const String&  name)
{
   // ====== Try to create InternetAddress object ===========================
   InternetAddress* internetAddress;
   if(flags & PF_HidePort) {
      internetAddress = new InternetAddress(name,0);
   }
   else {
      internetAddress = new InternetAddress(name);
   }
   if(internetAddress == NULL) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SocketAddress::createSocketAddress(name) - Out of memory!" << endl;
#endif
   }
   if(internetAddress->isValid()) {
      return(internetAddress);
   }
   delete internetAddress;


   // ====== Try to create InternetAddress object ===========================
   UnixAddress* unixAddress = new UnixAddress(name);
   if(unixAddress == NULL) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SocketAddress::createSocketAddress(name) - Out of memory!" << endl;
#endif
   }
   if(unixAddress->isValid()) {
      return(unixAddress);
   }
   delete unixAddress;


   return(NULL);
}


// ###### Create SocketAddress object with port number ######################
SocketAddress* SocketAddress::createSocketAddress(const cardinal flags,
                                                  const String&  name,
                                                  const card16   port)
{
   // ====== Try to create InternetAddress object ===========================
   InternetAddress* internetAddress = new InternetAddress(name,port);
   if(internetAddress == NULL) {
#ifndef DISABLE_WARNINGS
      cerr << "ERROR: SocketAddress::createSocketAddress(name,port) - Out of memory!" << endl;
#endif
   }
   if(internetAddress->isValid()) {
      return(internetAddress);
   }
   delete internetAddress;

   return(NULL);
}


// ###### Create SocketAddress object from system's sockaddr structure ######
SocketAddress* SocketAddress::createSocketAddress(const cardinal  flags,
                                                  sockaddr*       address,
                                                  const socklen_t length)
{
   switch(address->sa_family) {
      case AF_INET:
      case AF_INET6: {
            InternetAddress* internetAddress = new InternetAddress(address,length);
            if(internetAddress == NULL) {
#ifndef DISABLE_WARNINGS
               cerr << "ERROR: SocketAddress::createSocketAddress(sockaddr) - Out of memory!" << endl;
#endif
            }
            if(internetAddress->isValid()) {
               return(internetAddress);
            }
            delete internetAddress;
         }
       break;
      case AF_UNIX: {
            UnixAddress* unixAddress = new UnixAddress(address,length);
            if(unixAddress == NULL) {
#ifndef DISABLE_WARNINGS
               cerr << "ERROR: SocketAddress::createSocketAddress(sockaddr) - Out of memory!" << endl;
#endif
            }
            if(unixAddress->isValid()) {
               return(unixAddress);
            }
            delete unixAddress;
         }
       break;
      default:
#ifndef DISABLE_WARNINGS
          cerr << "ERROR: SocketAddress::createSocketAddress(sockaddr) - "
                  "Unknown address family " << address->sa_family << "!" << endl;
#endif
       break;
   }
   return(NULL);
}


}
