/*
 *  $Id: dhcpclient.h,v 1.2 2001/11/29 12:25:22 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: DHCP Client
 *
 */


#ifndef DHCPCLIENT_H
#define DHCPCLIENT_H


#include <tdsystem.h>
#include <tdsocket.h>
#include <internetaddress.h>
#include <randomizer.h>
#include <thread.h>
#include <timedthread.h>
#include <networkinterface.h>

#include <set.h>

#include <sys/ioctl.h>
#include <net/ethernet.h>
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <linux/if.h>
#include "dhcp.h"


namespace Coral {



struct DHCPConfiguration
{
   InternetAddress Address;
   InternetAddress Broadcast;
   InternetAddress SubnetMask;
   InternetAddress Router;
   InternetAddress NameServer[3];
   cardinal        NameServers;
   String          Host;
   String          Domain;
   cardinal        LeaseTime;
   cardinal        MessageType;
   InternetAddress Server;
   InternetAddress ServerID;
   card32          TransactionID;
};


/**
  * Output operator.
  */
ostream& operator<<(ostream& os, const DHCPConfiguration& config);



/**
 * This class implements a DHCP client for multiple interfaces as
 * a singleton object. The DHCP client for an interface has the
 * following states: Down, FastDiscovery, SlowDiscovery, Request,
 * Renew, Up and UserShutdown. Each of these states has its own
 * timeout (infinite for UserShutdown). From Down after DownTimeout,
 * the state changes to FastDiscovery. In this state, up to FastDiscoveryTrials
 * DHCP Discovery packets are broadcasted. If no answer is received,
 * the state changes to SlowDiscovery, which continues discovery with
 * a higher timeout. After an acceptable offer is received, the state
 * changes to Request and requests are sent up to RequestTrials. If no
 * DHCP Ack is received, the state changes to Down. Otherwise, the
 * interface is configured and the state changed to Up. After UpTimeout,
 * the state changes to Renew and DHCP Requests are sent up to RenewTimeout
 * to renew the configuration.
 *
 * @short   DHCP Client
 * @author  Thomas Dreibholz (dreibh@exp-math.uni-essen.de)
 * @version 1.0
 */
struct DHCPClient : public SingleTimerThread
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   DHCPClient();

   /**
     * Destructor.
     */
   ~DHCPClient();


   // ====== Start/stop =====================================================
   /**
     * Reimplementation of MultiTimerThread::start();
     *
     * @see MultiTimerThread#start
     */
   bool start(const char* name = "DHCPClient");

   /**
     * Reimplementation of MultiTimerThread::stop();
     *
     * @see MultiTimerThread#stop
     */
   void* stop();


   // ====== Add/delete interface ===========================================
   /**
     * Add interface. The interface configuration will be started automatically.
     *
     * @param name Name (e.g eth0).
     * @param downTimeout Down timeout in microseconds.
     * @param fastDiscoveryTimeout FastDiscovery timeout in microseconds.
     * @param fastDiscoveryTrials FastDiscovery trials.
     * @param slowDiscoveryTimeout SlowDiscovery timeout in microseconds.
     * @param slowDiscoveryTrials SlowDiscovery trials.
     * @param RequestTimeout Request timeout in microseconds.
     * @param requestTrials Request trials.
     * @param renewTimeout Renew timeout in microseconds.
     * @param renewTrials Renew trials.
     * @param upTimeout Up timeout in microseconds.
     */
   bool addInterface(const char*    name,
                     const card64   downTimeout           = 15000000,
                     const card64   fastDiscoveryTimeout  =   500000,
                     const cardinal fastDiscoveryTrials   =        5,
                     const card64   slowDiscoveryTimeout  =  1000000,
                     const cardinal slowDiscoveryTrials   =        5,
                     const card64   requestTimeout        =   400000,
                     const cardinal requestTrials         =       10,
                     const card64   upTimeout             = 15000000,
                     const card64   renewTimeout          =   500000,
                     const cardinal renewTrials           =       10);

   /**
     * Delete interface.
     *
     * @param name Name (e.g eth0).
     * @return true, if interface has been added; false otherwise.
     */
   bool deleteInterface(const char* name);

   /**
     * Start interface configuration by DHCP.
     *
     * @param name Name (e.g eth0).
     * @return true, if interface configuration has been started; false otherwise.
     */
   bool startInterface(const char* name);

   /**
     * Stop interface configuration by DHCP.
     *
     * @param name Name (e.g eth0).
     * @return true, if interface configuration has been stopped; false otherwise.
     */
   bool stopInterface(const char* name);


   // ====== Get configuration ==============================================
   /**
     * Get configuration for interface.
     *
     * @param name Name (e.g eth0).
     * @param config Space to write configuration to.
     * @return true, if there is a vaild configuration; false otherwise.
     */
   bool getConfiguration(const char*        name,
                         DHCPConfiguration& config);


   // ====== Private data ===================================================
   private:
   class ReplyReaderThread : public Thread {
      public:
      ReplyReaderThread();
      ~ReplyReaderThread();
      void run();
      DHCPClient* Client;
   };
   friend class ReplyReaderThread;


   struct DHCPClientInterface {
      card64   NextAction;

      card64   DownTimeout;
      card64   FastDiscoveryTimeout;
      cardinal FastDiscoveryTrials;
      card64   SlowDiscoveryTimeout;
      cardinal SlowDiscoveryTrials;
      card64   RequestTimeout;
      cardinal RequestTrials;
      card64   UpTimeout;
      card64   RenewTimeout;
      cardinal RenewTrials;
      cardinal DesiredLeaseTime;

      enum DHCPInterfaceState {
         DIS_Down          = 0,
         DIS_FastDiscovery = 1,
         DIS_SlowDiscovery = 2,
         DIS_Requesting    = 3,
         DIS_Up            = 4,
         DIS_Renewing      = 5,
         DIS_UserShutdown  = 6
      };

      DHCPInterfaceState State;
      card32             TransactionID;
      cardinal           Trial;
      char               Name[IFNAMSIZ + 1];
      DHCPConfiguration  Config;
   };

   struct RawDHCPPacket
   {
      ether_header Ethernet;
      ip           IP;
      udphdr       UDP;
      dhcp_packet  DHCP;
   }  __attribute__ ((__packed__));


   static const InternetAddress NullAddress;
   static const InternetAddress BroadcastAddress;
   static const InternetAddress LocalOutputAddress;
   static const InternetAddress LocalInputAddress;

   static const card64          IdleTimeout = 1000000;

   static ReplyReaderThread         ReplyReader;
   static set<DHCPClientInterface*> InterfaceSet;
   static Socket                    InputSocket;
   static Socket                    OutputSocket;
   static Randomizer                Random;
   static bool                      Ready;
   static bool                      Initialized;


   void timerEvent(const cardinal timer);
   DHCPClientInterface* findInterface(const char* name);
   static bool sendDiscovery(DHCPClientInterface* interface,
                             card32&              transactionID);
   static bool packetToConfig(const dhcp_packet* packet,
                              const cardinal     length,
                              DHCPConfiguration& config);
   static bool sendRequest(DHCPClientInterface* interface,
                           DHCPConfiguration&   config);
   static cardinal receiveReply(DHCPConfiguration& config);
   static bool configureInterface(DHCPClientInterface* interface);
};


}


#endif
