/*
 *  $Id: gnuplot.cc,v 1.3 2001/11/29 13:18:26 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: GNU Plot Writer
 *
 */

#include "tdsystem.h"
#include "gnuplot.h"
#include "tools.h"


#include <time.h>
#include <fstream.h>


namespace Coral {



// ##########################################################################
// #### GNUPlotData                                                      ####
// ##########################################################################



// ###### Constructor #######################################################
GNUPlotData::GNUPlotData()
{
}


// ###### Destructor ########################################################
GNUPlotData::~GNUPlotData()
{
   close();
}


// ###### Get name ##########################################################
String GNUPlotData::getName() const
{
   return(Name);
}


// ###### Set name ##########################################################
void GNUPlotData::setName(const String& name)
{
   Name = name;
}


// ###### Open ##############################################################
bool GNUPlotData::create(const String& prefix, const cardinal prot)
{
   Name = prefix + ".data";
   ofstream::open(Name.getData(),ios::out,prot);
   return(good());
}



// ##########################################################################
// #### GNUPlotScript                                                    ####
// ##########################################################################



// ###### Constructor #######################################################
GNUPlotScript::GNUPlotScript()
{
   Ready = false;
}


// ###### Destructor ########################################################
GNUPlotScript::~GNUPlotScript()
{
   close();
}


// ###### Open ##############################################################
bool GNUPlotScript::create(const String& prefix)
{
   Name = prefix + ".gp";
   Stream.open(Name.getData());
   Ready = Stream.good();
   if(Ready) {
      Usage      = 0;
      PageNumber = 1;

      // ====== Prepare information string ==================================
      TimeStamp = getMicroTime();
      Stream << "set terminal postscript enhanced color dashed" << endl;
      Stream << "set encoding iso_8859_1" << endl;
      time_t timeStamp = TimeStamp / 1000000;
      const struct tm* timeptr = localtime(&timeStamp);
      char str[256];
      strftime((char*)&str,256,"%A, %B %d %G %T %Z",timeptr);
      InfoString = "Generated on " + (String)str;
      for(cardinal i = 1;i < 65;i++) {
         Stream << "set linestyle " << i << " lt " << i << " lw 5" << endl;
      }

      // ====== Prepare default copyright string ============================
      if(getUserName((char*)&str,sizeof(str),true,getuid()) == true) {
         Copyright = String("Copyright {/Symbol \\323} ") +
                        (1900 + timeptr->tm_year) + " " + str;
      }
      else {
         Copyright = "";
      }
   }
   return(Ready);
}


// ###### Close #############################################################
void GNUPlotScript::close()
{
   Ready = false;
   Stream.close();
}


// ###### Begin new page ####################################################
void GNUPlotScript::newPage(const char* title,
                            const char* xlabel,
                            const char* ylabel,
                            const char* info)
{
   String infoText;
   if(info == NULL) {
      infoText = Copyright;
   }
   else {
      infoText = String(info);
   }
   String infoString = String("Page #") + PageNumber + ", " + InfoString + " - " + infoText;
   Stream << "set timestamp \"" << infoString << "\" 0,0 \"Times-Roman,12\"" << endl;

   const cardinal titleLength = strlen(title);
   cardinal size;
   if(titleLength < 60) {
      size = 16;
   }
   else if(titleLength < 80) {
      size = 14;
   }
   else if(titleLength < 100) {
      size = 13;
   }
   else if(titleLength < 120) {
      size = 12;
   }
   else {
      size = 11;
   }
   Stream << "set title \"" << title << "\" \"Times-Bold," << size << "\"" << endl;
   Stream << "set grid" << endl;
   Stream << "set xlabel \"" << xlabel << "\"" << endl;
   Stream << "set ylabel \"" << ylabel << "\"" << endl;

   PageNumber++;
}


// ###### Plot command ######################################################
void GNUPlotScript::plot(const char*        pagetitle,
                         const char*        linetitle,
                         const char*        xlabel,
                         const char*        ylabel,
                         const char*        info,

                         const bool         ownPage,

                         const GNUPlotData& data,
                         const cardinal     column,

                         const char*        mode,

                         const double       xRangeMin,
                         const double       xRangeMax,
                         const double       yRangeMin,
                         const double       yRangeMax)
{
   if((Usage == 0) || (ownPage == true)) {
      newPage(pagetitle,xlabel,ylabel,info);
      Stream << "plot ";
      if(xRangeMin != xRangeMax) {
         Stream << "[" << xRangeMin << ":" << xRangeMax << "] ";
      }
      else {
         Stream << "[] ";
      }
      if(yRangeMin != yRangeMax) {
         const double yRangeMax2 = 1.1 * yRangeMax;
         Stream << "[" << yRangeMin << ":" << yRangeMax2 << "] ";
      }
      else {
         Stream << "[] ";
      }
   }
   else {
      Stream << ", ";
   }
   Usage++;
   Stream << "\"" << data.Name <<  "\" using 1:" << column << " title \"" << linetitle << "\" with " << mode << " ls " << Usage;
   if(ownPage) {
      Stream << endl;
   }
}


// ###### End multi-line plot command #######################################
void GNUPlotScript::done()
{
   Usage = 0;
   Stream << endl;
}


}
