/*
 *  $Id: gnuplot.h,v 1.2 2001/11/29 12:25:22 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: GNU Plot Writer
 *
 */


#ifndef GNUPLOT_H
#define GNUPLOT_H


#include "tdsystem.h"
#include "strings.h"

#include <fstream.h>


namespace Coral {


/**
  * This is a GNUplot data writer class.
  *
  * @short   GNUPlot Data
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class GNUPlotData : public ofstream
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   GNUPlotData();

   /**
     * Destructor.
     */
   ~GNUPlotData();


   // ====== Open/close =====================================================
   /**
     * Create new data file. The extension ".data" will be added automatically.
     *
     * @param prefix Name prefix.
     * @param prot Protection (default: 664).
     * @return true, if file has been opened; false otherwise.
     */
   bool create(const String& prefix, const cardinal prot = 0664);


   // ====== File name ======================================================
   /**
     * Get file name.
     *
     * @return File name.
     */
   String getName() const;

   /**
     * Set file name.
     *
     * @param name File name.
     */
   void setName(const String& name);


   // ====== Private data ===================================================
   public:
   /**
     * Name of the output file.
     */
   String Name;
};


/**
  * This is a GNUplot script writer class.
  *
  * @short   GNUPlot Data
  * @author  Thomas Dreibholz (Dreibholz@bigfoot.com)
  * @version 1.0
  */
class GNUPlotScript
{
   // ====== Constructor/Destructor =========================================
   public:
   /**
     * Constructor.
     */
   GNUPlotScript();

   /**
     * Destructor.
     */
   ~GNUPlotScript();


   // ====== Open/close =====================================================
   /**
     * Open script file. The extension ".gp" will be added automatically.
     *
     * @param prefix Name prefix.
     * @return true, if file has been opened; false otherwise.
     */
   bool create(const String& prefix);

   /**
     * Close script file.
     */
   void close();


   // ====== Plot ===========================================================
   /**
     * Write "plot" command.
     *
     * @param pagetitle Page title.
     * @param linetitle Line title.
     * @param xlabel X axis label.
     * @param ylabel Y axis label.
     * @param info Information string.
     * @param ownPage true, to begin own page for this plot; false otherwise.
     * @param data GNUPlotData containing data to plot.
     * @pram column Column of data to plot.
     * @param mode Drawing mode for GNU plot. Default is "lines".
     * @param xRangeMin Minimum X value; 0.0 = Auto-scale.
     * @param xRangeMax Maximum X value; 0.0 = Auto-scale.
     * @param yRangeMin Minimum Y value; 0.0 = Auto-scale.
     * @param yRangeMax Maximum Y value; 0.0 = Auto-scale.
     */
   void plot(const char*        pagetitle,
             const char*        linetitle,
             const char*        xlabel,
             const char*        ylabel,
             const char*        info,

             const bool         ownPage,

             const GNUPlotData& data,
             const cardinal     column,

             const char*        mode = "lines",

             const double       xRangeMin = 0.0,
             const double       xRangeMax = 0.0,
             const double       yRangeMin = 0.0,
             const double       yRangeMax = 0.0);

   /**
     * End multi-line plot command.
     */
   void done();


   // ====== Private data ===================================================
   private:
   void newPage(const char* title,
                const char* xlabel,
                const char* ylabel,
                const char* info);


   cardinal Usage;
   cardinal PageNumber;
   String   Name;
   String   Copyright;
   ofstream Stream;
   card64   TimeStamp;
   String   InfoString;
   bool     Ready;
};


}


#endif
