/*
 *  $Id: networkinterface.cc,v 1.1 2001/11/28 11:46:49 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Network Interface
 *
 */


#include "tdsystem.h"
#include "networkinterface.h"
#include "internetaddress.h"

#include <sys/ioctl.h>



namespace Coral {



// ###### Static data #######################################################
int NetworkInterface::ControlSocket = socket(AF_INET,SOCK_DGRAM,0);


// ###### Constructor #######################################################
NetworkInterface::NetworkInterface(const char* name)
{
   strncpy(Name,name,IFNAMSIZ);
   Name[IFNAMSIZ] = 0x00;
}


// ###### Destructor ########################################################
NetworkInterface::~NetworkInterface()
{
}


// ###### Get name ##########################################################
const char* NetworkInterface::getName() const
{
   return(Name);
}


// ###### Check, if interface is installed ##################################
bool NetworkInterface::isInstalled()
{
   unsigned int flags;
   return(getFlags(flags));
}


// ###### Get flags #########################################################
bool NetworkInterface::getFlags(unsigned int& flags)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIFFLAGS,&ifr) < 0) {
      flags = 0;
      return(false);
   }
   flags = ifr.ifr_flags;
   return(true);
}


// ###### Set flags #########################################################
bool NetworkInterface::setFlags(const unsigned int flags)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   ifr.ifr_flags = flags;
   if(ioctl(ControlSocket,SIOCSIFFLAGS,&ifr) < 0) {
      return(false);
   }
   return(true);
}


// ###### Get MTU ###########################################################
bool NetworkInterface::getMTU(unsigned int& mtu)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIFMTU,&ifr) < 0) {
      mtu = 0;
      return(false);
   }
   mtu = ifr.ifr_mtu;
   return(true);
}


// ###### Set MTU ###########################################################
bool NetworkInterface::setMTU(const unsigned int mtu)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   ifr.ifr_mtu = mtu;
   if(ioctl(ControlSocket,SIOCSIFMTU,&ifr) < 0) {
      return(false);
   }
   return(true);
}


// ###### Get hardware address ##############################################
bool NetworkInterface::getHardwareAddress(sockaddr& address)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIFHWADDR,&ifr) < 0) {
      memset((char*)&address,0,sizeof(sockaddr));
      return(false);
   }
   address = ifr.ifr_hwaddr;
   return(true);
}


// ###### Set handware address ##############################################
bool NetworkInterface::setHardwareAddress(const sockaddr& address)
{
   ifreq ifr;
   strcpy(ifr.ifr_name,Name);
   ifr.ifr_hwaddr = address;
   if(ioctl(ControlSocket,SIOCSIFHWADDR,&ifr) < 0) {
      return(false);
   }
   return(true);
}


// ###### Set interface up ##################################################
bool NetworkInterface::up()
{
   unsigned int flags;
   if(getFlags(flags) == true) {
      flags |= IFF_UP|IFF_RUNNING;
      return(setFlags(flags));
   }
   return(false);
}


// ###### Set interface down ################################################
bool NetworkInterface::down()
{
   unsigned int flags;
   if(getFlags(flags) == true) {
      flags &= ~(IFF_UP|IFF_RUNNING);
      return(setFlags(flags));
   }
   return(false);
}


// ###### Add address #######################################################
bool NetworkInterface::addAddress(const AddressType    addrType,
                                  const SocketAddress& address,
                                  const cardinal       prefixLength)
{
   sockaddr_storage sa;
   int              family = address.getFamily();
   if(family == AF_INET6) {
      if(address.getSystemAddress((sockaddr*)&sa,sizeof(sa),AF_INET)) {
         family = AF_INET;
      }
   }

   if(address.getSystemAddress((sockaddr*)&sa,sizeof(sa),family)) {
      const int sockfd = socket(family,SOCK_DGRAM,0);
      if(socket >= 0) {
         if(family != AF_INET6) {
            ifreq ifr;
            strcpy(ifr.ifr_name,Name);
            memcpy(&ifr.ifr_addr,&sa,sizeof(struct sockaddr));
            int command;
            switch(addrType) {
               case AT_Netmask:
                  command = SIOCSIFNETMASK;
                break;
               case AT_Broadcast:
                  command = SIOCSIFBRDADDR;
                break;
               case AT_Destination:
                  command = SIOCSIFDSTADDR;
                break;
               case AT_Address:
               default:
                  command = SIOCSIFADDR;
                break;
            }
            const int result = ioctl(sockfd,command,&ifr);
            close(sockfd);
            return((result < 0) ? false : true);
         }
         else {
            ifreq ifr;
            strcpy(ifr.ifr_name,Name);
            if(ioctl(sockfd,SIOCGIFINDEX,&ifr) >= 0) {
               in6_ifreq ifr6;
               ifr6.ifr6_ifindex = ifr.ifr_ifindex;
               sockaddr_in6* sa6 = (sockaddr_in6*)&sa;
               memcpy((char*)&ifr6.ifr6_addr,(char*)&sa6->sin6_addr,sizeof(in6_addr));
               ifr6.ifr6_prefixlen = (int)prefixLength;
               const int result = ioctl(sockfd,SIOCSIFADDR,&ifr6);
               close(sockfd);
               return((result < 0) ? false : true);
            }
            close(sockfd);
            return(false);
         }
      }
   }
   return(false);
}


// ###### Delete address ####################################################
bool NetworkInterface::deleteAddress(const AddressType    addrType,
                                     const SocketAddress& address,
                                     const cardinal       prefixLength)
{
   int family = address.getFamily();
   if(family != AF_INET6) {
   puts("BAD TYPE!");
      return(false);
   }

   sockaddr_storage sa;
   if(address.getSystemAddress((sockaddr*)&sa,sizeof(sa),family)) {
      const int sockfd = socket(family,SOCK_DGRAM,0);
      if(socket >= 0) {
         ifreq ifr;
         strcpy(ifr.ifr_name,Name);
         if(ioctl(sockfd,SIOCGIFINDEX,&ifr) >= 0) {
            in6_ifreq ifr6;
            ifr6.ifr6_ifindex = ifr.ifr_ifindex;
            sockaddr_in6* sa6 = (sockaddr_in6*)&sa;
            memcpy((char*)&ifr6.ifr6_addr,(char*)&sa6->sin6_addr,sizeof(in6_addr));
            ifr6.ifr6_prefixlen = (int)prefixLength;
            const int result = ioctl(sockfd,SIOCDIFADDR,&ifr6);
            close(sockfd);
            return((result < 0) ? false : true);
         }
         close(sockfd);
         return(false);
      }
   }
   return(false);
}


// ###### Get address #######################################################
SocketAddress* NetworkInterface::getAddress(const AddressType addrType,
                                            const int         afFamily,
                                            const cardinal    index,
                                            cardinal*         prefixLength)
{
   if(afFamily != AF_INET6) {
      if(index != 0) {
         return(NULL);
      }

      const int sockfd = socket(afFamily,SOCK_DGRAM,0);
      if(socket < 0) {
         return(NULL);
      }

      ifreq ifr;
      strcpy(ifr.ifr_name,Name);
      int command;
      switch(addrType) {
         case AT_Netmask:
            command = SIOCGIFNETMASK;
          break;
         case AT_Broadcast:
            command = SIOCGIFBRDADDR;
          break;
         case AT_Destination:
            command = SIOCGIFDSTADDR;
          break;
         case AT_Address:
         default:
            command = SIOCGIFADDR;
          break;

      }
      const int result = ioctl(sockfd,command,&ifr);
      close(sockfd);
      if(result < 0) {
         return(NULL);
      }

      sockaddr* saddr;
      switch(addrType) {
         case AT_Netmask:
            saddr = (sockaddr*)&ifr.ifr_netmask;
          break;
         case AT_Broadcast:
            saddr = (sockaddr*)&ifr.ifr_broadaddr;
          break;
         case AT_Destination:
            saddr = (sockaddr*)&ifr.ifr_dstaddr;
          break;
         case AT_Address:
         default:
            saddr = (sockaddr*)&ifr.ifr_addr;
          break;
      }
      SocketAddress* sa = SocketAddress::createSocketAddress(0,
                             saddr, sizeof(sockaddr));
      if(sa) {
         sa->setPrintFormat(SocketAddress::PF_Address|SocketAddress::PF_HidePort);
      }
      if(prefixLength) {
         *prefixLength = 0;
      }   
      return(sa);
   }
   else {
      FILE* fh = fopen("/proc/net/if_inet6","r");
      if(fh == NULL) {
         return(NULL);
      }

      char addr6[40];
      char devname[20];
      int  plen;
      int  scope;
      int  dad_status;
      int  if_idx;
      char addr6p[8][5];

      cardinal number = 0;
      while(fscanf(fh,
                   "%4s%4s%4s%4s%4s%4s%4s%4s %02x %02x %02x %02x %20s\n",
                   addr6p[0], addr6p[1], addr6p[2], addr6p[3],
                   addr6p[4], addr6p[5], addr6p[6], addr6p[7],
                   &if_idx, &plen, &scope, &dad_status, devname) != EOF) {
         if(!strcmp(devname,Name)) {
            if(number == index) {
               snprintf(addr6, sizeof(addr6),
                        "%s:%s:%s:%s:%s:%s:%s:%s",
                        addr6p[0], addr6p[1], addr6p[2], addr6p[3],
                        addr6p[4], addr6p[5], addr6p[6], addr6p[7]);
               fclose(fh);
               if(prefixLength) {
                  *prefixLength = (cardinal)plen;
               }
               SocketAddress *sa = new InternetAddress(addr6,0);
               if(sa) {
                  sa->setPrintFormat(SocketAddress::PF_Address|SocketAddress::PF_HidePort);
               }
               return(sa);
            }
            number++;
         }
      }

      fclose(fh);
      return(NULL);
   }
}


}
