/*
 *  $Id: wirelessnetworkinterface.cc,v 1.2 2001/12/05 12:47:02 dreibh Exp $
 *
 * SCTP implementation according to RFC 2960.
 * Copyright (C) 1999-2001 by Thomas Dreibholz
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * There are two mailinglists available at www.sctp.de which should be used for
 * any discussion related to this implementation.
 *
 * Contact: discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Wireless Network Interface
 *
 */


#include "tdsystem.h"
#include "wirelessnetworkinterface.h"


#if (SYSTEM == OS_Linux)
#include <ctype.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>
#endif


namespace Coral {



// ###### Constructor #######################################################
WirelessNetworkInterface::WirelessNetworkInterface(const char* name)
   : NetworkInterface(name)
{
}


// ###### Destructor ########################################################
WirelessNetworkInterface::~WirelessNetworkInterface()
{
}


// ###### Check, if interface is wireless ###################################
bool WirelessNetworkInterface::isWireless()
{
#if (SYSTEM == OS_Linux)
   iwreq wrq;
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWNAME,&wrq) < 0) {
      return(false);
   }
   return(true);
#else
   return(false);
#endif
}


// ###### Get configuration #################################################
bool WirelessNetworkInterface::getConfig(WirelessConfig& config)
{
#if (SYSTEM == OS_Linux)
   memset((char*)&config,0,sizeof(config));

   // ====== Check, if wireless extensions are available ====================
   iwreq wrq;
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWNAME,&wrq) < 0) {
      return(false);
   }
   else {
      strcpy((char*)&config.Name,wrq.u.name);
   }

   // ====== Get range ======================================================
   strcpy(wrq.ifr_name,Name);
   wrq.u.data.pointer = (caddr_t)((char*)&config.Range);
   wrq.u.data.length  = 0;
   wrq.u.data.flags   = 0;
   if(ioctl(ControlSocket,SIOCGIWRANGE,&wrq) >= 0) {
      config.HasRange = true;

      if((WIRELESS_EXT > 10) && (wrq.u.data.length >= 300)) {
#if WIRELESS_EXT > 10
         if(config.Range.we_version_compiled != WIRELESS_EXT) {
            cerr << "WARNING: Wireless version incompatibility for device "
                 << Name << "!" << endl;
            return(false);
         }
#endif
      }
      else {
         if(wrq.u.data.length != sizeof(iw_range)) {
            cerr << "WARNING: Wireless version incompatibility for device "
                 << Name << "!" << endl;
            return(false);
         }
      }
   }

   // ====== Get NWID =======================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWNWID,&wrq) >= 0) {
      config.HasNWID = true;
      config.NWID    = wrq.u.nwid;
   }

   // ====== Get frequency ==================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWFREQ,&wrq) >= 0) {
      config.HasFrequency = true;
      config.Frequency    = (double)wrq.u.freq.m * pow(10,wrq.u.freq.e);
   }

   // ====== Get sensitivity ================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWSENS,&wrq) >= 0) {
      config.HasSensitivity = true;
      config.Sensitivity    = wrq.u.sens;
   }

   // ====== Get key ========================================================
   strcpy(wrq.ifr_name,Name);
   wrq.u.data.pointer = (caddr_t)config.Key;
   wrq.u.data.length  = 0;
   wrq.u.data.flags   = 0;
   if(ioctl(ControlSocket,SIOCGIWENCODE,&wrq) >= 0) {
      config.HasKey   = true;
      config.KeySize  = wrq.u.data.length;
      config.KeyFlags = wrq.u.data.flags;
   }

   // ====== Get ESSID ======================================================
   strcpy(wrq.ifr_name,Name);
   wrq.u.data.pointer = (caddr_t)config.ESSID;
   wrq.u.data.length  = 0;
   wrq.u.data.flags   = 0;
   if(ioctl(ControlSocket,SIOCGIWESSID,&wrq) >= 0) {
      config.HasESSID = true;
      config.ESSID_On = wrq.u.data.flags;
   }

   // ====== Get AP address =================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWAP,&wrq) >= 0) {
      config.HasAPAddress = true;
      config.APAddress    = wrq.u.ap_addr;
   }

   // ====== Get nickname ===================================================
   strcpy(wrq.ifr_name,Name);
   wrq.u.data.pointer = (caddr_t)config.Nickname;
   wrq.u.data.length  = 0;
   wrq.u.data.flags   = 0;
   if(ioctl(ControlSocket,SIOCGIWNICKN,&wrq) >= 0) {
      if(wrq.u.data.length > 1) {
         config.HasNickname = true;
      }
   }

   // ====== Get bitrate ====================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWRATE,&wrq) >= 0) {
      config.HasBitrate = true;
      config.Bitrate    = wrq.u.bitrate;
   }

   // ====== Get RTS ========================================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWRTS,&wrq) >= 0) {
      config.HasRTSThreshold = true;
      config.RTSThreshold    = wrq.u.rts;
   }

   // ====== Get fragmentation threshold ====================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWFRAG,&wrq) >= 0) {
      config.HasFragmentationThreshold = true;
      config.FragmentationThreshold    = wrq.u.frag;
   }

   // ====== Get operation mode =============================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWMODE,&wrq) >= 0) {
      if((wrq.u.mode < 6) && (wrq.u.mode >= 0)) {
         config.HasOperationMode = true;
         config.OperationMode    = wrq.u.mode;
      }
   }

   // ====== Get power management settings ==================================
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWPOWER,&wrq) >= 0) {
      config.HasPowerManagement = true;
      config.PowerManagement    = wrq.u.power;
   }

   // ====== Get transmit power =============================================
#if WIRELESS_EXT > 9
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWTXPOW,&wrq) >= 0) {
      config.HasTransmitPower = true;
      config.TransmitPower    = wrq.u.txpower;
   }
#else
#warning You are using an old version of Wireless Extensions (<= 9). Update to 11 or better.
#endif

   // ====== Get retry settings =============================================
#if WIRELESS_EXT > 9
   strcpy(wrq.ifr_name,Name);
   if(ioctl(ControlSocket,SIOCGIWRETRY,&wrq) >= 0) {
      config.HasRetry = true;
      config.Retry    = wrq.u.retry;
   }
#else
#warning You are using an old version of Wireless Extensions (<= 10). Update to 11 or better.
#endif

   return(true);
#else
   return(false);
#endif
}


// ###### Get link status ###################################################
bool WirelessNetworkInterface::getLinkStatus(WirelessLinkStatus& status)
{
#if (SYSTEM == OS_Linux)
   iwreq wrq;
   iw_range range;
   strcpy(wrq.ifr_name,Name);
   wrq.u.data.pointer = (caddr_t)((char*)&range);
   wrq.u.data.length  = 0;
   wrq.u.data.flags   = 0;
   if(ioctl(ControlSocket,SIOCGIWRANGE,&wrq) >= 0) {
      status.MaxLinkQuality = range.max_qual.qual;
      status.MaxSignal      = range.max_qual.level;
      status.MaxNoise       = range.max_qual.noise;
   }
   else {
      status.MaxLinkQuality = 0;
      status.MaxSignal      = 0;
      status.MaxNoise       = 0;
   }

   FILE* input = fopen("/proc/net/wireless","r");
   if(input == NULL) {
      return(false);
   }
   bool result = false;
   char buffer[256];
   while(fgets(buffer,sizeof(buffer) - 1,input)) {
      char* bp = (char*)&buffer;
      while((*bp) && (isspace(*bp))) {
         bp++;
         if((strncmp(bp,Name,strlen(Name)) == 0) &&
            (bp[strlen(Name)] == ':')) {
            bp = strchr(bp,':');
            bp++;

            unsigned int ifstatus     = 0x0000;
            unsigned int linkQuality  = 0;
            unsigned int linkSignal   = 0;
            unsigned int linkNoise    = 0;
            unsigned int discardNWID  = 0;
            unsigned int discardCrypt = 0;
            unsigned int discardMisc  = 0;
            unsigned int pos          = 0;
            const int ok = sscanf(bp,"%X %d. %d. %d. %d %d %d%n",
                              &ifstatus, &linkQuality, &linkSignal, &linkNoise,
                              &discardNWID, &discardCrypt, &discardMisc,
                              &pos);
            if(ok >= 1) {
               status.Status       = ifstatus;
               status.LinkQuality  = linkQuality;
               status.Signal       = linkSignal;
               status.Noise        = linkNoise;
               status.DiscardNWID  = discardNWID;
               status.DiscardCrypt = discardCrypt;
               status.DiscardMisc  = discardMisc;
               result = true;
            }
            break;
          }
       }
   }
   fclose(input);
   return(result);
#else
   return(false);
#endif
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const WirelessConfig& config)
{
#if (SYSTEM == OS_Linux)
   char str[256];
   static const char* operationMode[] = {
      "Auto",
      "Ad-Hoc",
      "Managed",
      "Master",
      "Repeater",
      "Secondary"
   };

   os << "   Name          = " << config.Name << endl;
   if(config.HasNickname) {
      os << "   Nickname      = " << config.Nickname << endl;
   }

   if(config.HasOperationMode) {
      os << "   Op Mode       = " << operationMode[config.OperationMode] << endl;
   }

   if(config.HasESSID) {
      os << "   ESSID         = ";
      if(config.ESSID_On) {
         if((config.ESSID_On & IW_ENCODE_INDEX) > 1) {
            os << config.ESSID << " [" << (config.ESSID_On & IW_ENCODE_INDEX) << "]" << endl;
         }
         else {
            os << config.ESSID << endl;
         }
      }
      else {
         os << "off" << endl;
      }
   }

   if(config.HasNWID) {
      if(config.NWID.disabled) {
         strcpy((char*)&str,"off/any");
      }
      else {
         snprintf((char*)&str,sizeof(str),"$%08x",config.NWID.value);
      }
      os << "   NWID           = " << str << endl;
   }

   if(config.HasSensitivity) {
      if(config.HasRange) {
         os << "   Sensitivity   = " << config.Sensitivity.value << "/"
                                     << config.Range.sensitivity << endl;
      }
      else {
         os << "   Sensitivity   = " << config.Sensitivity.value << " [dBm]" << endl;
      }
   }

   if(config.HasBitrate) {
      os << "   Bitrate       = " << config.Bitrate.value << " [Bits/s]" << endl;
   }

   if(config.HasFrequency) {
      if(config.Frequency < 1000) {
         os << "   Channel       = " << config.Frequency << endl;
      }
      else {
         os << "   Frequency     = " << config.Frequency / 1000000000.0 << " [GHz]" << endl;
      }
   }

   if(config.HasRTSThreshold) {
      os << "   RTS Threshold = ";
      if(config.RTSThreshold.disabled) {
         os << "off";
      }
      else {
         os << config.RTSThreshold.value;
         if(config.RTSThreshold.fixed) {
            os << " fixed";
         }
      }
      os << endl;
   }

   if(config.HasFragmentationThreshold) {
      os << "   Frgmt. Thres. = ";
      if(config.FragmentationThreshold.disabled) {
         os << "off";
      }
      else {
         os << config.FragmentationThreshold.value;
         if(config.FragmentationThreshold.fixed) {
            os << " fixed";
         }
      }
      os << endl;
   }

#if WIRELESS_EXT > 9
   if(config.HasTransmitPower) {
      os << "   Transmit Pwr. = ";
      if(config.TransmitPower.disabled) {
         os << "off";
      }
      else {
         os << config.TransmitPower.value;
         if(config.TransmitPower.fixed) {
            os << " fixed";
         }
         if(config.TransmitPower.flags & IW_TXPOW_MWATT) {
            os << " [mWatt]";
         }
         else {
            os << " [dBm]";
         }
      }
      os << endl;
   }
#endif

   if(config.HasRetry) {
      os << "   Retry = " << config.Retry.fixed << endl;
   }
#endif
   return(os);
}


// ###### Output operator ###################################################
ostream& operator<<(ostream& os, const WirelessLinkStatus& status)
{
#if (SYSTEM == OS_Linux)
   char str[256];

   snprintf((char*)&str,sizeof(str),"$%04x",status.Status);
   os << "   Status        = " << str                 << endl;
   os << "   Link Quality  = " << status.LinkQuality  << "/"
                               << status.MaxLinkQuality << endl;
   if((status.MaxLinkQuality) && (status.Signal <= (integer)status.MaxLinkQuality)) {
      os << "   Signal        = " << status.Signal << "/"
                                  << status.MaxSignal << endl
         << "   Noise         = " << status.Noise << "/"
                                  << status.MaxNoise  << endl;
   }
   else {
      os << "   Signal        = " << status.Signal - 0x100 << " [dBm]" << endl
         << "   Noise         = " << status.Noise - 0x100  << " [dBm]" << endl;
   }
   os << "   Discard NWID  = " << status.DiscardNWID  << endl
      << "   Discard Crypt = " << status.DiscardCrypt << endl
      << "   Discard Misc  = " << status.DiscardMisc  << endl;
#endif
   return(os);
}


}
